#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2013, 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import time

import upload_package
import watch_ppa

from ci_utils import (
    amqp_utils,
    amqp_worker,
)

logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)

TIME_BETWEEN_CHECKS = 10


def create_result_file(datastore, content):
    '''Creates a result file based on the final upload and build status.

    The input content has the following format:
        {
            'message': 'The high level status of the upload',
            'subticket_status': [{
                'name': source_package_name,
                'version': source_package_version,
                'step': numeric_workflow_step,
                'step_text': text_workflow_step,
                'status': numeric_workflow_status,
                'status_text': text_workflow_status,
                'message': subticket_specific_message,
            }]
        }

    The return is an artifact entry for the progress message.'''
    package_build_result_filename = 'package_build.output.log'
    body = '{}\n\n'.format(content['message'])
    body += 'Uploaded source packages:'
    for subticket_status in content['subticket_status']:
        body += '\nname: {}\n'.format(subticket_status['name'])
        body += 'version: {}\n'.format(subticket_status['version'])
        body += 'status: {}\n'.format(subticket_status['status_text'])
        if subticket_status.get('message'):
            body += 'message: {}\n'.format(subticket_status['message'])
    location = datastore.put_file(
        package_build_result_filename, str(body), 'text/plain')
    return {
        'name': package_build_result_filename,
        'type': 'LOGS',
        'reference': location,
    }


class BSBuilderWorker(amqp_worker.AMQPWorker):
    def __init__(self):
        super(BSBuilderWorker, self).__init__('bsbuilder')

    def handle_request(self, log, params):
        subtickets = params['subtickets']
        series = params['series']
        ppa = params['ppa']
        archive = params['archive']
        trigger = params['progress_trigger']

        datastore = self._create_data_store(params['ticket_id'])

        # Setup the output data to send back to the caller
        out_data = {
            'message': 'Build in progress.',
            'ppa': ppa,
            'archive': archive,
            'artifacts': [],
        }

        upload_list = upload_package.create_upload_list(subtickets)
        log.info('upload_list: {}'.format(upload_list))

        if params.get('cancelled'):
            log.error('request has been cancelled, exiting')
            return amqp_utils.progress_failed, out_data

        # Check that an upload isn't going to fail due to version
        (failed, subticket_status) = watch_ppa.check_ppa(
            series, ppa, archive, None, upload_list)
        if failed:
            # This is the end of the line, the upload is blocked.
            out_data['subticket_status'] = subticket_status
            message = 'Source package upload failed.'
            log.error(message)
            out_data['message'] = message
            out_data['artifacts'].append(
                create_result_file(datastore, out_data))
            amqp_utils.progress_failed(trigger, out_data)
        else:
            # After checking the PPA, it's now safe to do the upload
            upload_package.upload_source_packages(ppa, upload_list)

            # Set ret to 1, the upload is failed until it succeeds
            ret = 1
            start_time = time.time()
            while True:
                if params.get('cancelled'):
                    log.error('request has been cancelled, exiting')
                    return amqp_utils.progress_failed, out_data

                (ret, subticket_status) = watch_ppa.watch_ppa(
                    datastore, start_time, series, ppa, archive, None,
                    upload_list)
                log.info('subticket_status: {}'.format(subticket_status))
                out_data['subticket_status'] = subticket_status
                if ret == -1:
                    # -1 indicates that the task is not finished
                    # Send regular progress back to the lander
                    amqp_utils.progress_update(trigger, out_data)
                    log.info('Going to sleep for '
                             '{}'.format(TIME_BETWEEN_CHECKS))
                    time.sleep(TIME_BETWEEN_CHECKS)
                else:
                    log.info('All done')
                    break

            if ret == 1:
                # '1' indicates a failure occurred during the upload/build
                message = 'A package failed to upload or build.'
                log.error(message)
                out_data['message'] = message
                out_data['artifacts'].append(
                    create_result_file(datastore, out_data))
                amqp_utils.progress_failed(trigger, out_data)
            else:
                # '0' indicates the build was successful
                message = 'The package build was successful.'
                log.error(message)
                out_data['message'] = message
                out_data['artifacts'].append(
                    create_result_file(datastore, out_data))
                amqp_utils.progress_completed(trigger, out_data)


if __name__ == '__main__':
    BSBuilderWorker().main('bsbuilder')
