#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import atexit
from dput.changes import parse_changes_file
import logging
import os
import re
import shutil
import sys
import tempfile
import urllib2

from cupstream2distro import packagemanager


def parse_arguments():
    parser = argparse.ArgumentParser(
        description='Source package upload handler.')
    parser.add_argument('-p', '--ppa',
                        required=True,
                        help='Target PPA for source package upload.')
    parser.add_argument('-s', '--source',
                        required=True,
                        action='append',
                        help='A source package file to include in the upload.')
    return parser.parse_args()


def cleanup(directory):
    if os.path.exists(directory) and os.path.isdir(directory):
        logging.info('Removing temp directory: {}'.format(directory))
        shutil.rmtree(directory)


def create_temp_dir():
    directory = tempfile.mkdtemp()
    atexit.register(cleanup, directory)
    return directory


def get_url_contents(url):
    try:
        req = urllib2.Request(url)
        f = urllib2.urlopen(req)
        response = f.read()
        f.close()
        return response
    except IOError as e:
        raise EnvironmentError('Failed to open url [{}]: {}'.format(url, e))


def get_source_files(subtickets):
    '''Retrieves the source package files and creates initial upload_list.'''
    upload_list = []
    for subticket in subtickets:
        package = {}
        spu = subticket['source_package_upload']
        package['id'] = spu['id']
        package['name'] = spu['sourcepackage']['name']
        package['version'] = spu['version']
        package['resource'] = subticket['resource_uri']
        package['location'] = create_temp_dir()
        package['files'] = []
        for artifact in subticket['artifact']:
            source = artifact['reference']
            file_name = os.path.basename(source)
            file_path = os.path.join(package['location'], file_name)
            logging.info('Retrieving source file: {}'.format(file_name))
            with open(file_path, 'w') as out_file:
                out_file.write(get_url_contents(source))
            package['files'].append(file_name)
        upload_list.append(package)
    return upload_list


def parse_dsc_file(dsc_file):
    regexp = re.compile("^Architecture: (.*)\n")
    with open(dsc_file) as in_file:
        for line in in_file:
            arch_lists = regexp.findall(line)
            if arch_lists:
                return arch_lists[0]


def parse_source_files(upload_list):
    '''Perform a second pass of the upload list and parse out more info.

    The purpose of this is to make sure all source package files have
    been downloaded and identify the source package architecture.
    This must be done after all files in a subticket have been acquired.'''
    for package in upload_list:
        for source in package['files']:
            source_name = os.path.join(package['location'], source)
            if source.endswith('changes'):
                changes = parse_changes_file(filename=source_name,
                                             directory=package['location'])
                for package_file in changes.get_files():
                    if os.path.exists(package_file):
                        if package_file.endswith('.dsc'):
                            package['architecture'] = parse_dsc_file(
                                package_file)
                    else:
                        raise EnvironmentError(
                            'Not found: {}'.format(package_file))


def create_upload_list(subtickets):
    '''Parses the subtickets into the upload list.'''
    upload_list = get_source_files(subtickets)
    parse_source_files(upload_list)
    return upload_list


def upload_source_packages(ppa, upload_list):
    '''Attempts source file upload into the PPA.'''
    logging.info('Upload to the ppa: {}'.format(ppa))
    for package in upload_list:
        packagemanager.upload_package(package['name'], package['version'],
                                      ppa, package['location'])

        directory = package['location']
        if os.path.exists(directory) and os.path.isdir(directory):
            logging.info('Removing temp directory: {}'.format(directory))
            shutil.rmtree(directory)


def main():
    logging.basicConfig(level=logging.INFO, format="%(asctime)s %(levelname)s "
                        "%(message)s")

    args = parse_arguments()
    upload_source_packages(args.ppa, args.source)


if __name__ == '__main__':
    sys.exit(main())
