#!/usr/bin/env python

import base64
import os
import json
import shutil
import subprocess
import sys
import tempfile
import textwrap
import charmhelpers.fetch


def juju_info(msg):
    subprocess.check_call(['juju-log', '-l', 'INFO', msg])
    pass


def _config():
    output = subprocess.check_output(['config-get', '--format=json'])
    return json.loads(output)


def _relation_get(key):
    return subprocess.check_output(['relation-get', key]).strip()


def _relation_set(keyvalues, relation_id=None):
    args = ['relation-set']
    if relation_id:
        args.extend(['-r', relation_id])
    args.extend(["{}={}".format(k, v or '') for k, v in keyvalues.items()])
    subprocess.check_call(args)


def _service_name(config):
    unit = os.environ['JUJU_UNIT_NAME'].split('/')[0]
    for x in (':', '-', '/', '"', "'"):
        unit = unit.replace(x, '_')
    return unit


def _service_dir(config):
    return os.path.join(config['install_root'], _service_name(config))


def pip_install(package):
    cmd_line = ['pip', 'install', '-b', '/tmp/']
    if package.startswith('svn+') or package.startswith('git+') or \
       package.startswith('hg+') or package.startswith('bzr+'):
           cmd_line.append('-e')
    cmd_line.append(package)
    return(subprocess.call(cmd_line))


def _install_from_tarball(config):
    juju_info('grabbing service from tarball...')
    sdir = _service_dir(config)
    if os.path.exists(sdir):
        juju_info('deleting pre-existing service directory: %s' % sdir)
        shutil.rmtree(sdir)
    os.mkdir(sdir)
    cmd = 'curl %s | tar -xzC %s' % (config['tarball'], sdir)
    subprocess.check_call(cmd, shell=True)


def _install_from_bzr(config):
    juju_info('grabbing service from bzr...')

    sdir = _service_dir(config)
    if os.path.exists(sdir):
        juju_info('deleting pre-existing service directory: %s' % sdir)
        shutil.rmtree(sdir)

    args = ['bzr', 'branch']
    rev = config.get('revno', '')
    if rev:
        args.extend(['-r', rev])
    args.append(config['branch'])
    args.append(sdir)
    subprocess.check_call(args)


def install(config):
    charmhelpers.fetch.configure_sources(update=True)

    pkgs = [x for x in config.get('packages', '').split(' ') if x]
    pkgs.append('python-amqplib')
    pkgs.append('python-pip')
    pkgs.append('bzr')

    juju_info('installing apt packages...')
    charmhelpers.fetch.apt_install(pkgs)

    pip_pkgs = [x for x in config.get('pip-packages', '').split(' ') if x]
    if pip_pkgs:
        juju_info('installing pip packages...')
        for package in pip_pkgs:
            pip_install(package.strip())

    if config.get('vcs') == 'tarball':
        _install_from_tarball(config)
    else:
        _install_from_bzr(config)

    unit_config = config.get('unit-config')
    if unit_config:
        with open(os.path.join(_service_dir(config), 'unit_config'), 'w') as f:
            f.write(base64.b64decode(unit_config))


def _create_upstart(config):
    template = textwrap.dedent('''
    #--------------------------------------------------------------
    # This file is managed by the rabbitmq-worker Juju charm
    #--------------------------------------------------------------
    description "starts a rabbitmq worker"
    start on (local-filesystems and net-device-up IFACE=eth0)
    stop on runlevel [!12345]

    # use sigint so python code just needs to catch KeyboardInterrupt
    kill signal SIGINT

    # If the process quits unexpectadly trigger respawn it, unless it
    # fails 15 times within 5 seconds
    respawn
    respawn limit 15 5

    setuid {uid}
    setgid {gid}
    chdir {sdir}

    exec {main}
    ''')
    params = {
        'main': config['main'],
        'sdir': _service_dir(config),
        'uid': config.get('uid', 'nobody'),
        'gid': config.get('gid', 'nogroup'),
    }
    path = '/etc/init/{}.conf'.format(_service_name(config))
    with tempfile.NamedTemporaryFile('w', delete=False) as f:
        f.write(template.format(**params))
    # Rename to final destination
    os.rename(f.name, path)
    os.chmod(path, 0644)


def _create_amqp_config(config):
    path = os.path.join(_service_dir(config), 'amqp_config.py')
    with tempfile.NamedTemporaryFile('w', delete=False) as f:
        f.write('# DO NOT EDIT. Generated by restish charm hook\n')
        f.write('AMQP_USER = "%s"\n' % config['amqp-user'])
        f.write('AMQP_VHOST = "%s"\n' % config['amqp-vhost'])
        f.write('AMQP_HOST = "%s"\n' % _relation_get('private-address'))
        f.write('AMQP_PASSWORD = "%s"\n' % _relation_get('password'))
    # Now that the file is complete (and closed) rename it to its final
    # destination. This avoids races where another process try to import that
    # file as a python module (creating a .pyc) and ends up having close enough
    # time stamps to fool python (which won't reload the completed file).
    os.rename(f.name, path)
    os.chmod(path, 0644)


def amqp_relation_joined(config):
    _relation_set({
        'username': config['amqp-user'],
        'vhost': config['amqp-vhost'],
    })


def amqp_relation_changed(config):
    _create_upstart(config)
    _create_amqp_config(config)
    try:
        subprocess.check_call(['/sbin/restart', _service_name(config)])
    except subprocess.CalledProcessError:
        # if it wasn't running, restart fails, so just try to start
        subprocess.check_call(['/sbin/start', _service_name(config)])


def amqp_relation_broken(config):
    subprocess.call(['/sbin/stop', _service_name(config)])
    os.unlink(os.path.join(_service_dir(config), 'amqp_config.py'))


def main():
    hook = os.path.basename(sys.argv[0])
    juju_info("Running hook: %s" % hook)

    hook_py = hook.replace('-', '_')
    funcs = globals()
    if hook_py not in funcs:
        print("Unknown hook: %s" % hook)
        return 1

    config = _config()
    try:
        return funcs[hook_py](config)
    except subprocess.CalledProcessError as e:
        juju_info('Error running: %s: %s' % (e.cmd, e.output))
        return e.returncode


if __name__ == '__main__':
    exit(main())
