#!/usr/bin/env python

import os
import json
import shutil
import subprocess
import sys
import time

import charmhelpers.fetch


def juju_info(msg):
    subprocess.check_call(['juju-log', '-l', 'INFO', msg])
    pass


def _config():
    output = subprocess.check_output(['config-get', '--format=json'])
    return json.loads(output)


def _relation_ids(relation_types=('db',)):
    # accept strings or iterators
    if isinstance(relation_types, basestring):
        reltypes = [relation_types, ]
    else:
        reltypes = relation_types
    relids = []
    for reltype in reltypes:
        relid_cmd_line = ['relation-ids', '--format=json', reltype]
        json_relids = subprocess.check_output(relid_cmd_line).strip()
        if json_relids:
            relids.extend(json.loads(json_relids))
    return relids


def _relation_get(key):
    return subprocess.check_output(['relation-get', key]).strip()


def _relation_set(keyvalues, relation_id=None):
    args = ['relation-set']
    if relation_id:
        args.extend(['-r', relation_id])
    args.extend(["{}={}".format(k, v or '') for k, v in keyvalues.items()])
    subprocess.check_call(args)


def _service_dir(config):
    unit = os.environ['JUJU_UNIT_NAME'].split('/')[0]
    for x in (':', '-', '/', '"', "'"):
        unit = unit.replace(x, '_')
    return os.path.join(config['install_root'], unit)


def _install_from_tarball(config):
    juju_info('grabbing service from tarball...')
    sdir = _service_dir(config)
    if os.path.exists(sdir):
        juju_info('deleting pre-existing service directory: %s' % sdir)
        shutil.rmtree(sdir)
    os.mkdir(sdir)
    cmd = 'curl %s | tar -xzC %s' % (config['tarball'], sdir)
    subprocess.check_call(cmd, shell=True)


def _install_from_bzr(config):
    juju_info('grabbing service from bzr...')

    sdir = _service_dir(config)
    if os.path.exists(sdir):
        juju_info('deleting pre-existing service directory: %s' % sdir)
        shutil.rmtree(sdir)

    args = ['bzr', 'branch']
    rev = config.get('revno', '')
    if rev:
        args.extend(['-r', rev])
    args.append(config['branch'])
    args.append(sdir)
    subprocess.check_call(args)


def install(config):
    pkgs = [x for x in config['packages'].split(' ') if x]
    pkgs.append('bzr')
    pkgs.append('python-restish')

    juju_info('installing apt packages...')
    charmhelpers.fetch.configure_sources(update=True)
    charmhelpers.fetch.apt_install(pkgs)

    if config.get('vcs') == 'tarball':
        _install_from_tarball(config)
    else:
        _install_from_bzr(config)


def start(config):
    service = 'restish'
    conf = '/etc/init/%s.sh' % service
    if os.path.exists(conf):
        try:
            subprocess.check_call(['service', 'restart', service])
        except subprocess.CalledProcessError:
            # if it wasn't running, restart fails, so just try to start
            subprocess.check_call(['service', 'start', service])


def stop(config):
    service = 'restish'
    conf = '/etc/init/%s.sh' % service
    if os.path.exists(conf):
        subprocess.check_call(['service', 'stop', service])


def config_changed(config):
    # Trigger WSGI reloading
    for relid in _relation_ids('wsgi'):
        _relation_set({'wsgi_timestamp': time.time()}, relation_id=relid)


def website_relation_joined(config):
    host = subprocess.check_output(['unit-get', 'private-address']).strip()
    _relation_set({'port': config["port"], 'hostname': host})


def website_relation_changed(config):
    return website_relation_joined(config)


def website_relation_broken(config):
    pass


def json_status_relation_joined(config):
    path = config.get('json_status_path')
    if path:
        port = config['port']
        host = subprocess.check_output(['unit-get', 'public-address']).strip()
        if not host:
            juju_info('no public address found, using private')
            host = subprocess.check_output(
                ['unit-get', 'private-address']).strip()
        url = 'http://%s:%s/%s' % (host, port, path)
        _relation_set({'status-url': url})


def wsgi_relation_joined(config):
    sdir = _service_dir(config)

    _relation_set({'working_dir': sdir})

    for var in config:
        if var.startswith('wsgi_') or var in ['listen_ip', 'port']:
            _relation_set({var: config[var]})

    path = config.get('python_path')
    if path:
        full_path = []
        for p in path.split(':'):
            if p[0] != '/':
                p = os.path.join(sdir, p)
            full_path.append(p)
        _relation_set({'python_path': ':'.join(full_path)})

    port = config.get('port')
    if port:
        subprocess.call(['open-port', "%s/TCP" % port])


def wsgi_relation_changed(config):
    return wsgi_relation_joined(config)


def wsgi_relation_broken(config):
    port = config.get('port')
    if port:
        subprocess.call(['close-port', "%s/TCP" % port])


def amqp_relation_joined(config):
    _relation_set({
        'username': config['amqp-user'],
        'vhost': config['amqp-vhost'],
    })


def amqp_relation_changed(config):
    with open(os.path.join(_service_dir(config), 'amqp_config.py'), 'w') as f:
        f.write('# DO NOT EDIT. Generated by restish charm hook\n')
        f.write('AMQP_USER = "%s"\n' % config['amqp-user'])
        f.write('AMQP_VHOST = "%s"\n' % config['amqp-vhost'])
        f.write('AMQP_HOST = "%s"\n' % _relation_get('private-address'))
        f.write('AMQP_PASSWORD = "%s"\n' % _relation_get('password'))


def amqp_relation_broken(config):
    os.unlink(os.path.join(_service_dir(config), 'amqp_config.py'))


def lander_jenkins_relation_joined(config):
    host = subprocess.check_output(['unit-get', 'private-address']).strip()
    _relation_set({
        'lander_api_hostname': host,
    })


def lander_jenkins_relation_changed(config):
    data = {
        "lander_username": _relation_get('lander_username'),
        "lander_password": _relation_get('lander_password'),
        "lander_hostname": _relation_get('lander_hostname'),
        "lander_port": _relation_get('lander_port'),
        "lander_master_job": _relation_get('lander_master_job'),
        "lander_job_token": _relation_get('lander_job_token'),
    }
    with open(os.path.join(_service_dir(config), 'lander_config.json'),
              'w') as f:
        f.write(json.dumps(data))
        f.write('\n')


def lander_jenkins_relation_broken(config):
    os.unlink(os.path.join(_service_dir(config), 'lander_config.json'))


def main():
    hook = os.path.basename(sys.argv[0])
    juju_info("Running hook: %s" % hook)

    hook_py = hook.replace('-', '_')
    funcs = globals()
    if hook_py not in funcs:
        print("Unknown hook: %s" % hook)
        return 1

    config = _config()
    try:
        return funcs[hook_py](config)
    except subprocess.CalledProcessError as e:
        juju_info('Error running: %s: %s' % (e.cmd, e.output))
        return e.returncode


if __name__ == '__main__':
    exit(main())
