# Ubuntu CI Engine
# Copyright 2013, 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import logging
import os
import socket
import time

from amqplib import client_0_8 as amqp

log = logging.getLogger(__name__)


def get_config():
    '''Load the rabbit config created by the restish charm'''
    config = None
    try:
        # try and find the config file, should be in the root of our bzr branch
        f = os.path.join(os.path.dirname(__file__), '../../amqp_config.py')
        if os.path.exists(f):
            import imp
            config = imp.load_source('amqp_config', f)
        else:
            log.warn('No amqp_config found at: %s' % os.path.abspath(f))
    except:
        log.exception('ERROR detecting rabbit args')
    return config


def connection(rabbit_config):
    return amqp.Connection(
        userid=rabbit_config.AMQP_USER,
        virtual_host=rabbit_config.AMQP_VHOST,
        host=rabbit_config.AMQP_HOST,
        password=rabbit_config.AMQP_PASSWORD
    )


def send(queue, msg):
    config = get_config()
    if not config:
        return 'rabbitmq settings not available.'
    con = channel = None
    try:
        con = connection(config)
        channel = con.channel()
        channel.queue_declare(queue=queue, durable=True, auto_delete=False)
        body = amqp.Message(msg)
        body.properties['delivery_mode'] = 2  # Persistent
        channel.basic_publish(body, routing_key=queue)
    except Exception as e:
        logging.exception('unable to queue up message: %s', e)
        return str(e)
    finally:
        if channel:
            channel.close()
        if con:
            con.close()
    return None


def _run_forever(channel, queue, callback, retry_period=120):
    tag = channel.basic_consume(callback=callback, queue=queue)
    try:
        timeout = time.time()
        while channel.callbacks and time.time() < timeout + retry_period:
            try:
                channel.wait()
                timeout = time.time()
            except (amqp.AMQPConnectionException, socket.error):
                logging.error('lost connection to Rabbit')
                # TODO metrics.meter('lost_rabbit_connection')
                # Don't probe immediately, give the network/process
                # time to come back.
                time.sleep(0.1)
        if channel.callbacks:
            logging.error('Rabbit did not reappear quickly enough.')
    except KeyboardInterrupt:
        pass
    finally:
        if channel and channel.callbacks and channel.is_open:
            channel.basic_cancel(tag)


def process_queue(config, queue, callback, delete=False):
    conn = channel = None
    try:
        conn = connection(config)
        channel = conn.channel()
        channel.queue_declare(queue=queue, durable=True, auto_delete=False)
        channel.basic_qos(0, 1, False)
        logging.info('Waiting for messages. ^C to exit.')
        _run_forever(channel, queue, callback)
    finally:
        if channel:
            if delete:
                channel.queue_delete(queue)
            channel.close()
        if conn:
            conn.close()


def progress_waiting(progress_trigger, data=None):
    if data is None:
        data = {}
    data['state'] = 'WAITING'
    return send(progress_trigger, json.dumps(data))


def progress_update(progress_trigger, data=None):
    if data is None:
        data = {}
    data['state'] = 'STATUS'
    return send(progress_trigger, json.dumps(data))


def progress_completed(progress_trigger, data=None):
    if data is None:
        data = {}
    data['state'] = 'COMPLETED'
    data['result'] = 'PASSED'
    data['exit'] = True
    return send(progress_trigger, json.dumps(data))


def progress_failed(progress_trigger, data=None):
    if data is None:
        data = {}
    data['state'] = 'COMPLETED'
    data['result'] = 'FAILED'
    data['exit'] = True
    return send(progress_trigger, json.dumps(data))
