# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import logging
import os
import StringIO
import subprocess
import threading
import traceback
import urllib2
import yaml

from ci_utils import amqp_utils, data_store, dump_stack

logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


def _get(url):
    try:
        resp = urllib2.urlopen(url)
        resp = resp.read()
        return json.loads(resp)
    except:
        logging.exception('error checking url')
        return {}


class _timer(threading.Thread):
    '''similar to threading.Timer, but repeats until cancel is called'''
    def __init__(self, interval, cb):
        super(_timer, self).__init__()
        self.interval = interval
        self.cb = cb
        self.finished = threading.Event()
        self.exit = False

    def cancel(self):
        """Stop the timer if it hasn't finished yet"""
        self.finished.set()
        self.exit = True

    def run(self):
        while not self.exit:
            self.finished.wait(self.interval)
            if not self.finished.is_set():
                self.cb()
        self.finished.set()


class AMQPWorker(object):
    '''Base class that handles the more complex issues of a rabbit worker.

    This allows a user to subclass this and provide a custom "handle_request"
    method.
    '''

    def __init__(self, logger_name):
        self.logger_name = logger_name

    # how often to check if the current task should be cancelled
    cancel_interval = 120

    def handle_request(self, logger, params):
        '''To be implemented by the subclass to do their work

        amqp_cb should either be one of:
          * amqp_utils.progress_failed
          * amqp_utils.progress_comple
        :return: tuple(ampq_cb, return_value_dictionary)
        '''
        raise NotImplementedError

    def main(self, queue):
        dump_stack.install_stack_dump_signal()
        config = amqp_utils.get_config()
        if not config:
            exit(1)  # the get_config code prints an error
        amqp_utils.process_queue(config, queue, self._on_message)

    def _create_worker_logger(self):
        '''Create a logger that captures output into StringIO.

        This provides an easy way to let workers get free logging that will
        get added to the ticket.
        '''
        log = logging.getLogger(self.logger_name)
        log.buffer = StringIO.StringIO()
        logstream = logging.StreamHandler(log.buffer)
        formatter = logging.Formatter(
            '[%(asctime)s] %(name)s:%(levelname)s:%(message)s')
        logstream.setFormatter(formatter)
        logstream.setLevel(logging.INFO)
        log.addHandler(logstream)
        return log

    def _create_data_store(self, ticket):
        # Note that this may be called multiple times
        fname = os.path.join(os.path.dirname(__file__), '../../unit_config')
        with open(fname) as f:
            config = yaml.safe_load(f)
            return data_store.create_for_ticket(ticket, config)

    def _store_worker_log(self, store, logger, retval):
        '''An exception safe mechanism to upload log files.

        This needs to catch any exception so that progress complete/fail
        call can be made on the queue.
        '''
        if not store or not logger:
            return
        content = logger.buffer.getvalue()
        if not content:
            log.info('no logging content from action, skipping upload')
        try:
            name = '{}.output.log'.format(logger.name)
            url = store.put_file(name, content, 'text/plain')
            self._create_artifact(name, url, retval)
        except:
            log.exception('unable to store worker log')

    def _create_artifact(self, name, url, retval, type='LOGS'):
        retval.setdefault('artifacts', []).append({
            'name': name,
            'reference': url,
            'type': type,
        })

    def _handle_cancel(self, params):
        '''provide an advisory mechansim to inform the worker to stop

        This checks a URL to determine if the job has been cancelled. If so,
        it updates the "params" object with a params['cancelled'] = True. Thus
        this is an advisory mechansim that the worker can choose how/if to
        deal with.
        '''
        url = params.get('cancel_url')
        if not url:
            return None

        def check_url():
            data = _get(url)
            if not data.get('building', True):
                params['cancelled'] = True
        t = _timer(self.cancel_interval, check_url)
        t.start()
        return t

    def _handle_exception(self, e, store, trigger, logger, ret):
        try:
            err = 'Unexpected exception occurred'
            if isinstance(e, KeyboardInterrupt):
                err = 'Worker terminated'
            elif isinstance(e, subprocess.CalledProcessError) and logger:
                err = 'error running {}: {}'.format(e.cmd, e.output)
            ret['error_message'] = err
            ret['traceback'] = traceback.format_exc()
            if logger:
                logger.error(err)
            self._store_worker_log(store, logger, ret)
        finally:
            if trigger:
                amqp_utils.progress_failed(trigger, ret)

    def _on_message(self, msg):
        log.info('on_message: %s', msg.body)
        ret = {}
        store = logger = trigger = cancel_thread = None
        try:
            params = json.loads(msg.body)
            trigger = params['progress_trigger']
            amqp_utils.progress_update(trigger, params)
            store = self._create_data_store(params['ticket_id'])

            cancel_thread = self._handle_cancel(params)

            logger = self._create_worker_logger()
            amqp_cb, ret = self.handle_request(logger, params)
            self._store_worker_log(store, logger, ret)
            amqp_cb(trigger, ret)
        except (KeyboardInterrupt, Exception) as e:
            self._handle_exception(e, store, trigger, logger, ret)
            if isinstance(e, KeyboardInterrupt):
                raise  # re-raise so amqp_utils.process_queue can exit
        finally:
            msg.channel.basic_ack(msg.delivery_tag)
            if cancel_thread:
                cancel_thread.cancel()
