# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import novaclient.v1_1.client as novaclient
import os
import subprocess
import time
from requests import ConnectionError


logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


class NovaClientException(Exception):
    pass


class NovaClient(object):

    NovaClientException = NovaClientException

    def __init__(self, auth_config, imagename, flavor='m1.small'):
        creds = self._get_nova_creds(auth_config)
        self.nova = novaclient.Client(**creds)
        image = self.nova.images.find(name=imagename)
        inst_flavor = self.nova.flavors.find(name=flavor)
        SSH_KEYNAME = os.path.expanduser('~/.ssh/id_rsa')
        if not os.path.exists(SSH_KEYNAME):
            subprocess.call(
                ['ssh-keygen', '-t', 'rsa', '-f', SSH_KEYNAME, '-N', ''])
        if not self.nova.keypairs.findall(name="sshkey"):
            with open(SSH_KEYNAME + '.pub') as sshkey:
                self.key = self.nova.keypairs.create(
                    name="sshkey", public_key=sshkey.read())
        self.instance = self.nova.servers.create(
            name="test", image=image, flavor=inst_flavor, key_name="sshkey")
        self.ip = self._get_ip()

    def _wait_for_instance(self):
        status = self.instance.status
        #Try for up to 5 minutes to create the instance
        timeout = time.time() + 300
        while status == 'BUILD' and time.time() < timeout:
            log.info("Waiting on instance to become available...")
            time.sleep(5)
            self._refresh_instance()
            status = self.instance.status
        if status == 'BUILD':
            raise NovaClientException("Instance creation timed out")

    def _refresh_instance(self):
        id = self.instance.id
        try:
            #You have to re-get the instance for it to update
            self.instance = self.nova.servers.get(id)
        except ConnectionError:
            log.warn("Received connection error, retrying")

    def _get_ip(self):
        timeout = time.time() + 300
        ip = None
        while not ip and time.time() < timeout:
            try:
                #You have to re-get the instance for it to update
                self._refresh_instance()
                #The value first instance in networks should contain the
                #value we need
                ip = self.instance.networks.values()[0][0]
            except IndexError:
                log.info("Waiting for IP to become available")
                time.sleep(5)
        if not ip:
            raise NovaClientException("IP address never found")
        log.info("IP is %s" % ip)
        return ip

    def runcmd(self, cmd):
        sshopts = ['-o', 'UserKnownHostsFile=/dev/null', '-o ',
                   'StrictHostKeyChecking=no']
        cmd = ['ssh'] + ['ubuntu@' + self.ip] + sshopts + cmd.split(' ')
        return subprocess.check_output(cmd)

    def scp(self, src, dest, recurse=False):
        sshopts = ['-o', 'UserKnownHostsFile=/dev/null', '-o ',
                   'StrictHostKeyChecking=no']
        if recurse:
            sshopts.append['-r']
        ['scp'] + ['ubuntu@' + self.ip] + sshopts + [src, dest]

    def delete(self):
        self.instance.delete()

    def _get_nova_creds(self, config):
        try:
            creds = {'auth_url': config['auth_url'],
                     'username': config['auth_user'],
                     'api_key': config['auth_password'],
                     'project_id': config['auth_tenant_name'],
                     'region_name': config['auth_region']}
        except KeyError:
            raise self.NovaClientException(
                "Missing or invalid authentication info."
            )
        return creds
