# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import mock
import unittest

from ci_utils import jenkins_utils


class TestJenkins(unittest.TestCase):
    @mock.patch('os.path.exists')
    def test_get_config_no_file(self, os_path_exists_mock):
        '''Assert that a missing config file returns None.'''
        os_path_exists_mock.return_value = False
        result = jenkins_utils.get_config()
        self.assertIsNone(result)

    @mock.patch('os.path.exists')
    @mock.patch('__builtin__.open')
    def test_get_config_file_error(self, open_mock, os_path_exists_mock):
        '''Assert that failure to open the config file returns None.'''
        file_mock = mock.Mock()
        file_mock.read.return_value = False
        context_manager_mock = mock.Mock()
        context_manager_mock.__enter__ = mock.Mock()
        context_manager_mock.__enter__.return_value = file_mock
        context_manager_mock.__exit__ = mock.Mock()
        open_mock.return_value = context_manager_mock
        os_path_exists_mock.return_value = True
        result = jenkins_utils.get_config()
        self.assertIsNone(result)

    @mock.patch('jenkins.Jenkins')
    def test_trigger_build(self, jenkins_mock):
        '''Test that a valid jenkins config can trigger a build.'''
        job_name = 'master_job'
        job_params = {'foo': 'bar'}
        job_token = 'token'
        jenkins_config = {'lander_hostname': 'http://localhost/',
                          'lander_port': 8080,
                          'lander_username': 'lander-bot',
                          'lander_password': 'lander-bot',
                          'lander_master_job': job_name,
                          'lander_job_token': job_token}
        build_job_mock = mock.Mock()
        build_job_mock.return_value = None
        jenkins_obj_mock = mock.Mock()
        jenkins_obj_mock.build_job = build_job_mock
        jenkins_mock.return_value = jenkins_obj_mock
        result = jenkins_utils.trigger_build(jenkins_config, job_params)
        build_job_mock.assert_called_once_with(job_name, job_params, job_token)
        self.assertIsNone(result)

    def test_get_api_json_url(self):
        '''Verify correct expansion of the jenkins api json path.'''
        url = 'http://jenkins:8080/job/lander_master'
        self.assertEqual(url + '/api/json',
                         jenkins_utils.get_api_json_url(url))

    def test_get_console_log_url(self):
        '''Verify correct expansion of the jenkins console log path.'''
        url = 'http://jenkins:8080/job/lander_master'
        self.assertEqual(url + '/consoleText',
                         jenkins_utils.get_console_log_url(url))

    def test_get_artifact_url(self):
        '''Verify correct expansion of the jenkins artifact path.'''
        url = 'http://jenkins:8080/job/lander_master'
        name = 'artifact.1'
        self.assertEqual('{}/{}/{}'.format(url, 'artifact', name),
                         jenkins_utils.get_artifact_url(url, name))

    @mock.patch('urllib2.Request')
    def test_get_url_contents(self, request_mock):
        '''Verify generation of EnvironmentError on a failed url load.'''
        request_mock.side_effect = IOError
        self.assertRaises(
            EnvironmentError,
            jenkins_utils.get_url_contents,
            'http://jenkins:8080/bad-url/')

    @mock.patch('ci_utils.jenkins_utils.get_url_contents')
    def test_get_json(self, get_url_mock):
        url = 'http://jenkins:8080/job/lander_master'
        json_data = {'name': 'lander_master'}
        get_url_mock.return_value = json.dumps(json_data)
        self.assertEqual(json_data, jenkins_utils.get_json(url))

    @mock.patch('ci_utils.jenkins_utils.get_url_contents')
    def test_is_master_job_running_no_prior_build(self, get_url_mock):
        jenkins_config = {'lander_hostname': 'localhost',
                          'lander_port': 8080,
                          'lander_master_job': 'lander_master'}
        job_data = {'name': 'lander_master', 'lastBuild': False}
        last_build_data = {'name': 'lander_master', 'building': True}
        get_url_mock.side_effect = [json.dumps(job_data),
                                    json.dumps(last_build_data)]
        self.assertFalse(jenkins_utils.is_master_job_running(jenkins_config))

    @mock.patch('ci_utils.jenkins_utils.get_url_contents')
    def test_is_master_job_running(self, get_url_mock):
        jenkins_config = {'lander_hostname': 'localhost',
                          'lander_port': 8080,
                          'lander_master_job': 'lander_master'}
        job_data = {'name': 'lander_master', 'lastBuild': True}
        last_build_data = {'name': 'lander_master', 'building': True}
        get_url_mock.side_effect = [json.dumps(job_data),
                                    json.dumps(last_build_data)]
        self.assertTrue(jenkins_utils.is_master_job_running(jenkins_config))

    @mock.patch('ci_utils.jenkins_utils.get_url_contents')
    def test_is_master_job_running_False(self, get_url_mock):
        jenkins_config = {'lander_hostname': 'localhost',
                          'lander_port': 8080,
                          'lander_master_job': 'lander_master'}
        job_data = {'name': 'lander_master', 'lastBuild': True}
        last_build_data = {'name': 'lander_master', 'building': False}
        get_url_mock.side_effect = [json.dumps(job_data),
                                    json.dumps(last_build_data)]
        self.assertFalse(jenkins_utils.is_master_job_running(jenkins_config))
