# Ubuntu CI Engine
# Copyright 2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import unittest

import webtest

from restish import resource
from restish.app import RestishApp

from ci_utils import restish_utils


def _no_param_action():
    return restish_utils.json_ok({'func': _no_param_action.__name__})


def _two_param_action(p1, p2):
    return restish_utils.json_ok({'func': _two_param_action.__name__})


class TestGetResource(unittest.TestCase):
    '''Test to ensure the http_get_resource works properly'''

    def setUp(self):
        super(TestGetResource, self).setUp()

        class TR(resource.Resource):
            @resource.child()
            def empty(self, request, segments):
                return restish_utils.http_get_resource(_no_param_action)

            @resource.child()
            def two_param_get(self, request, segments):
                return restish_utils.http_get_resource(_two_param_action)

        app = RestishApp(TR())
        self.app = webtest.TestApp(app, relative_to='.')

    def testGetNoParam(self):
        resp = self.app.get('/empty', status=200)
        data = json.loads(resp.body)
        self.assertEqual(data['func'], _no_param_action.__name__)

    def testGetTwoParam(self):
        resp = self.app.get('/two_param_get', {'p1': 1, 'p2': 2}, status=200)
        data = json.loads(resp.body)
        self.assertEqual(data['func'], _two_param_action.__name__)

        # ensure a missing parameter returns a 400 error
        resp = self.app.get('/two_param_get', {'p1': 1}, status=400)
        self.assertTrue('p2' in resp.body)

    def testLeafEnforced(self):
        '''ensure we only respond to the leaf and not additional segments'''
        self.app.get('/empty/', status=404)

    def testAllowedMethods(self):
        self.app.post_json('/empty', {'key': 'val'}, status=405)

    def testJSONP(self):
        resp = self.app.get('/empty?format=jsonp', status=200)
        body = resp.body.strip()
        self.assertTrue(body.startswith('callback({'))
        self.assertTrue(body.endswith(');'))

        resp = self.app.get('/empty?format=jsonp&callback=foo', status=200)
        body = resp.body.strip()
        self.assertTrue(body.startswith('foo({'))
        self.assertTrue(body.endswith(');'))


class TestPostResource(unittest.TestCase):
    '''Test to ensure the http_post_resource works properly'''

    def setUp(self):
        super(TestPostResource, self).setUp()

        class TR(resource.Resource):
            @resource.child()
            def empty(self, request, segments):
                return restish_utils.http_post_resource(_no_param_action)

            @resource.child()
            def two_param_get(self, request, segments):
                return restish_utils.http_post_resource(_two_param_action)

        app = RestishApp(TR())
        self.app = webtest.TestApp(app, relative_to='.')

    def testPostNoParam(self):
        resp = self.app.post_json('/empty', {}, status=200)
        data = json.loads(resp.body)
        self.assertEqual(data['func'], _no_param_action.__name__)

    def testPostTwoParam(self):
        resp = self.app.post_json('/two_param_get', {'p1': 1, 'p2': 2})
        data = json.loads(resp.body)
        self.assertEqual(data['func'], _two_param_action.__name__)

        # ensure a missing parameter returns a 400 error
        resp = self.app.post_json('/two_param_get', {'p1': 1}, status=400)
        self.assertTrue('p2' in resp.body)

    def testLeafEnforced(self):
        '''ensure we only respond to the leaf and not additional segments'''
        self.app.post_json('/empty/', {}, status=404)

    def testAllowedMethods(self):
        self.app.get('/empty', {'p1': 1, 'p2': 2}, status=405)
