# Ubuntu Continuous Integration Engine
# Copyright 2013, 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.

# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import logging

from dput.changes import logger, parse_changes_file, ChangesFileException
log = logging.getLogger()

from ci_libs import utils
from ci_utils import data_store


def set_log_level(log_ref):
    """Set log level from caller's logger so reporting is consitent."""
    log.setLevel(log_ref.level)
    logger.setLevel(log_ref.level)


class FileToUploadNotFound(IOError):
    """Raised if a file mentioned in .changes is not found."""


class UploadDirNotFound(IOError):
    """Raised if the provided upload directory is not found."""


class ChangesFileNotFound(IOError):
    """Raised if .changes file is not found."""


class NotAChangesFileError(Exception):
    """Raised if provided file isn't a .changes."""


class ChangesParseError(ChangesFileException):
    """Raised when .parse() fails."""


class ChangesValidationError(Exception):
    """Raised when .changes file fails validation."""


class ChangesProcessor:
    """Class to parse, validate and upload source files to the datastore."""

    def __init__(self, changes_filepath, files_to_upload_dir=None):
        # Needs to be provided with the full path to the changes file.
        if not os.path.exists(changes_filepath):
            raise ChangesFileNotFound(changes_filepath)
        self.changes_filepath = changes_filepath

        if files_to_upload_dir is not None:
            if not os.path.exists(files_to_upload_dir):
                raise UploadDirNotFound(files_to_upload_dir)
            self.files_to_upload_dir = files_to_upload_dir
        else:
            self.files_to_upload_dir = os.path.dirname(changes_filepath)
        self.files_to_upload = []
        self.changes = None

    @property
    def source_package_name(self):
        return self.changes._data["Source"]

    @property
    def source_version(self):
        return self.changes._data["Version"]

    def validate(self, check_signature=True):
        changes = self.changes
        try:
            if changes.get('Distribution') == "UNRELEASED":
                raise ChangesValidationError(
                    "Invalid 'Distribution' in .changes: UNRELEASED")
            # Enabling signature check to avoid bug #1283186
            if check_signature:
                try:
                    changes.validate_signature()
                except ChangesFileException as exc:
                    raise ChangesValidationError(
                        "failed to verify signature. Is it signed? "
                        "({})".format(exc))
            changes.validate_checksums()
        except Exception, exc:
            raise exc

    def parse(self):
        try:
            self.changes = parse_changes_file(
                filename=self.changes_filepath,
                directory=self.files_to_upload_dir)
            # Check if all files to upload mentioned by .changes can be found
            # at the provided directory. If not, abort the upload.
            for file_ in self.changes.get_files():
                if os.path.exists(file_):
                    self.files_to_upload.append(file_)
                else:
                    raise FileToUploadNotFound(file_)
            self.files_to_upload.append(self.changes.get_changes_file())
        except (FileToUploadNotFound, ChangesFileException) as exc:
            raise exc
        except KeyError as exc:
            # This happens when a dsc file is provided accidentally.
            raise ChangesParseError("Maybe this is a .dsc file?")

    def process(self, check_signature=True):
        self.parse()
        self.validate(check_signature=check_signature)
        return self.files_to_upload


def upload_files(file_path, ticket_id):
    ds = data_store.create_for_ticket(ticket_id, utils.AUTH_CONFIG)
    with open(file_path, 'r') as fp:
        location = ds.put_file(file_path, fp)
    return location
