# Ubuntu Continuous Integration Engine
# Copyright 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import logging

from ci_libs.file_handler import ChangesProcessor, upload_files
from ci_libs import utils
from ci_utils.ticket_states import TicketWorkflowStep, TicketWorkflowStepStatus

log = logging.getLogger()


class SubTicket():

    def __init__(self, owner, ticket_id):
        self.owner = owner
        self.ticket_id = str(ticket_id)
        self.sourcepackage = ''
        self.files = ''
        self.version = ''

        self.spu_uri = ''
        self.subticket_uri = ''
        self.ticket_uri = utils.TICKET_BASE + self.ticket_id + '/'

    def _parse_changes(self, changes_filepath):
        log.info("Parsing %s..." % changes_filepath)
        changes = ChangesProcessor(
            changes_filepath=changes_filepath)
        log.info("Validating .changes file...")
        self.files = changes.process()
        self.version = changes.source_version
        self.sourcepackage = changes.source_package_name
        self.version = changes.source_version
        log.info("Done. %s_%s will be uploaded." % (self.sourcepackage,
                                                    self.version))

    def _process(self):
        self._create_spu()
        self._create_subticket()
        for file_path in self.files:
            location = upload_files(file_path, self.ticket_id)
            log.info("File uploaded to %s" % location)
            self._create_artifact(os.path.basename(file_path), location)

    def _create_sourcepackage(self):
        data = {
            "name": self.sourcepackage,
        }
        location = utils.post(
            utils.CI_URL + utils.SOURCEPACKAGE_BASE, data=data)
        sourcepackage_id = utils.parse_id(location)
        return utils.SOURCEPACKAGE_BASE + sourcepackage_id + '/'

    def _create_spu(self):
        sourcepackage_uri = utils.get_sourcepackage_uri(self.sourcepackage)
        if sourcepackage_uri == '':
            sourcepackage_uri = self._create_sourcepackage()
        data = {
            "sourcepackage": sourcepackage_uri,
            "version": self.version,
        }
        location = utils.post(utils.CI_URL + utils.SPU_BASE, data=data)
        self.spu_id = utils.parse_id(location)
        self.spu_uri = utils.SPU_BASE + self.spu_id + '/'
        log.info("Created source package upload: %s" % location)
        return location

    def _create_subticket(self):
        data = {
            "ticket": self.ticket_uri,
            "assignee": self.owner,
            "source_package_upload": self.spu_uri,
        }
        location = utils.post(utils.CI_URL + utils.SUBTICKET_BASE, data=data)
        self.subticket_id = utils.parse_id(location)
        self.subticket_uri = utils.SUBTICKET_BASE + self.subticket_id + '/'
        log.info("Created subticket: %s" % location)
        return location

    def _create_artifact(self, file, location):
        data = {
            "type": "SPU",
            "reference": location,
            "name": file,
            "subticket": self.subticket_uri,
        }
        location_ = utils.post(utils.CI_URL + utils.ARTIFACT_BASE, data=data)
        log.info("Created artifact: %s" % location_)


class Ticket():

    def __init__(self):
        self.ticket_id = ''
        self.ticket_uri = ''
        self.uploads = []
        self.subtickets = {}

    def _create_ticket(self, args):
        data = {
            "title": args.title,
            "owner": args.owner,
            "description": args.description,
            "bug_id": args.bug,
            "added_binaries": args.add,
            "removed_binaries": args.remove,
        }
        location = utils.post(utils.CI_URL + utils.TICKET_BASE, data=data)
        self.ticket_id = utils.parse_id(location)
        self.ticket_uri = utils.TICKET_BASE + self.ticket_id + '/'
        return self.ticket_id

    def _update_ticket(self):
        """
        After a ticket has been created, it needs to be updated to have a
        status of QUEUED in order to be processed.
        """
        data = {
            "current_workflow_step": TicketWorkflowStep.QUEUED.value,
            "status": TicketWorkflowStepStatus.PENDING.value,
        }
        urlbase = utils.CI_URL + utils.API_URL + 'updateticketstatus/'
        utils.patch(urlbase + self.ticket_id + '/', data=data)
        print("You have successfully submitted a ticket to the Ubuntu CI"
              " Engine. Your ticket number is %s." % self.ticket_id)

    def add_new_ticket(self, args):
        ticket_id = self._create_ticket(args)
        for source in args.sources:
            subticket = SubTicket(owner=args.owner, ticket_id=ticket_id)
            subticket._parse_changes(source)
            subticket._process()
            self.subtickets[subticket.sourcepackage] = subticket
        self._update_ticket()


def new_ticket(args):
    new_ticket = Ticket()
    new_ticket.add_new_ticket(args)
    return new_ticket
