# Ubuntu Continuous Integration Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.

# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import gnupg
import unittest
import sys
import tempfile
import shutil

from cStringIO import StringIO
from contextlib import contextmanager
from ucitests import fixtures

from ci_libs import utils


KEY_DATA = """
-----BEGIN PGP PUBLIC KEY BLOCK-----
Version: GnuPG v1

mI0EUx0vDgEEAMFUFhO0YmshgBTeOPrkwPFQvZhX0G6JG4BlLsb7W+LLrF24oLwn
8TCgs/wzyO7HnRum6bP+38f76Bm/i5y2po4pg9upWEBVWhi2iMMUyFvBrFtWSHD3
eD4c7S265eELoxTqE9aU6vXUbLdRvSHSLLID/ERxUSY7FiX7aopeOANJABEBAAG0
IUZvbyBCYXIgVGVzdCA8Zm9vYmFyQGV4YW1wbGUuY29tPoi3BBMBCAAhBQJTHS8O
AhsDBQsJCAcDBRUKCQgLBRYCAwEAAh4BAheAAAoJENmiqDEqvvzb0IcEAIDan3yE
p9LofKHiVwgebQf/0GO6ftIZ+0p17EMfX8m/Hfj9XEPu4cwRJB/XhdnJP0kLMcES
/Pa+03CdwdHTZpUyR8HjmoYrUR09euUQYuzXttmT9MwGU30caRvpXlPV58LcK9ff
aNcuLEe/RGHoQsDyr8EhpCTEyGkkB/DqjOb8uI0EUx0vDgEEALKNaE0J72U9UH9/
yg0sJYVmKG5obR8PqxxjKF4L8oBI3eei3ijt+YJXm2uW66utH/KwqQmp7hDdJQD4
QYZx8Z2nG+OYDwfFRJhdQ6FZZ1G2k8M2KxmqYLjd8rh1olZragXMJowQAGFC0zVb
spatmahbZPvZE8oHW84+5ug3cCXPABEBAAGInwQYAQgACQUCUx0vDgIbDAAKCRDZ
oqgxKr782wLmA/9wfkMBzzdrYd6wquJnbyPezbvGu7zOsgalb5UJyzXeMjynkBzU
siiyVtPzSZN9xC86zCced73e4H5yRcmMrje3QLSvUoicVUyRreftdLPBsAHwTbTb
1O4wls94mwLDRpH21HY8eizc3hpuIx09uXLdWwZUH8AwnOYGL7I1xwPw+w==
=0EYm
-----END PGP PUBLIC KEY BLOCK-----

"""


@contextmanager
def capture_stderr(command, *args, **kwargs):
    out, sys.stderr = sys.stderr, StringIO()
    command(*args, **kwargs)
    sys.stderr.seek(0)
    yield sys.stderr.read()
    sys.stderr = out


@contextmanager
def capture_stdout(command, *args, **kwargs):
    out, sys.stdout = sys.stdout, StringIO()
    command(*args, **kwargs)
    sys.stdout.seek(0)
    yield sys.stdout.read()
    sys.stdout = out


def mock_load_config():
    # mock up settings needed by tests that normally come from config
    utils.AUTH_CONFIG = {"ci_url": "foo",
                         "auth_url": "test",
                         "auth_user": "test",
                         "auth_password": "test",
                         "auth_tenant_name": "test",
                         }
    utils.CI_URL = utils.AUTH_CONFIG['ci_url']


class TestCaseWithGnupg(unittest.TestCase):
    '''Provides the necessary setup for testing with custom keyring.'''
    def setUp(self):
        super(TestCaseWithGnupg, self).setUp()
        self.gnupg_home = tempfile.mkdtemp()
        self._set_gnupg_home()
        self.gnupg = gnupg.GPG(gnupghome=self.gnupg_home)
        self.assertEquals(len(self.gnupg.list_keys()), 0)
        self.gnupg.import_keys(KEY_DATA)
        self.assertEquals(len(self.gnupg.list_keys()), 1)
        self.addCleanup(self._destroy_gnupg_env)

    def _set_gnupg_home(self):
        fixtures.override_env(self, 'GNUPGHOME', self.gnupg_home)

    def _destroy_gnupg_env(self):
        shutil.rmtree(self.gnupg_home)
