#!/usr/bin/env python
# Ubuntu Continuous Integration Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.

# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import httplib
import imp
import json
import logging
import mock
import os
import socket

from testfixtures import LogCapture
import urllib2

from ci_utils.ticket_states import TicketWorkflowStep, TicketWorkflowStepStatus
from ci_libs import utils
from ci_libs.ticket import new_ticket
from tests import (TestCaseWithGnupg, capture_stderr, capture_stdout,
                   mock_load_config)


class MainScriptTestCase(TestCaseWithGnupg):

    def setUp(self):
        super(MainScriptTestCase, self).setUp()
        curdir = os.path.dirname(__file__)
        self.main_script = os.path.join(curdir, "../ubuntu-ci")
        self.cli = imp.load_source(
            "ubuntu-ci",
            self.main_script,
        )


class CLICommandsTestCase(MainScriptTestCase):
    """Tests cli basic commands."""

    @mock.patch('urllib2.urlopen')
    @mock.patch('ci_libs.utils.load_config', side_effect=mock_load_config())
    def test_ticket_status_one_ticket(self, mock_load_config, mock_urlopen):
        """Verify cli status command when providing ticket number."""
        data = {
            "current_workflow_step": str(TicketWorkflowStep.IMAGE_BUILDING),
            "status": str(TicketWorkflowStepStatus.IMAGE_BUILDING_INPROGRESS),
            "title": "My cool feature",
            "id": '5',
            "description": "my new feature",
            "owner": "user@example.com",
            "created": "Wed, 12 Mar 2014 02:18:28 +0000",
            "updated": "Wed, 12 Mar 2014 02:30:22 +0000",
            "bug_id": "12345",
            "added_binaries": "mynewpackage",
            "removed_binaries": "myoldpackage",
        }
        resp = mock.Mock()
        resp.read.return_value = json.dumps(data)
        mock_urlopen.return_value = resp
        args = ['status', '-t', '5']
        with capture_stdout(self.cli.main, args) as cm:
            self.assertEquals(
                cm,
                'Ticket #5 - My cool feature\n'
                'Curent landing step: {} - {}\n'
                'Description: my new feature\n'
                'Owner: user@example.com\n'
                'Created: Wed, 12 Mar 2014 02:18:28 +0000\n'
                'Updated: Wed, 12 Mar 2014 02:30:22 +0000\n'
                'Bug #: 12345\n'
                'Added binaries: mynewpackage\n'
                'Removed binaries: myoldpackage\n'.format(
                    str(TicketWorkflowStep.IMAGE_BUILDING),
                    str(TicketWorkflowStepStatus.IMAGE_BUILDING_INPROGRESS)))
        mock_urlopen.assert_called_once()

    @mock.patch('urllib2.urlopen')
    @mock.patch('ci_libs.utils.load_config', side_effect=mock_load_config())
    def test_ticket_status_all_tickets(self, mock_load_config, mock_urlopen):
        """Verify cli status command for multiple tickets."""
        data = {
            "objects": [
                {
                    "id": 5,
                    "current_workflow_step": 100,
                    "status": 110,
                    "title": "My cool feature",
                },
                {
                    "id": 6,
                    "current_workflow_step": 100,
                    "status": 110,
                    "title": "Ubuntu bug fix",
                },
            ]
        }
        resp = mock.Mock()
        resp.read.return_value = json.dumps(data)
        mock_urlopen.return_value = resp
        args = ['status']
        with capture_stdout(self.cli.main, args) as cm:
            self.assertEquals(
                cm,
                "Ticket #5, My cool feature is 100 110\n" +
                "Ticket #6, Ubuntu bug fix is 100 110\n")
        mock_urlopen.assert_called_once()

    @mock.patch('ci_libs.ticket.SubTicket._process')
    @mock.patch('ci_libs.utils.post',
                return_value='http://www.example.com/api/v1/ticket/38/')
    @mock.patch('ci_libs.utils.patch')
    def test_add_new_ticket(self, mock_patch, mock_post, mock_process):
        """Verify cli create_ticket command for one .changes."""
        changes = os.path.join(os.path.dirname(__file__),
                               'data/foobar_0.1-1_source.changes')
        args = ['create_ticket', '-t', 'New feature', '-b', '1234', '-o',
                'someone@example.com', '-d', 'This is a cool new feature',
                '-a', 'foo', '-r', 'bar,baz', '-s', changes]

        args = self.cli.parse_arguments(args)
        new_ticket_ = new_ticket(args)
        self.assertEquals(mock_process.call_count, 1)
        self.assertEquals(mock_post.call_count, 1)
        self.assertEquals(mock_patch.call_count, 1)
        self.assertEquals(len(new_ticket_.subtickets), 1)
        self.assertTrue("foobar" in new_ticket_.subtickets)

    @mock.patch('ci_libs.ticket.SubTicket._process')
    @mock.patch('ci_libs.utils.post',
                return_value='http://www.example.com/api/v1/ticket/39/')
    @mock.patch('ci_libs.utils.patch')
    def test_add_new_ticket_multiple_subtickets(self, mock_patch, mock_post,
                                                mock_process):
        """Verify cli create_ticket command for multiple .changes."""
        changes_1 = os.path.join(os.path.dirname(__file__),
                                 'data/foobar_0.1-1_source.changes')
        changes_2 = os.path.join(os.path.dirname(__file__),
                                 'data/barfoo_0.1-1_source.changes')

        args = ['create_ticket', '-t', '"New feature"', '-b', '4321', '-o',
                'someone@example.com', '-d', '"This is a cool new feature"',
                '-a', 'foo,bar', '-r', 'baz', '-s', changes_1, '-s', changes_2]

        args = self.cli.parse_arguments(args)
        new_ticket_ = new_ticket(args)
        self.assertEquals(mock_process.call_count, 2)
        self.assertEquals(mock_post.call_count, 1)
        self.assertEquals(mock_patch.call_count, 1)

        self.assertEquals(len(new_ticket_.subtickets), 2)
        self.assertTrue("foobar" in new_ticket_.subtickets)
        self.assertTrue("barfoo" in new_ticket_.subtickets)

    @mock.patch('urllib2.urlopen')
    @mock.patch('ci_libs.utils.load_config', side_effect=mock_load_config())
    def test_ts_server_returns_urlerror(self, mock_load_config, mock_urlopen):
        """Test cli response to ts server refusing connection (111)."""
        mock_urlopen.side_effect = urllib2.URLError(reason=socket.error(111,
                                                    "Connection refused"))
        args = ['status']
        with LogCapture() as lc:
            logger = logging.getLogger()
            self.cli.main(args, log=logger)
        mock_urlopen.assert_called_once()
        self.assertTrue("Cannot reach the server at %s." % utils.CI_URL in
                        lc.records[0].msg)
        self.assertTrue("Connection refused" in lc.records[0].msg)

    @mock.patch('urllib2.urlopen')
    @mock.patch('ci_libs.utils.load_config', side_effect=mock_load_config())
    def test_ts_server_returns_internal_server_error(self, mock_load_config,
                                                     mock_urlopen):
        """Test cli response to ts server returning error 500."""
        mock_urlopen.side_effect = urllib2.HTTPError(None, 500,
                                                     "INTERNAL SERVER ERROR",
                                                     None, None)
        args = ['status']
        with LogCapture() as lc:
            logger = logging.getLogger()
            self.cli.main(args, log=logger)
        mock_urlopen.assert_called_once()
        self.assertEquals("Server at {} reported internal error. Maybe a "
                          "database hiccup?".format(utils.CI_URL),
                          lc.records[0].msg)

    @mock.patch('urllib2.urlopen')
    @mock.patch('ci_libs.utils.load_config', side_effect=mock_load_config())
    def test_ts_server_returns_url_not_found(self, mock_load_config,
                                             mock_urlopen):
        """Test cli response to ts server returning 404 error."""
        mock_urlopen.side_effect = urllib2.HTTPError(None, 404,
                                                     "Server Not Found",
                                                     None, None)
        args = ['status']
        with LogCapture() as lc:
            logger = logging.getLogger()
            self.cli.main(args, log=logger)
        mock_urlopen.assert_called_once()
        self.assertEquals("Server at {} not found. Is 'ci_url' correctly "
                          "set? (HTTP Error 404: Server Not Found)".format(
                              utils.CI_URL), lc.records[0].msg)

    @mock.patch('urllib2.urlopen')
    @mock.patch('ci_libs.utils.load_config', side_effect=mock_load_config())
    def test_ts_server_returns_bad_status_line(self, mock_load_config,
                                               mock_urlopen):
        """Test cli response to ts server returning BadStatusLine."""
        args = ['status']
        mock_urlopen.side_effect = httplib.BadStatusLine('')
        with LogCapture() as lc:
            logger = logging.getLogger()
            self.cli.main(args, log=logger)
        mock_urlopen.assert_called_once()
        self.assertEquals("Server at {} replied with an empty response. Is "
                          "'ci_url' pointing to the correct service?".format(
                              utils.CI_URL), lc.records[0].msg)

    @mock.patch('urllib2.urlopen')
    @mock.patch('ci_libs.utils.load_config', side_effect=mock_load_config())
    def test_ts_server_returns_other_httperror(self, mock_load_config,
                                               mock_urlopen):
        """Test cli response to ts server returning HTTPError."""
        mock_urlopen.side_effect = urllib2.HTTPError(None, 403, "Forbidden",
                                                     None, None)
        args = ['status']
        with LogCapture() as lc:
            logger = logging.getLogger()
            self.cli.main(args, log=logger)
        mock_urlopen.assert_called_once()
        self.assertEquals("HTTP Error 403: Forbidden", lc.records[0].msg)


class CliArgumentsTestCase(MainScriptTestCase):
    """Test cli arguments validation."""

    def _test_parser_missing_arguments(self, args, expected_msg):
        """Helper to capture error output when cli fails."""
        with capture_stderr(self.assertRaises, SystemExit, self.cli.main,
                            args) as cm:
            self.assertNotEqual(cm, None)
        self.assertTrue(expected_msg in cm)

    @mock.patch('ci_libs.utils.load_config', side_effect=mock_load_config())
    def test_cli_arguments_are_required(self, mock_load_config):
        """Verify cli arguments are required."""
        args = ['create_ticket']
        self._test_parser_missing_arguments(
            args, "create_ticket: error: argument -t/--title is required")

        args.extend(['-t', '"New feature"'])
        self._test_parser_missing_arguments(
            args,
            "create_ticket: error: argument -d/--description is required")

        args.extend(['-d', '"This is a cool new feature"'])
        self._test_parser_missing_arguments(
            args, "create_ticket: error: argument -o/--owner is required")

        args.extend(['-o', 'someone@example.com'])
        self._test_parser_missing_arguments(
            args, "create_ticket: error: argument -a/--add is required")

        args.extend(['-a', 'foo'])
        self._test_parser_missing_arguments(
            args, "create_ticket: error: argument -s/--sources is required")

        # All arguments were provided. It should fail because file is not
        # found.
        args.extend(['-s', 'foobar.changesss'])
        with LogCapture() as lc:
            logger = logging.getLogger()
            self.cli.main(args, log=logger)
        self.assertIsNotNone(lc)
        self.assertTrue("Changes file not found" in lc.records[0].msg)

    def test_cli_only_source_files_pass(self):
        """Verify cli fails in case file isn't prefixed .changes."""
        wrong_changes = os.path.join(os.path.dirname(__file__),
                                     'data/foobar_0.1-1.dsc')
        args = ['create_ticket', '-t', '"New feature"', '-d',
                '"This is a cool new feature"', '-o', 'someone@example.com',
                '-a', 'foo', '-s', wrong_changes]
        with LogCapture() as lc:
            logger = logging.getLogger()
            self.cli.main(args, log=logger)
        self.assertIsNotNone(lc)
        self.assertEquals(
            lc.records[0].msg,
            "Upload file must be a .changes; you provided '{}'".format(
                wrong_changes))

    @mock.patch('ci_libs.utils.post',
                return_value='http://www.example.com/api/v1/ticket/39/')
    def test_is_valid_package_list_add_space(self, mock_patch):
        """Test a ticket adding test binaries and not removing any fails."""
        changes = os.path.join(os.path.dirname(__file__),
                               'data/foobar_0.1-1_source.changes')
        args = ['create_ticket', '-t', '"New feature"', '-d',
                '"This is a cool new feature"', '-o', 'someone@example.com',
                '-a', 'foo, bar', '-s', changes]
        with LogCapture() as lc:
            logger = logging.getLogger()
            self.cli.main(args, log=logger)
        self.assertIsNotNone(lc)
        self.assertEquals(
            lc.records[0].msg[0],
            "Added and removed packages have to be comma-separated lists "
            "with no spaces.")

    @mock.patch('ci_libs.utils.post',
                return_value='http://www.example.com/api/v1/ticket/39/')
    def test_is_valid_package_list_remove_space(self, mock_patch):
        """Test a ticket removing two test binaries fails."""
        changes = os.path.join(os.path.dirname(__file__),
                               'data/foobar_0.1-1_source.changes')
        args = ['create_ticket', '-t', '"New feature"', '-d',
                '"This is a cool new feature"', '-o', 'someone@example.com',
                '-a', 'foo', '-r', 'bar, baz', '-s', changes]
        with LogCapture() as lc:
            logger = logging.getLogger()
            self.cli.main(args, log=logger)
        self.assertIsNotNone(lc)
        self.assertEquals(
            lc.records[0].msg[0],
            "Added and removed packages have to be comma-separated lists "
            "with no spaces.")


class ChangesValidationTestCase(MainScriptTestCase):
    """Test how cli handles malformed, unreleased or unsigned .changes."""

    @mock.patch('ci_libs.utils.load_config', side_effect=mock_load_config())
    @mock.patch('ci_libs.utils.post',
                return_value='http://www.example.com/api/v1/ticket/39/')
    def test_cli_dsc_file_in_disguise(self, mock_patch, mock_load_config):
        """Verify cli behavior when .changes file is a .dsc."""
        wrong_changes = os.path.join(os.path.dirname(__file__),
                                     'data/foobar_0.1-1_dsc.changes')
        args = ['create_ticket', '-t', '"New feature"', '-d',
                '"This is a cool new feature"', '-o', 'someone@example.com',
                '-a', 'foo', '-s', wrong_changes]
        with LogCapture() as lc:
            logger = logging.getLogger()
            self.cli.main(args, log=logger)
        self.assertIsNotNone(lc)
        self.assertEquals(
            lc.records[0].msg,
            "Problem parsing .changes file, are you sure it's a valid "
            ".changes? (Maybe this is a .dsc file?)")

    @mock.patch('ci_libs.utils.load_config', side_effect=mock_load_config())
    @mock.patch('ci_libs.utils.post',
                return_value='http://www.example.com/api/v1/ticket/39/')
    def test_cli_changes_file_is_something_else(self, mock_patch,
                                                mock_load_config):
        """Verify cli behavior when provided .changes file isn't a .changes."""
        wrong_changes = os.path.join(os.path.dirname(__file__),
                                     'data/foobar_0.1-1_random.changes')
        args = ['create_ticket', '-t', '"New feature"', '-d',
                '"This is a cool new feature"', '-o', 'someone@example.com',
                '-a', 'foo', '-s', wrong_changes]
        with LogCapture() as lc:
            logger = logging.getLogger()
            self.cli.main(args, log=logger)
        self.assertIsNotNone(lc)
        self.assertEquals(
            lc.records[0].msg,
            "Problem parsing .changes file, are you sure it's a valid "
            ".changes? (Changes file could not be parsed.)")

    @mock.patch('ci_libs.utils.load_config', side_effect=mock_load_config())
    @mock.patch('ci_libs.utils.post',
                return_value='http://www.example.com/api/v1/ticket/40/')
    def test_cli_changes_file_unsigned_fails(self, mock_patch,
                                             mock_load_config):
        """Verify cli behavior when provided .changes isn't signed."""
        unsigned_changes = os.path.join(os.path.dirname(__file__),
                                        'data/foobar_0.1-1_unsigned.changes')
        args = ['create_ticket', '-t', '"New feature"', '-d',
                '"This is a cool new feature"', '-o', 'someone@example.com',
                '-a', 'foo', '-s', unsigned_changes]
        with LogCapture() as lc:
            logger = logging.getLogger()
            self.cli.main(args, log=logger)
        self.assertIsNotNone(lc)
        self.assertEquals(
            lc.records[0].msg,
            "Problem validating .changes file: failed to verify signature. Is "
            "it signed? (No signature on)")

    @mock.patch('ci_libs.utils.load_config', side_effect=mock_load_config())
    @mock.patch('ci_libs.utils.post',
                return_value='http://www.example.com/api/v1/ticket/39/')
    def test_cli_changes_file_is_unreleased(self, mock_patch,
                                            mock_load_config):
        """Verify cli behavior when provided .changes file is UNRELEASED."""
        wrong_changes = os.path.join(os.path.dirname(__file__),
                                     'data/foobar_0.1-1_unreleased.changes')
        args = ['create_ticket', '-t', '"New feature"', '-d',
                '"This is a cool new feature"', '-o', 'someone@example.com',
                '-a', 'foo', '-s', wrong_changes]
        with LogCapture() as lc:
            logger = logging.getLogger()
            self.cli.main(args, log=logger)
        self.assertIsNotNone(lc)
        self.assertEquals(
            lc.records[0].msg,
            "Problem validating .changes file: Invalid 'Distribution' in "
            ".changes: UNRELEASED")
