#!/usr/bin/env python
# Ubuntu Continuous Integration Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.

# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from ci_libs.file_handler import (ChangesProcessor, ChangesFileNotFound,
                                  ChangesParseError, ChangesValidationError,
                                  FileToUploadNotFound, UploadDirNotFound)
from dput.changes import Changes, ChangesFileException
from tests import TestCaseWithGnupg

TEST_DATA_ROOT = os.path.join(os.path.dirname(__file__), 'data')


class ChangesProcessorTestCase(TestCaseWithGnupg):
    def setUp(self):
        super(ChangesProcessorTestCase, self).setUp()
        self.create_test_info()

    def _get_test_fullpath(self, filename):
        return os.path.join(TEST_DATA_ROOT, filename)

    def create_test_info(self, source_name="foobar", source_version="0.1",
                         packaging_version="1"):
        self.source_name = source_name
        self.source_version = source_version
        self.packaging_version = packaging_version
        self.full_version = "%s-%s" % (self.source_version,
                                       self.packaging_version)
        self.changes_file = (
            "%s_%s-%s_source.changes" % (self.source_name, self.source_version,
                                         self.packaging_version))
        self.dsc_file = (
            "%s_%s-%s.dsc" % (self.source_name, self.source_version,
                              self.packaging_version))
        self.debian_tar_gz_file = (
            "%s_%s-%s.debian.tar.gz" % (self.source_name, self.source_version,
                                        self.packaging_version))
        self.orig_file = "%s_%s.orig.tar.gz" % (self.source_name,
                                                self.source_version)

    def test_parse(self):
        processor = ChangesProcessor(self._get_test_fullpath(
                                     self.changes_file))
        processor.parse()
        self.assertIsInstance(processor.changes, Changes)
        self.assertEqual(processor.changes.get_package_name(),
                         self.source_name)

        self.assertEqual(processor.source_package_name, self.source_name)
        self.assertEqual(processor.source_version, self.full_version)

        self.assertEqual(self._get_test_fullpath(self.dsc_file),
                         processor.changes.get_dsc())
        self.assertEqual(self._get_test_fullpath(self.debian_tar_gz_file),
                         processor.changes.get_diff())
        self.assertTrue(self._get_test_fullpath(self.orig_file) in
                        processor.changes.get_files())

    def test_parse_changes_file_not_found(self):
        self.assertRaises(ChangesFileNotFound, ChangesProcessor,
                          "barbaz_source.changes")

    def test_parse_changes_file_is_a_dsc(self):
        # We're only able to tell a file is a real .changes during parsing.
        processor = ChangesProcessor(
            self._get_test_fullpath("foobar_0.1-1_dsc.changes"))
        self.assertRaises(ChangesParseError, processor.parse)

    def test_parse_changes_file_is_something_else(self):
        # We're only able to tell a file is a real .changes during parsing.
        processor = ChangesProcessor(
            self._get_test_fullpath("foobar_0.1-1_random.changes"))
        self.assertRaises(ChangesFileException, processor.parse)

    def test_upload_dir_not_found(self):
        self.assertRaises(UploadDirNotFound, ChangesProcessor,
                          self._get_test_fullpath(self.changes_file),
                          "foo/bar/baz")

    def test_files_to_upload_dir_found(self):
        processor = ChangesProcessor(
            self._get_test_fullpath(self.changes_file),
            files_to_upload_dir=TEST_DATA_ROOT)
        self.assertEquals(processor.files_to_upload_dir, TEST_DATA_ROOT)

    def test_file_to_upload_not_found(self):
        processor = ChangesProcessor(
            self._get_test_fullpath(self.changes_file),
            files_to_upload_dir="tests")
        self.assertRaises(FileToUploadNotFound, processor.parse)

    def test_all_files_found(self):
        processor = ChangesProcessor(
            self._get_test_fullpath(self.changes_file))
        self.assertEqual(len(processor.files_to_upload), 0)
        processor.parse()
        # The three files (dsc, debian.tar.gz and orig) plus changes file
        self.assertEqual(len(processor.files_to_upload), 4)
        processor.validate()
        self.assertTrue(
            processor.changes.get_dsc() in processor.files_to_upload)
        self.assertTrue(
            processor.changes.get_diff() in processor.files_to_upload)
        self.assertTrue(
            self._get_test_fullpath(self.orig_file) in
            processor.files_to_upload)

    def test_files_to_upload_dir_not_provided(self):
        processor = ChangesProcessor(
            self._get_test_fullpath(self.changes_file))
        self.assertEqual(processor.files_to_upload_dir, TEST_DATA_ROOT)

    def test_signature_check_fails(self):
        processor = ChangesProcessor(
            self._get_test_fullpath("foobar_0.1-1_unsigned.changes"))
        processor.parse()
        self.assertRaises(ChangesValidationError, processor.validate)

    def test_validate_distribution_unreleased(self):
        processor = ChangesProcessor(
            self._get_test_fullpath("foobar_0.1-1_unreleased.changes"))
        processor.parse()
        self.assertRaises(ChangesValidationError, processor.validate)

    def test_process(self):
        # Simple test to guarantee minimum integrity of this method, as the
        # parts are being tested elsewhere.
        processor = ChangesProcessor(
            self._get_test_fullpath(self.changes_file))
        processor.process()
        self.assertIsInstance(processor.changes, Changes)
        self.assertEqual(processor.changes.get_package_name(),
                         self.source_name)

        self.assertEqual(len(processor.files_to_upload), 4)
