#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import logging
import os
import re
import shutil
import string
import subprocess
import textwrap
import time
import urllib
import urlparse
import uuid

import ci_utils

from imagebuilder import ImageException
from ci_utils.image_store import ImageStore
from ci_utils import dump_stack, data_store, launchpad


log = logging.getLogger('Imagebuilder')


def _get_args():
    parser = argparse.ArgumentParser(description='Modify a cloud image')
    parser.add_argument(
        "-i", "--image", required=True,
        help="URL for the cloud image to modify"),
    parser.add_argument(
        "-r", "--repository", required=True, action='append',
        help="List of PPAs to add")
    parser.add_argument(
        "-p", "--package", required=True, action='append',
        help="List of extra packages to install")
    parser.add_argument(
        "-s", "--series", required=True,
        help="Series of Ubuntu the image is based on")

    args = parser.parse_args()
    return args


def _parse_ppas(lp, ppalist, series):
    aptlines = []
    keys = {}
    for ppa in ppalist:
        ppamatch = re.match("ppa:(.+)/(.+)", ppa)
        #This will throw an exception for malformed ppas, intentional for now
        (team, repo) = ppamatch.groups()
        ppa = lp.people[team].getPPAByName(name=repo)
        keys[ppa.name] = ppa.signing_key_fingerprint
        if ppa.private:
            #lp.me won't work, you have to get yourself by person
            user = lp.people[launchpad.get_lp_user_name()]
            url = user.getArchiveSubscriptionURL(archive=ppa)
        else:
            url = 'http://ppa.launchpad.net/{}/{}/ubuntu'.format(team, repo)
        aptlines.append('deb {} {} main'.format(url, series))
        aptlines.append('deb-src {} {} main'.format(url, series))
    return aptlines, keys


def _chrooted(chroot, cmd):
    log.info('Running command in image: {0}'.format(cmd))
    if type(cmd) is not list:
        cmd = cmd.split(' ')
    output = subprocess.check_output(['chroot', chroot] + cmd)
    if output:
        log.info(output)


def _apt_get_update(chroot):
    # apt-get update isn't always reliable give it 3 tries:
    for x in (2, 1, 0):
        try:
            _chrooted(chroot, "/usr/bin/apt-get update")
            break
        except subprocess.CalledProcessError as e:
            if x == 0:
                raise e
            else:
                log.info('apt-get update failed, retrying %d more times', x)


def _setup_ppas(chroot, repos, series, status_cb):
    status_cb('importing ppa signing keys...')
    lp = launchpad.lp_login()

    status_cb('grabbing ppa subscription urls and keys...')
    deblines, keys = _parse_ppas(lp, repos, series)

    for ppa, key in keys.iteritems():
        status_cb('importing signing key for: %s' % ppa)
        cmd_base = 'apt-key adv --keyserver keyserver.ubuntu.com --recv-keys '
        _chrooted(chroot, cmd_base + key)

    status_cb('adding ppas to apt list...')
    with open('%s/etc/apt/sources.list.d/ci.list' % chroot, 'a+') as f:
        for debline in deblines:
            f.write('\n%s' % debline)

    status_cb('running apt-get-update...')
    _apt_get_update(chroot)


def _run_chroot_cmds(chroot, pkglist, status_cb):
    '''Run commands in the chroot to do update the sources and add packages
    '''
    status_cb('Adding requested packages to the image...')
    os.environ['DEBIAN_FRONTEND'] = 'noninteractive'
    _chrooted(chroot,
              ['/bin/bash', '-c',
               ('/usr/bin/apt-get -o Dpkg::Options::=\"--force-confdef\" '
                '-o Dpkg::Options::=\"--force-confold\" install -y '
                '{pkglist}'.format(pkglist=string.join(pkglist, " ")))])
    _chrooted(chroot, "/usr/bin/apt-get clean")

    #create a big null data file and remove it to make it more compressable
    status_cb('updating chroot to be more compressible...')
    zerofile = '%s/zero.tmp' % chroot
    filldata = '\0' * 4096
    with open(zerofile, 'wb') as f:
        while True:
            try:
                f.write(filldata)
            except IOError:
                break
    os.unlink(zerofile)


def _download(url, path):
    urlpath = urlparse.urlsplit(url).path
    filename = os.path.basename(urlpath)
    filename = os.path.join(path, filename)
    urllib.urlretrieve(url, filename)
    return filename


def _load_nbd():
        '''load the nbd kernel module, ok even if it's loaded already'''
        runcmd('modprobe nbd')


def runcmd(cmd):
    log.info('Running command: {0}'.format(cmd))
    output = subprocess.check_output(cmd.split(' '))
    if output:
        log.info(output)


class CloudImage(object):
    def __init__(self, image_path, tmpdir):
        self.mountpoint = self._mount_image(image_path, tmpdir)

    def __enter__(self):
        return self.mountpoint

    def __exit__(self, exc, value, tb):
        self._umount_image()

    def _mount_image(self, image_path, tmpdir):
        #Find all the unused network block devices
        mountpoint = '%s/mountpoint' % tmpdir
        os.mkdir(mountpoint)
        nbd_devs = [
            dev for dev in os.listdir('/sys/block') if dev.startswith('nbd')
            and not os.path.exists('/sys/block/%s/pid' % dev)
        ]
        self.nbd_dev = nbd_devs[0]
        runcmd('qemu-nbd -c /dev/%s %s' % (self.nbd_dev, image_path))
        for x in xrange(10):
            if os.path.exists('/sys/block/%s/pid' % self.nbd_dev):
                break
            time.sleep(1)
        else:
            raise ImageException('Block device %s never became available' %
                                 self.nbd_dev)
        runcmd('udevadm settle')
        runcmd('mount /dev/%sp1 %s' % (self.nbd_dev, mountpoint))
        #Mount /proc /sys and /dev in the image
        for bmount in ['/proc', '/sys', '/dev']:
            runcmd('mount --bind %s %s/%s' % (bmount, mountpoint, bmount))
        #Backup resolv.conf and copy in the one from the local system
        if os.path.lexists('%s/etc/resolv.conf' % mountpoint):
            shutil.move('%s/etc/resolv.conf' % mountpoint,
                        '%s/etc/resolv.conf.orig' % mountpoint)
        shutil.copy('/etc/resolv.conf', '%s/etc/resolv.conf' % mountpoint)
        policyfile = textwrap.dedent("""\
            #!/bin/sh
            while true; do
            case "$1" in
                -*) shift ;;
                makedev) exit 0;;
                x11-common) exit 0;;
                *) exit 101;;
            esac
            done""")
        with open('%s/usr/sbin/policy-rc.d' % mountpoint, 'w') as f:
            f.write(policyfile)
        os.chmod('%s/usr/sbin/policy-rc.d' % mountpoint, 0o755)

        return mountpoint

    def _umount_image(self):
        for bmount in ['/proc', '/sys', '/dev']:
            runcmd('umount %s/%s' % (self.mountpoint, bmount))
        if os.path.lexists('%s/etc/resolv.conf.orig' % self.mountpoint):
            shutil.move('%s/etc/resolv.conf.orig' % self.mountpoint,
                        '%s/etc/resolv.conf' % self.mountpoint)
        os.unlink('%s/usr/sbin/policy-rc.d' % self.mountpoint)
        runcmd('umount %s' % self.mountpoint)
        runcmd('qemu-nbd -d /dev/%s' % self.nbd_dev)


def modify_image(image, repos, series, packages, ds, status_cb=None):
    '''Add packges and PPAs to a cloud image mounted in a chroot'''
    if not status_cb:
        status_cb = lambda x: 0   # just provide a no-op

    status_cb('loading nbd...')
    _load_nbd()
    with ci_utils.TmpDir() as tmpdir:
        status_cb('downloading %s...' % image)
        image_path = _download(image, tmpdir)
        with CloudImage(image_path, tmpdir) as mountpoint:
            status_cb('updating image %s...' % image)

            _setup_ppas(mountpoint, repos, series, status_cb)

            _run_chroot_cmds(mountpoint, packages, status_cb)

        glance = ImageStore(ci_utils.unit_config.get_auth_config())
        converted_image_path = os.path.join(tmpdir, 'new.img')
        status_cb('Shrinking the converted image...')
        runcmd('qemu-img convert -O qcow2 -c %s %s' % (
            image_path, converted_image_path))
        upload_filename = str(uuid.uuid1()) + '.img'
        status_cb(
            'Uploading converted image to glance as %s' % upload_filename)
        glance.image_create(upload_filename, converted_image_path)
        try:
            ds.change_visibility()
            with open(converted_image_path, 'r') as fp:
                location = ds.put_file(upload_filename, fp)
        except data_store.DataStoreException:
            status_cb('ERROR: unable to upload image to swift')
        return upload_filename, location


def main():
    dump_stack.install_stack_dump_signal()
    args = _get_args()
    return modify_image(args.image, args.repository, args.series,
                        args.package)

if __name__ == '__main__':
    main()
