#!/usr/bin/python

import argparse
import hashlib
import os
import sys
import subprocess
import stat
import atexit
import shutil
import tempfile
import textwrap
from distutils import spawn
from itertools import chain

import bzrlib
from bzrlib import (
    builtins,
    plugin,
    workingtree,
)


HERE = os.path.abspath(os.path.dirname(__file__))

# the juju-deployer/ directory.
deployer_dir = os.path.abspath(os.path.dirname(sys.argv[0]))
# ci-utils probably isn't in our pythonpath
sys.path.append(os.path.join(os.path.dirname(__file__), '../ci-utils'))
from ci_utils import dump_stack


# Initialize bzrlib and its plugins
bzrlib.initialize()
plugin.load_plugins()


def check_and_set_ppa_privacy(private_ppas_only):
    current_value = os.environ['CI_PRIVATE_PPAS_ONLY']
    if private_ppas_only:
        value = 'True'
        status = 'ENABLED'
    elif current_value not in ['0', '1']:
        print('Invalid CI_PRIVATE_PPAS_ONLY value. Options are 0 (disabled) '
              'and 1 (enabled). Provided value is: "{}"'.format(current_value))
        sys.exit(1)
    else:
        if current_value == '0':
            value = 'False'
            status = 'DISABLED'
        else:
            value = 'True'
            status = 'ENABLED'
    os.environ['CI_PRIVATE_PPAS_ONLY'] = value
    print('Private PPAs only is {} (CI_PRIVATE_PPAS_ONLY: {})'.format(status,
          value))


def check_environment():
    '''Cheetah requires environment variables for populating our templates'''
    cheetah_vars = {
        'OS_USERNAME': None,
        'CI_CODE_SOURCE': 'branch',
        'CI_BRANCH': 'lp:ubuntu-ci-services-itself',
        'CI_PAYLOAD_URL': 'n/a',
        'CI_LAUNCHPAD_USER': None,
        'CI_LAUNCHPAD_PPA_OWNER': None,
        'CI_PRIVATE_PPAS_ONLY': '0',
        'CI_OAUTH_CONSUMER_KEY': None,
        'CI_OAUTH_TOKEN': None,
        'CI_OAUTH_TOKEN_SECRET': None,
        'CI_PPA': 'ppa:canonical-ci-engineering/ci-airline-phase-0',
        'CI_MASTER_PPA': 'ppa:ci-engineering-airline/ci-archive',
    }
    missing_required = []
    for key in cheetah_vars:
        if key not in os.environ or len(os.environ[key]) == 0:
            default = cheetah_vars[key]
            if default:
                os.environ[key] = default
            else:
                missing_required.append(key)

    if len(missing_required) > 0:
        template = textwrap.dedent('''
        ERROR: Missing required environment variables:
          {}
        Please ensure the novarc file has been sourced.''')
        exit(template.format('\n  '.join(missing_required)))


def check_bootstrapped():
    try:
        # XXX: Added a print to give user feedback in case sshuttle isn't
        # running and it "hangs", maybe correct behavior is to add a timeout
        # here, but too much work for the benefit, so I'm only providing a
        # hint now -- Ursinha (2014-02-17)
        print 'Checking juju status'
        subprocess.check_output(['juju', 'status'])
    except subprocess.CalledProcessError:
        # There's no need to print an error message as juju status already
        # does that.
        sys.exit(1)


def check_dependencies():
    '''Exit the application if we are missing any dependency programs'''

    needed_commands = ['cheetah']
    for cmd in needed_commands:
        if not spawn.find_executable(cmd):
            err = "%s is missing. Please install it before continuing." % cmd
            print >>sys.stderr, err
            sys.exit(1)


def populate_templates(deployer_dir, working_dir):
    # The command to populate the template.
    cmd_base = ['cheetah', 'fill', '--env', '-p']

    for tmpl in os.listdir(deployer_dir):
        path = os.path.join(deployer_dir, tmpl)
        if os.path.isdir(path):
            src = os.path.join(deployer_dir, tmpl)
            dst = os.path.join(working_dir, tmpl)
            populate_templates(src, dst)
            continue
        if not tmpl.endswith('.yaml.tmpl'):
            continue
        # Strip the .tmpl off .yaml.tmpl and write a novarc-populated version
        # of the yaml to our working directory.
        with open(os.path.join(working_dir, tmpl[:-5]), 'w') as out:
            args = [os.path.join(deployer_dir, tmpl)]
            subprocess.check_call(cmd_base + args, stdout=out)


def generate_argument_list(d, specific=None):
    for f in os.listdir(d):
        if not f.endswith('.yaml'):
            continue
        # If we're looking for a specific set of components to deploy and this
        # isn't one of them, continue.
        if specific and f[:-5] not in specific:
            continue
        yield ('-c', os.path.join(d, f))


def prepend_path(env_var, path):
    """Put our dependencies paths first so they don't get hidden."""
    val = os.environ.get(env_var, None)
    paths = (val and val.split(os.path.pathsep)) or []
    paths.insert(0, path)
    os.environ[env_var] = os.path.pathsep.join(paths)


def get_or_pull_branch(src, dst, revision=None, out=None, err=None):
    """Get a branch at a given revision.

    If the branch already exists, pulls the given revision.

    :param src: The url to get the branch from (any url recognized by bzr).

    :param dst: The directory where the branch will reside.

    :param revision: The optional revision to set the branch to.
    """
    with bzrlib.initialize(stdout=out, stderr=err):
        if os.path.exists(dst):
            cmd = builtins.cmd_pull()
            args = [src, '-d', dst, '--overwrite']
        else:
            cmd = builtins.cmd_branch()
            args = [src, dst]
        if revision is not None:
            args.extend(['--revision', revision])
        cmd.run_argv_aliases(args)


def install_websocket_client(where=None):
    """Install python-websocket-client from bzr, as jujuclient needs it."""
    if where is None:
        where = os.path.join(HERE, '..', 'branches')
    dest = os.path.abspath(os.path.join(where, 'websocket-client'))
    # r106 is 0.12.0
    get_or_pull_branch('lp:websocket-client', dest, 'revno:106')
    prepend_path('PYTHONPATH', dest)
    return dest


def install_jujuclient(where=None):
    """Install python-jujuclient from bzr. r28 fixes bugs we were seeing around
    resetting the websocket connection."""
    if where is None:
        where = os.path.join(HERE, '..', 'branches')
    dest = os.path.abspath(os.path.join(where, 'jujuclient'))
    get_or_pull_branch('lp:python-jujuclient', dest, 'revno:28')
    prepend_path('PYTHONPATH', dest)
    return dest


def install_deployer(where=None):
    """Install juju-deployer from bzr."""
    if where is None:
        where = os.path.join(HERE, '..', 'branches')
    dest = os.path.abspath(os.path.join(where, 'juju-deployer'))
    # For an obscure reason tag:0.3.1 is not in lp:juju-deployer
    get_or_pull_branch('lp:juju-deployer', dest, revision='revno:103')
    prepend_path('PYTHONPATH', dest)
    # FIXME: This duplicates setup.py and shouldn't be needed. This specific
    # piece (juju-deployer) is a weird beast, it installs from source (not pip
    # nor package), it requires amulet which itself requires urllib3 (among
    # others), it's used both in unit tests that are executed in a venv and in
    # ./tests/run.py which is not executed in venv. My head hurts -- vila
    # 2014-02-12
    deployer_bin = os.path.join(dest, 'juju-deployer')
    with open(deployer_bin, 'w') as fp:
        tmpl = '#!/bin/sh\nPYTHONPATH={}\nexec python {}/deployer/cli.py $*'
        fp.write(tmpl.format(os.environ['PYTHONPATH'], dest))
    st = os.stat(deployer_bin)
    mode = st.st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH
    os.chmod(deployer_bin, mode)
    prepend_path('PATH', dest)
    return dest


def build_charms(charmsdir=None):
    # juju needs to know about our local charms.
    # and some of our charms require us to call "make" on them
    if not charmsdir:
        charmsdir = os.path.abspath(
            os.path.join(os.path.dirname(__file__), '../charms'))
    os.environ['JUJU_REPOSITORY'] = charmsdir

    def _cleanup(local_charms):
        # delete any charms pulled down by juju-deployer into our local tree
        if not os.path.exists(charmsdir):
            return  # handle special case for unit test harness
        for release in local_charms:
            charms = os.listdir(os.path.join(charmsdir, release))
            remove = set(charms) - set(local_charms[release])
            for r in remove:
                shutil.rmtree(os.path.join(charmsdir, release, r))

    # Feed local_charms with our charms so that if anything goes wrong during
    # the building phase, we'll never remove them.
    local_charms = {'precise': ['lander-jenkins', 'python-django',
                                'rabbitmq-worker', 'restish', 'webui']}
    clean_func = lambda: _cleanup(local_charms)
    atexit.register(clean_func)

    for release in os.listdir(charmsdir):
        if local_charms.get(release) is None:
            local_charms[release] = []
        for charm in os.listdir(os.path.join(charmsdir, release)):
            local_charms[release].append(charm)
            charmdir = os.path.join(charmsdir, release, charm)
            if os.path.exists(os.path.join(charmdir, 'Makefile')):
                print('Building charm: {}'.format(charm))
                subprocess.check_call(['make'], cwd=charmdir)
    return clean_func


def build_payload(path):
    """Build the payload as a compressed tarball at the given path.

    :returns: A fingerprint uniquely identifying the payload.
    """
    payload_path = '.'
    cmd = builtins.cmd_export()
    # We want uncommitted changes. We take file timestamps as they carry useful
    # information (dir timestamps are broken though).
    cmd.run_argv_aliases([path, '--root', payload_path,
                          '--per-file-timestamps',
                          '--format=tgz', '--uncommitted'])
    # Build the fingerprint from everything included in the tarball, dirs have
    # no fingerprint (but paths are included so we don't miss renames).
    sha1 = hashlib.sha1()
    wt, _ = workingtree.WorkingTree.open_containing(payload_path)
    wt.lock_read()
    try:
        ids = wt.all_file_ids()
        for i in ids:
            i_sha = wt.get_file_sha1(i)
            if i_sha is not None:  # dirs have no sha
                sha1.update(i_sha)
            sha1.update(wt.id2path(i))
    finally:
        wt.unlock()
    return sha1.hexdigest()


def upload_payload_to_swift():
    """Upload the versions controlled code tree to swift.

    Also set the relevant environment variables used to populate the deployer
    scripts and charm configurations.
    """
    base = os.path.join(os.path.dirname(__file__), '..')
    print('Preparing local branch upload...')

    ciutils = os.path.join(base, 'ci-utils')
    sys.path.append(ciutils)
    from ci_utils import data_store

    config = {
        'auth_url': os.environ['OS_AUTH_URL'],
        'auth_user': os.environ['OS_USERNAME'],
        'auth_password': os.environ['OS_PASSWORD'],
        'auth_tenant_name': os.environ['OS_TENANT_NAME'],
        'auth_region': os.environ['OS_REGION_NAME']
    }
    ds = data_store.DataStore('ci-payload', config, public=True)

    with tempfile.NamedTemporaryFile() as ball:
        fingerprint = build_payload(ball.name)
        # The sha-1 fingerprint is used as the name of the file in swift so we
        # can avoid uploading if it's already there.
        if fingerprint not in ds.list_files():
            with open(ball.name) as f:
                print('Uploading local branch,'
                      ' fingerprint {}'.format(fingerprint))
                payload_url = ds.put_file(fingerprint, f)
        else:
            print('Reusing local branch, fingerprint {}'.format(fingerprint))
            payload_url = ds.file_path(fingerprint)
    # the django charm uses branch for both meanings of payload
    os.environ['CI_CODE_SOURCE'] = 'tarball'
    os.environ['CI_PAYLOAD_URL'] = payload_url
    os.environ['CI_BRANCH'] = payload_url


def _get_args(args=None):
    parser = argparse.ArgumentParser(
        description='use juju-deployer to deploy services')
    parser.add_argument('--branch-payload', action='store_true',
                        help='''Use the current tip of this payload branch for
                        the code-base rather than a tarball of the current
                        directory.''')
    parser.add_argument('--reuse-payload', action='store_true',
                        help='''Use the already uploaded payload for
                        the code-base rather than a tarball of the current
                        directory. Use at your own risk.''')
    parser.add_argument('--private-ppas-only', action='store_true',
                        help='''Sets the environment to use only private PPAs
                        when building packages.''')
    parser.add_argument('specific', metavar='<config.yaml>', nargs='*',
                        help='Use a specific list of juju-deployer configs.')

    return parser.parse_args(args)


def main():
    args = _get_args()

    check_environment()

    # You can configure the system to use private PPAs only or by providing
    # the option --private-ppas-only or by setting the environment variable
    # CI_PRIVATE_PPAS_ONLY.
    # If provided, --private-ppas-only option overrides the environment
    # variable value.
    check_and_set_ppa_privacy(args.private_ppas_only)

    check_bootstrapped()
    check_dependencies()

    install_websocket_client()
    install_jujuclient()
    install_deployer()

    # A place to put our generated .yaml files.
    temp_dir = tempfile.mkdtemp()
    atexit.register(shutil.rmtree, temp_dir)
    working_dir = os.path.join(temp_dir, 'deployer')
    # Copy all existing .yaml files, they don't need to be populated but
    # juju-deployer expect them in 'working_dir'.
    shutil.copytree(deployer_dir, working_dir)

    if not args.branch_payload and not args.reuse_payload:
        upload_payload_to_swift()

    # Create .yaml files from .yaml.tmpl files.
    populate_templates(deployer_dir, working_dir)

    # Generate an argument list of -c FILE for each .yaml file in our working
    # directory.
    i = generate_argument_list(working_dir, args.specific)
    generated = list(chain.from_iterable(i))

    build_charms()

    args = ['juju-deployer', '-v'] + generated + ['ci-airline-staging']
    print 'Running {}'.format(' '.join(args))
    try:
        subprocess.check_call(args,
                              cwd=os.path.join(deployer_dir, '../charms'))
    except Exception as exc:
        print 'Problem deploying "ci-airline-staging": {}'.format(str(exc))

if __name__ == '__main__':
    dump_stack.install_stack_dump_signal()
    main()
