#!/usr/bin/env python

import mock
import unittest
import os
import sys
import tempfile
import shutil
import subprocess


from ucitests import fixtures


sys.path.append(os.path.abspath(os.path.dirname(sys.argv[0])))
import deploy


class TestDeploy(unittest.TestCase):

    def test_check_environment(self):
        # If we haven't sourced a novarc, this program should bail out.
        with mock.patch.dict('os.environ', {'OS_USERNAME': ''}):
            with self.assertRaises(SystemExit):
                deploy.check_environment()

        # check that some of our values can be defaulted correctly:
        env_required = {
            'OS_USERNAME': 'os-user',
            'CI_OAUTH_CONSUMER_KEY': 'key',
            'CI_OAUTH_TOKEN': 'token',
            'CI_OAUTH_TOKEN_SECRET': 'secret',
            'CI_LAUNCHPAD_USER': 'lp-user',
            'CI_LAUNCHPAD_PPA_OWNER': 'lp-ppa-owner',
        }
        with mock.patch.dict('os.environ', env_required):
            deploy.check_environment()
            for key, val in env_required.iteritems():
                self.assertEqual(val, os.environ[key])
            # defaulted by code
            self.assertEqual(os.environ['CI_CODE_SOURCE'], 'branch')
            self.assertEqual(os.environ['CI_PRIVATE_PPAS_ONLY'], '0')

    def test_check_dependencies(self):
        # If we cannot find any of our dependencies, this program should bail
        # out.
        with mock.patch.dict('os.environ', {'PATH': ''}):
            with self.assertRaises(SystemExit):
                deploy.check_dependencies()

    def test_check_not_bootstrapped(self):
        # If juju environment isn't bootstrapped, this program should bail
        # out.
        with mock.patch('subprocess.check_output',
                        side_effect=subprocess.CalledProcessError(
                            -1, 'juju status')):
            with self.assertRaises(SystemExit):
                deploy.check_bootstrapped()

    def test_check_bootstrapped(self):
        with mock.patch('subprocess.check_output', return_value='ok'):
            # If juju status returns ok, this should just work.
            deploy.check_bootstrapped()

    def test_install_deployer(self):
        fixtures.set_uniq_cwd(self)
        with mock.patch.dict('os.environ', {'PYTHONPATH': ''}):
            deployer_dir = deploy.install_deployer('.')
            self.assertTrue(os.path.exists(os.path.join(deployer_dir, '.bzr')))
            self.assertEqual(deployer_dir, os.environ['PYTHONPATH'])
            deployer_script = os.path.join(deployer_dir, 'juju-deployer')
            self.assertTrue(os.access(deployer_script, os.X_OK))
            self.assertTrue(os.path.exists(deployer_script))

    def test_generate_empty_list(self):
        working = tempfile.mkdtemp()
        self.addCleanup(lambda: shutil.rmtree(working))

        # Calling without anything in the directory should produce an empty
        # list.
        self.assertEqual(list(deploy.generate_argument_list(working)), [])

    def test_generate_explicit_list(self):
        working = tempfile.mkdtemp()
        self.addCleanup(lambda: shutil.rmtree(working))

        expected = os.path.join(working, 'foo.yaml')
        with open(expected, 'w') as fp:
            fp.write('nothing')
        # Calling with a basic configuration should produce the full path to
        # that configuration in the form ('-c', path).
        gen = deploy.generate_argument_list(working)
        self.assertEqual([('-c', expected)], list(gen))

    def test_populate_templates_empty_list(self):
        fake_deployer_dir = tempfile.mkdtemp()
        self.addCleanup(lambda: shutil.rmtree(fake_deployer_dir))
        working = tempfile.mkdtemp()
        self.addCleanup(lambda: shutil.rmtree(working))

        deploy.populate_templates(fake_deployer_dir, working)
        # Nothing happened
        self.assertEqual([], os.listdir(working))

    @mock.patch.dict('os.environ', {'USER': 'foo'})
    def test_populate_templates_explicit_list(self, *args):
        fake_deployer_dir = tempfile.mkdtemp()
        self.addCleanup(lambda: shutil.rmtree(fake_deployer_dir))
        working = tempfile.mkdtemp()
        self.addCleanup(lambda: shutil.rmtree(working))

        test_files = ['config.yaml', 'configs/unit_config.yaml']
        for f in test_files:
            src = os.path.join(fake_deployer_dir, f)
            dst = os.path.join(working, f)
            if not os.path.exists(os.path.dirname(src)):
                os.mkdir(os.path.dirname(src))
            if not os.path.exists(os.path.dirname(dst)):
                os.mkdir(os.path.dirname(dst))

            with open(src + '.tmpl', 'w') as f:
                f.write('me: ${USER}')

        deploy.populate_templates(fake_deployer_dir, working)

        for f in test_files:
            f = os.path.join(working, f)
            content = open(f).read()
            self.assertEqual('me: {}'.format(os.environ['USER']), content)

    def test_build_charms(self):
        charmsdir = tempfile.mkdtemp()
        self.addCleanup(lambda: shutil.rmtree(charmsdir))
        p = os.path.join(charmsdir, 'precise')
        os.mkdir(p)
        charm1 = os.path.join(p, 'charm1')
        os.mkdir(charm1)
        charm2 = os.path.join(p, 'charm2')
        os.mkdir(charm2)
        with open(os.path.join(charm2, 'Makefile'), 'w') as f:
            f.write('all:\n')
            f.write('\tmkdir ../build_charms_should_delete\n')
            # webui is precious and should not be deleted
            f.write('\tmkdir ../webui\n')
            f.write('\ttouch make_called\n')

        clean = deploy.build_charms(charmsdir)
        clean()

        self.assertTrue(os.path.exists(os.path.join(charm2, 'make_called')))
        self.assertItemsEqual(['charm1', 'charm2', 'webui'], os.listdir(p))

if __name__ == '__main__':
    unittest.main()
