#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Why are we pinning branches to specific revisions?
# --------------------------------------------------
# * Always deploying the charms from tip breaks test isolation.
# * We need to ensure we always deploy the same code to production, code that
#   we've completely tested.

# TODO: We should wire update.py --assert-pinned into the tests, once this has
# landed and we've done a follow up change to actually run it and pin.
# -- ev 2014-02-04

import argparse
import logging
import os
import re
import sys
import yaml
from ci_utils import dump_stack

from bzrlib import (
    branch,
    plugin,
)


logging.basicConfig(level=logging.INFO, format='%(msg)s')
log = logging.getLogger(__name__)
# We need to load the Launchpad bzr plugin.
plugin.load_plugins()


class UpdateArgParser(argparse.ArgumentParser):
    """A parser for the uci-run-tests script."""

    def __init__(self):
        description = 'Update juju-deployer branch references in configs.'
        super(UpdateArgParser, self).__init__(
            prog='update.py', description=description)
        # Optional arguments
        self.add_argument(
            '--assert-pinned', action='store_true',
            dest='assert_pinned',
            help='Fails if some branches are not revno-pinned.')
        self.add_argument(
            '--check', action='store_true',
            dest='check',
            help='Check the branches without updating the configs.')

    def parse_args(self, args=None, out=None, err=None):
        """Parse arguments, overridding stdout/stderr if provided.

        Overridding stdout/stderr is provided for tests.

        :params args: The arguments to the script.

        :param out: Default to sys.stdout.

        :param err: Default to sys.stderr.

        :return: The populated namespace.
        """
        out_orig = sys.stdout
        err_orig = sys.stderr
        try:
            if out is not None:
                sys.stdout = out
            if err is not None:
                sys.stderr = err
            return super(UpdateArgParser, self).parse_args(args)
        finally:
            sys.stdout = out_orig
            sys.stderr = err_orig


def get_deployer_configs(deployer_dir):
    for x in os.listdir(deployer_dir):
        if x.endswith('.yaml') or x.endswith('.yaml.tmpl'):
            yield os.path.join(deployer_dir, x)


def _iter_branches_and_revnos(paths):
    for cfg in paths:
        with open(cfg) as fp:
            y = yaml.safe_load(fp.read())

        for deployment in y.itervalues():
            services = deployment.get('services', {})
            for service in services.itervalues():
                service_branch = service.get('branch')
                if not service_branch:
                    continue

                branch_url, _, cur = service_branch.partition('@')
                yield (branch_url, cur)


def ensure_all_branches_are_pinned(paths, warn=False):
    pinned = True
    for branch_url, revno in _iter_branches_and_revnos(paths):
        if not revno:
            if warn:
                log.error('{} is not pinned to a revno.'.format(branch_url))
            pinned = False
    return pinned


def get_branches_and_revnos(paths, warn=False):
    branches = {}
    for branch_url, cur in _iter_branches_and_revnos(paths):
        if branch_url not in branches:
            b = branch.Branch.open(branch_url)
            revno = int(b.revno())
            branches[branch_url] = revno

            if warn and not cur:
                t = '{} is not pinned (latest {})'
                log.warning(t.format(branch_url, revno))

            if warn and cur and int(cur) < revno:
                t = '{} is out of date (currently {}, latest {})'
                log.warning(t.format(branch_url, int(cur), revno))

    return branches


def set_branches_and_revnos(paths, branches):
    """Update the yaml files with the new revnos.

    We cannot use pyyaml for this, as it doesn't preserve order or preserve
    comments.
    """

    for cfg in paths:
        with open(cfg) as fp:
            body = fp.read()

        # Start with the original body
        new_body = body
        for branch_url in branches:
            # Replace all occurences of the naked branch url with one
            # specifying the revno
            pattern = '{}[@0-9]*'.format(branch_url)
            repl = '{}@{}'.format(branch_url, branches[branch_url])
            new_body = re.sub(pattern, repl, new_body)

        if body != new_body:
            # Save the modified body
            with open(cfg, 'w') as fp:
                fp.write(new_body)


def main(args=None):
    if args is None:
        args = sys.argv[1:]

    parser = UpdateArgParser()
    ns = parser.parse_args(args)

    deployer_dir = os.path.dirname(sys.argv[0])
    paths = list(get_deployer_configs(deployer_dir))

    if ns.assert_pinned:
        all_pinned = ensure_all_branches_are_pinned(paths, True)
        return not bool(all_pinned)
    else:
        branches = get_branches_and_revnos(paths, ns.check)
        if not ns.check:
            set_branches_and_revnos(paths, branches)
        return 0


if __name__ == '__main__':
    dump_stack.install_stack_dump_signal()
    sys.exit(main())
