#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2013, 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import logging
import os
import sys
import yaml

sys.path.append(os.path.join(os.path.dirname(__file__), '../../ci-utils'))
from ci_utils import data_store
from ci_utils import jenkins_utils
from ci_utils import dump_stack

from ticket_api import TicketApi

DEFAULT_DATASTORE_CONFIG = os.path.join(
    os.path.dirname(__file__), '../../unit_config')


class LanderArchiver():
    '''Handles archiving of artifacts generated by the lander.'''

    def __init__(self):
        logging.basicConfig(level=logging.INFO)
        self.log = logging.getLogger(__name__)

    def parse_arguments(self):
        parser = argparse.ArgumentParser(
            description='''Archives the provided job artifacts to the
            data store.''')
        parser.add_argument('--source-url',
                            help='The source build to archive.',
                            required=True)
        parser.add_argument('--request-id',
                            help='The originating request id.',
                            required=True)
        parser.add_argument('--ticket-system',
                            help='The ticket-system URL.',
                            required=True)
        parser.add_argument('--job-name',
                            help='The originating job name.',
                            required=True)
        parser.add_argument('--datastore-config',
                            help='''Location of the datastore auth
                            configuration file.''',
                            default=DEFAULT_DATASTORE_CONFIG)
        return parser.parse_args()

    def get_datastore_config_contents(self, filename):
        with open(filename) as f:
            return yaml.safe_load(f)

    def get_datastore_config(self, filename):
        try:
            return self.get_datastore_config_contents(filename)
        except IOError as e:
            self.log.error('Failed to load datastore configuration: '
                           '{}'.format(e))

    def get_datastore(self, datastore_config, request_id):
        try:
            return data_store.create_for_ticket(request_id, datastore_config)
        except data_store.DataStoreException as e:
            self.log.error('Failed to create datastore: {}'.format(e))

    def get_artifact_list(self, url):
        data = jenkins_utils.get_json(url)
        artifact_list = []
        for artifact in data['artifacts']:
            artifact_list.append(artifact['relativePath'])
        return artifact_list

    def store_artifact(self, datastore, name, url, content_type=None):
        try:
            return datastore.put_file(name,
                                      jenkins_utils.get_url_contents(url),
                                      content_type)
        except data_store.DataStoreException as e:
            self.log.error('Failed to archive artifact [{}]: '
                           '{}'.format(name, e))

    def archive_console_log(self, datastore, job_name, source_url):
        artifact_name = '{}.log'.format(job_name)
        location = self.store_artifact(
            datastore, artifact_name,
            jenkins_utils.get_console_log_url(source_url),
            'text/plain')
        self.log.info('Console log: {}'.format(location))
        self.ticket_api.add_artifact(artifact_name, location)

    def archive_artifacts(self, datastore, job_name, source_url):
        artifact_list = self.get_artifact_list(source_url)
        for artifact in artifact_list:
            artifact_name = '{}.{}'.format(job_name,
                                           artifact.replace('/', '_'))
            location = self.store_artifact(
                datastore, artifact_name,
                jenkins_utils.get_artifact_url(source_url, artifact))
            self.log.info('Artifact [{}]: {}'.format(artifact, location))
            self.ticket_api.add_artifact(artifact_name, location)

    def __call__(self):
        args = self.parse_arguments()
        datastore_config = self.get_datastore_config(
            args.datastore_config)
        if not datastore_config:
            return -1
        datastore = self.get_datastore(datastore_config, args.request_id)
        if not datastore:
            return -2

        self.ticket_api = TicketApi(args.ticket_system, args.request_id)

        try:
            self.archive_console_log(datastore, args.job_name, args.source_url)
            self.archive_artifacts(datastore, args.job_name, args.source_url)
        except EnvironmentError as e:
            self.log.error('{}'.format(e))
            return -3

        return 0


if __name__ == '__main__':
    dump_stack.install_stack_dump_signal()
    archiver = LanderArchiver()
    sys.exit(archiver())
