#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2013, 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import json
import logging
import os
import signal
import sys
import time
import urllib2
import yaml

# the worker might not have installed this module, so determine the path
# and add it, so we can always safely import stuff
sys.path.append(os.path.join(os.path.dirname(__file__), '../../ci-utils'))
from ci_utils import amqp_utils
from ci_utils import dump_stack

from ticket_api import TicketApi

format_list = {'json', 'yaml'}

# set up a root logger for everything
stdout_handler = logging.StreamHandler(stream=sys.stdout)
formatter = logging.Formatter(
    '%(asctime)s %(levelname)-5s %(name)s: %(message)s', datefmt='%H:%M:%S')
stdout_handler.setFormatter(formatter)
l = logging.getLogger('')
l.addHandler(stdout_handler)
l.setLevel(logging.DEBUG)

logger = logging.getLogger('lander_service_wrapper')


class MessageSync(object):
    progress = logging.getLogger('PROGRESS_TRIGGER')

    def __init__(self, ticket, started_cb=None, completed_cb=None,
                 message_cb=None):
        if not started_cb:
            started_cb = lambda: True
        if not completed_cb:
            completed_cb = lambda x: True
        if not message_cb:
            message_cb = lambda x: True
        self.ticket = ticket
        self.started_cb = started_cb
        self.completed_cb = completed_cb
        self.message_cb = message_cb

    def on_message(self, msg):
        data = json.loads(msg.body)
        state = data['state']

        # print "state" as a label for the rest of the items in the list
        params = {x: y for (x, y) in data.iteritems() if x != 'state'}
        if params:
            self.progress.info('%s: %s', state, params)
        else:
            self.progress.info(state)

        self.message_cb(data)

        if state == 'STATUS':
            self.started_cb()
        elif state == 'COMPLETED':
            self.ticket.process_artifacts(data)
            if data['result'] == 'PASSED':
                self.completed_cb(True)
            else:
                self.completed_cb(False)

        msg.channel.basic_ack(msg.delivery_tag)
        if data.get('exit'):
            self.data = data
            # this will kick us out of the run-forever worker loop
            msg.channel.basic_cancel(msg.consumer_tag)


def _drain_progress(amqp_config, queue, ticket, started_cb=None,
                    completed_cb=None, message_cb=None):
    sync = MessageSync(ticket, started_cb, completed_cb, message_cb)
    amqp_utils.process_queue(amqp_config, queue, sync.on_message, delete=True)
    return sync.data


def _post(url, data):
    logger.info('Calling via POST [{}]: {}'.format(url, data))
    req = urllib2.Request(url)
    req.add_header('Content-Type', 'application/json')
    resp = urllib2.urlopen(req, json.dumps(data))
    if resp.getcode() == 204:
        return None  # 204 is an empty post reply
    return resp.headers['location']


def _patch(url, data):
    logger.info('Calling via PATCH [{}]: {}'.format(url, data))
    req = urllib2.Request(url)
    req.add_header('Content-Type', 'application/json')
    req.get_method = lambda: 'PATCH'
    urllib2.urlopen(req, json.dumps(data))


def _get(url):
    logger.info('Calling via GET [{}]'.format(url))
    resp = urllib2.urlopen(url)
    resp = resp.read()
    return json.loads(resp)


def _ticket_api(config):
    master = config['master']
    api = TicketApi(master['ts_url'], master['request_id'])

    def _on_sigterm(signame, frame):
        logger.error('job was terminated, failing ticket')
        api.fail_ticket()

    signal.signal(signal.SIGTERM, _on_sigterm)
    return api


def _get_ppa(ppa_assigner_url, ticket_id):
    timeout = 15
    url = '%s/api/v1/ppa/' % ppa_assigner_url
    while True:
        try:
            loc = _post(url, {'ticket_id': ticket_id})
            return loc
        except urllib2.HTTPError as e:
            if e.getcode() != 404:
                raise e
            logging.warn('No free PPAs can be found. Retrying in %ds', timeout)
            time.sleep(timeout)


def _handle_ppa_assigner(args):
    config = _get_config(args)
    loc = _get_ppa(
        config['master']['ppa_assigner_url'], config['master']['request_id'])
    data = _get(loc)
    result = {
        'result': 'PASSED',
        'ppa': data['name'],
        'location': loc,
    }
    return result


def _handle_ppa_assigner_release(args):
    config = _get_config(args)
    if 'ppa_assigner' in config:
        url = config['ppa_assigner'].get('location', False)
        if url:
            _patch(url, {'state': 'dirty'})
        else:
            logger.info('ppa was never reserved by this job')
    else:
        logger.info('ppa_assigner was not run. nothing to do')

    return {'result': 'PASSED'}


def _handle_bsbuilder(args):
    config = _get_config(args)
    ticket = _ticket_api(config)
    url = '%s/api/v1/build_source' % config['master']['bsb_url']

    queue = '%s-bsbuilder' % config['master']['progress_trigger']

    request_parameters = ticket.get_full_ticket()
    params = {
        'ticket_id': config['master']['request_id'],
        'subtickets': request_parameters['subticket'],
        'cancel_url': os.environ['BUILD_URL'] + 'api/json',
        'series': request_parameters.get('series', ''),
        'ppa': config['ppa_assigner']['ppa'],
        'archive': request_parameters.get('master_ppa', ''),
        'progress_trigger': queue,
    }

    ticket.set_pkg_waiting()
    try:
        _post(url, params)

        logger.info('starting progress handler...')
        return _drain_progress(args.amqp_config, queue, ticket,
                               ticket.set_pkg_inprogress,
                               ticket.set_pkg_complete,
                               ticket.set_subticket_progress)
    except:
        ticket.set_pkg_complete(False)
        raise


def _get_binary_packages(ticket, request_parameters):
    # Start with the set of golden binary packages
    package_list = set([x['name'] for x in ticket.get_binaries()])

    # Extract the list of added packages from the ticket request
    try:
        package_list = package_list | set(
            request_parameters.get('added_binaries', None).split(','))
    except AttributeError:
        # Expected when added_binaries is None.
        pass

    # Finally remove the list of removed binaries from the ticket request
    try:
        removed_list = set(
            request_parameters.get('removed_binaries', None).split(','))
        # Log a warning if any package in the removed_binaries is not in the
        # package_list. This is not an error because multiple queued tickets
        # may attempt to remove the same package. However, This could be an
        # indicator of a mis-typed package name.
        diff_list = removed_list - package_list
        if diff_list:
            logger.warning('Package(s) to remove not found: '
                           '{}'.format(list(diff_list)))
        package_list = package_list - removed_list

    except AttributeError:
        # Expected when removed_binaries is None.
        pass

    return list(package_list)


def _handle_image_builder(args):
    config = _get_config(args)
    ticket = _ticket_api(config)

    url = '%s/api/v1/build_image' % config['master']['imgbuilder_url']

    queue = '%s-imgbuilder' % config['master']['progress_trigger']

    # Extract the image parameters from the master request parameters
    request_parameters = ticket.get_full_ticket()
    # TODO Remove the hardcoded cloud, do same as series and base_image
    image = {'image_type': 'cloud',
             'series': request_parameters.get('series', ''),
             'url_list': [request_parameters.get('base_image'), '']}
    ppa_list = [
        config['bsbuilder']['ppa'],
        request_parameters['master_ppa'],
    ]
    params = {
        'ticket_id': config['master']['request_id'],
        'cancel_url': os.environ['BUILD_URL'] + 'api/json',
        'base_image': image,
        'package_list': _get_binary_packages(ticket, request_parameters),
        'ppa_list': ppa_list,
        'progress_trigger': queue,
    }
    ticket.set_image_waiting()
    try:
        _post(url, params)
        logger.info('starting progress handler...')
        return _drain_progress(args.amqp_config, queue, ticket,
                               ticket.set_image_inprogress,
                               ticket.set_image_complete)
    except:
        ticket.set_image_complete(False)
        raise


def _handle_publisher(args):
    config = _get_config(args)
    ticket = _ticket_api(config)

    url = '%s/api/v1/ppa_copy/' % config['master']['ppa_assigner_url']

    request_parameters = ticket.get_full_ticket()
    packages = []
    try:
        for subticket in request_parameters['subticket']:
            name = subticket['source_package_upload']['sourcepackage']['name']
            packages.append(name)
    except AttributeError:
        # Expected if there is no source package upload
        pass

    params = {
        'series': request_parameters['series'],
        'packages': packages,
        'source': config['bsbuilder']['ppa'],
        'target': request_parameters['master_ppa'],
    }
    try:
        ticket.set_publishing_inprogress()
        _post(url, params)
        ticket.set_publishing_complete(True)
    except:
        ticket.set_publishing_complete(False)
        raise

    result = {
        'result': 'PASSED',
        'packages': params['packages'],
    }
    return result


def _handle_test_runner(args):
    config = _get_config(args)
    ticket = _ticket_api(config)

    url = '%s/api/v1/test_image' % config['master']['tr_url']

    # The queue should be unique
    queue = '%s-testrunner' % config['master']['progress_trigger']
    request_parameters = ticket.get_full_ticket()
    packages = []
    try:
        packages = request_parameters.get('added_binaries', None).split(',')
    except AttributeError:
        # Expected when added_binaries is None.
        pass

    params = {
        'ticket_id': config['master']['request_id'],
        'cancel_url': os.environ['BUILD_URL'] + 'api/json',
        'image_id': config['image_builder']['image_id'],
        'package_list': packages,
        'progress_trigger': queue,
    }
    ticket.set_testing_waiting()
    try:
        _post(url, params)
        logger.info('starting progress handler...')
        return _drain_progress(args.amqp_config, queue, ticket,
                               ticket.set_testing_inprogress,
                               ticket.set_testing_complete)
    except:
        ticket.set_testing_complete(False)
        raise


def _get_config(args):
    with open(args.input_file) as f:
        if args.input_format == 'json':
            config = json.loads(f.read())
        else:
            config = yaml.safe_load(f)
    return config


def _get_service_handlers():
    '''get the availble service handlers.

    This is a handy mechanism so we don't have to maintain a list of available
    service handlers. Any function prefixed with '_handle_' will be exposed to
    the user.
    '''
    return [x[8:] for x in globals() if x.startswith('_handle_')]


def _get_parser():
    parser = argparse.ArgumentParser(
        description='Wraps the REST API calls with a progress queue monitor.')
    parser.add_argument('--service',
                        required=True,
                        choices=_get_service_handlers(),
                        help='The service name to call.')
    parser.add_argument('--input-file',
                        required=True,
                        help='File containing all of the input parameters.')
    parser.add_argument('--input-format',
                        required=False,
                        choices=format_list,
                        default='json')
    parser.add_argument('--output-file',
                        required=True,
                        help='File to store the API call results.')
    parser.add_argument('--output-format',
                        required=False,
                        choices=format_list,
                        default='json')
    return parser


def main(args):
    args.amqp_config = amqp_utils.get_config()
    if not args.amqp_config:
        return 1  # the get_config code prints an error

    handler = globals()['_handle_' + args.service]
    logger.debug('Executing service wrapper: {}'.format(args.service))

    rc = 1
    try:
        results = handler(args)
        if results['result'] == 'PASSED':
            rc = 0
    except urllib2.HTTPError as e:
        body = e.read()
        logging.exception('Unable to handle request: %s: %s', e, body)
        results = {
            'result': 'ERROR',
            'resp_code': e.reason,
            'resp_body': body,
        }
    except Exception as e:
        logging.exception('Unexpected error: %s', e)
        results = {
            'result': 'FAILED',
            'error': e.message,
        }

    with open(args.output_file, 'w') as output_file:
        out_data = None
        if args.output_format == 'json':
            out_data = json.dumps(results)
        elif args.output_format == 'yaml':
            out_data = yaml.dump(results)
        output_file.write('{}\n'.format(out_data))

    return rc


if __name__ == '__main__':
    dump_stack.install_stack_dump_signal()
    args = _get_parser().parse_args()
    exit(main(args))
