# Ubuntu CI Engine
# Copyright 2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import logging
import urllib2

from ci_utils.ticket_states import (
    TicketWorkflowStep,
    TicketWorkflowStepStatus,
)


class TicketApi(object):

    def __init__(self, url, ticket_id):
        self._url = url + '/api/v1'
        self._ticket_id = ticket_id

    def _patch(self, resource, data):
        req = urllib2.Request(resource)
        req.add_header('Content-Type', 'application/json')
        req.get_method = lambda: 'PATCH'
        urllib2.urlopen(req, json.dumps(data))

    def _post(self, url, data):
        req = urllib2.Request(url)
        req.add_header('Content-Type', 'application/json')
        resp = urllib2.urlopen(req, json.dumps(data))
        return resp.headers['location']

    def _get(self, url):
        resp = urllib2.urlopen(url)
        resp = resp.read()
        return json.loads(resp)

    def fail_ticket(self):
        url = self._url + '/updateticketstatus/' + self._ticket_id + '/'
        self._patch(url, {'status': TicketWorkflowStep.FAILED.value})

    def _patch_pkg_status(self, status):
        step = TicketWorkflowStep.PKG_BUILDING.value
        status = status.value
        url = self._url + '/updateticketstatus/' + self._ticket_id + '/'
        self._patch(url, {'status': status, 'current_workflow_step': step})

    def set_pkg_waiting(self):
        self._patch_pkg_status(TicketWorkflowStepStatus.PKG_BUILDING_WAITING)

    def set_pkg_inprogress(self):
        self._patch_pkg_status(
            TicketWorkflowStepStatus.PKG_BUILDING_INPROGRESS)

    def set_pkg_complete(self, success=True):
        state = TicketWorkflowStepStatus.PKG_BUILDING_COMPLETED
        if not success:
            state = TicketWorkflowStepStatus.PKG_BUILDING_FAILED
        self._patch_pkg_status(state)

    def _patch_subticket_status(self, subticket_id, step, status):
        url = self._url + '/updatesubticketstatus/' + str(subticket_id) + '/'
        self._patch(url, {'status': status, 'current_workflow_step': step})

    def set_subticket_progress(self, data):
        '''Processes the content of messages with subticket_status.

        Not all messages will have this content.'''
        for subticket in data.get('subticket_status', []):
            self._patch_subticket_status(subticket['id'], subticket['step'],
                                         subticket['status'])
            self.process_artifacts(subticket, subticket['id'])

    def _patch_ticket_status(self, status):
        step = TicketWorkflowStep.IMAGE_BUILDING.value
        status = status.value
        url = self._url + '/updateticketstatus/' + self._ticket_id + '/'
        self._patch(url, {'status': status, 'current_workflow_step': step})

    def set_image_waiting(self):
        self._patch_ticket_status(
            TicketWorkflowStepStatus.IMAGE_BUILDING_WAITING)

    def set_image_inprogress(self):
        self._patch_ticket_status(
            TicketWorkflowStepStatus.IMAGE_BUILDING_INPROGRESS)

    def set_image_complete(self, success=True):
        state = TicketWorkflowStepStatus.IMAGE_BUILDING_COMPLETED
        if not success:
            state = TicketWorkflowStepStatus.IMAGE_BUILDING_FAILED
        self._patch_ticket_status(state)

    def _patch_testing_status(self, status):
        step = TicketWorkflowStep.IMAGE_TESTING.value
        status = status.value
        url = self._url + '/updateticketstatus/' + self._ticket_id + '/'
        self._patch(url, {'status': status, 'current_workflow_step': step})

    def set_testing_waiting(self):
        self._patch_testing_status(
            TicketWorkflowStepStatus.IMAGE_TESTING_WAITING)

    def set_testing_inprogress(self):
        self._patch_testing_status(
            TicketWorkflowStepStatus.IMAGE_TESTING_INPROGRESS)

    def set_testing_complete(self, success=True):
        state = TicketWorkflowStepStatus.IMAGE_TESTING_PASSED
        if not success:
            state = TicketWorkflowStepStatus.IMAGE_TESTING_FAILED
        self._patch_testing_status(state)

    def _patch_publishing_status(self, status):
        step = TicketWorkflowStep.PKG_PUBLISHING.value
        if status == TicketWorkflowStepStatus.PKG_PUBLISHING_COMPLETED:
            logging.info('publishing is final step, marking ticket completed')
            step = TicketWorkflowStep.COMPLETED.value
            status = TicketWorkflowStepStatus.COMPLETED
        status = status.value
        url = self._url + '/updateticketstatus/' + self._ticket_id + '/'
        self._patch(url, {'status': status, 'current_workflow_step': step})

    def set_publishing_waiting(self):
        self._patch_publishing_status(
            TicketWorkflowStepStatus.PKG_PUBLISHING_WAITING)

    def set_publishing_inprogress(self):
        self._patch_publishing_status(
            TicketWorkflowStepStatus.PKG_PUBLISHING_INPROGRESS)

    def set_publishing_complete(self, success=True):
        state = TicketWorkflowStepStatus.PKG_PUBLISHING_COMPLETED
        if not success:
            state = TicketWorkflowStepStatus.PKG_PUBLISHING_FAILED
        self._patch_publishing_status(state)

    def add_artifact(self, name, artifact_url, artifact_type='LOGS',
                     subticket_id=None):
        '''Adds an artifact to the ticket or subticket if defined.

        It is the caller's responsibility to not create duplicates.'''
        data = {
            'type': artifact_type,
            'name': name,
            'reference': artifact_url,
        }
        if subticket_id:
            url = self._url + '/subticketartifact/'
            data['subticket'] = '/api/v1/subticket/' + str(subticket_id) + '/'
        else:
            url = self._url + '/ticketartifact/'
            data['ticket'] = '/api/v1/ticket/' + self._ticket_id + '/'
        try:
            return self._post(url, data)
        except:
            logging.exception('Unable to add %s to ticket-system' % name)

    def get_artifact_list(self, subticket_id=None):
        '''Fetches the list of artifact names for the ticket or subticket.'''
        full_ticket = self.get_full_ticket()
        current_artifacts = None
        if subticket_id:
            for subticket in full_ticket.get('subticket', []):
                if str(subticket['id']) == str(subticket_id):
                    current_artifacts = subticket['artifact']
                    break
            if not current_artifacts:
                raise ValueError('subticket id not found: '
                                 '{}'.format(subticket_id))
        else:
            current_artifacts = full_ticket['artifact']
        return [artifact['name'] for artifact in current_artifacts]

    def process_artifacts(self, data, subticket_id=None):
        '''Process all of the artifacts enclosed in the data dictionary.

        This will register the artifact to the ticket or subticket (if
        defined).  Not all data input may have artifacts.
        Duplicate artifacts names are not registered, this allows a progress
        update routine to continue to send back the same artifact data over
        multiple updates. Attempts to create duplicate artifacts should be
        caught when the artifact is placed in the data store by the creator.'''
        current_artifacts = self.get_artifact_list(subticket_id)
        artifact_list = data.get('artifacts', None)
        if artifact_list:
            for artifact in artifact_list:
                if artifact['name'] not in current_artifacts:
                    self.add_artifact(artifact['name'], artifact['reference'],
                                      artifact['type'], subticket_id)

    def get_full_ticket(self):
        url = self._url + '/fullticket/' + self._ticket_id + '/'
        data = self._get(url)
        return data

    def get_binaries(self):
        url = self._url + '/binarypackage/'
        data = self._get(url)
        return data['objects']
