#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import logging
import json
import os
import re
import sys
import time
import urllib2

from ci_utils import jenkins_utils
from ci_utils import dump_stack

SERVICE_PATH = os.path.join(os.path.dirname(__file__), '../../juju-relations')

# set up a root logger for everything
stdout_handler = logging.StreamHandler(stream=sys.stdout)
formatter = logging.Formatter(
    '%(asctime)s %(levelname)-5s %(name)s: %(message)s', datefmt='%H:%M:%S')
stdout_handler.setFormatter(formatter)
l = logging.getLogger('')
l.addHandler(stdout_handler)
l.setLevel(logging.DEBUG)

logger = logging.getLogger('lander_get_next_ticket')


def _get(url):
    resp = urllib2.urlopen(url)
    resp = resp.read()
    return json.loads(resp)


def _get_parser():
    parser = argparse.ArgumentParser(
        description='Wraps the REST API calls with a progress queue monitor.')
    parser.add_argument('--service-name',
                        required=True,
                        help='The name of the service to connect.')
    parser.add_argument('--service-port',
                        required=True,
                        help='The port number of the service to connect.')
    parser.add_argument('--delay',
                        required=True,
                        type=int,
                        help='The delay in seconds between poll attempts.')
    return parser


def handle_next_ticket(jenkins_config, service_url):
    if jenkins_utils.is_master_job_running(jenkins_config):
        logging.info('{} is building, skip next ticket '
                     'request.'.format(jenkins_config['lander_master_job']))
    else:
        logging.info('Jenkins is available')
        url = '{}/api/v1/next/'.format(service_url)
        data = _get(url)
        return data


def poll(jenkins_config, service_url):
    rc = 1
    try:
        results = handle_next_ticket(jenkins_config, service_url)
        if results:
            out_data = results.get('objects', None)
            if out_data:
                out_data = out_data[0]
                request_id = out_data['id']
                logging.info('request_id: {}'.format(request_id))
                logging.info('request_parameters: {}'.format(out_data))
                job_parameters = {
                    'request_id': request_id,
                }
                jenkins_utils.trigger_build(jenkins_config, job_parameters)
        rc = 0
    except urllib2.HTTPError as e:
        body = e.read()
        logging.exception('Unable to handle request: %s: %s', e, body)
        results = {
            'result': 'ERROR',
            'resp_code': e.reason,
            'resp_body': body,
        }
    except Exception as e:
        logging.exception('Unexpected error: {}'.format(e))
        results = {
            'result': 'FAILED',
            'error': e.message,
        }

    return rc


def get_service_url(service_name, service_port):
    for root, dirs, files in os.walk(SERVICE_PATH):
        for name in files:
            with open(os.path.join(root, name)) as f:
                for line in f.readlines():
                    if re.match('%s:' % service_name, line):
                        (junk, url) = line.split(':')
                        return 'http://{}:{}'.format(url.strip(),
                                                     service_port)


def main(args):
    logging.info('Starting lander-jenkins worker.')
    while True:
        logging.info('Trying to acquire jenkins config.')
        jenkins_config = jenkins_utils.get_config()
        if not jenkins_config:
            logging.error('Unable to acquire jenkins configuration, retrying')
            time.sleep(args.delay)
        else:
            break

    while True:
        logging.info('Trying to acquire ticket service url.')
        service_url = get_service_url(args.service_name, args.service_port)
        if not service_url:
            logging.error('Unable to acquire ticket service url, retrying')
            time.sleep(args.delay)
        else:
            break

    while True:
        logging.info('Polling for a new ticket')
        ret = poll(jenkins_config, service_url)
        if ret:
            logging.error('Breaking out of polling loop')
            return ret
        time.sleep(args.delay)


if __name__ == '__main__':
    dump_stack.install_stack_dump_signal()
    args = _get_parser().parse_args()
    exit(main(args))
