# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import mock
import os
import sys
import unittest
import urllib2

sys.path.append(os.path.join(os.path.dirname(__file__), '../../bin'))
import lander_archiver
sys.path.append(os.path.join(os.path.dirname(__file__), '../../ci-utils'))
from ci_utils import data_store


class TestArchiver(unittest.TestCase):
    '''Tests for the lander_archiver.py script.'''

    def setUp(self):
        super(TestArchiver, self).setUp()
        self.archiver = lander_archiver.LanderArchiver()

    def test_get_datastore_config(self):
        '''Verify creation of a datastore configuration.'''
        config = {'username', 'user',
                  'password', 'pass'}
        self.archiver.get_datastore_config_contents = lambda x: config
        self.assertEqual(config,
                         self.archiver.get_datastore_config('filename'))

    def test_get_datastore_config_exception(self):
        '''Verify an error is logged on failure to load a datastore config.'''
        get_mock = mock.Mock()
        get_mock.side_effect = IOError
        self.archiver.get_datastore_config_contents = get_mock
        log_mock = mock.Mock()
        self.archiver.log = mock.Mock()
        self.archiver.log.error = log_mock
        self.assertIsNone(self.archiver.get_datastore_config('filename'))
        log_mock.assert_called_with('Failed to load datastore configuration: ')

    @mock.patch('lander_archiver.data_store.DataStore')
    def test_get_datastore(self, data_store_mock):
        '''Verify a data store is returned assuming mocked sources.'''
        datastore = 'Mocked datastore'
        request_id = 'request-1'
        data_store_mock.return_value = datastore
        self.assertEqual(datastore,
                         self.archiver.get_datastore(request_id, {}))

    @mock.patch('lander_archiver.data_store.DataStore')
    def test_get_datastore_exception(self, data_store_mock):
        '''Verify a data store is returned assuming mocked sources.'''
        request_id = 'request-1'
        data_store_mock.side_effect = data_store.DataStoreException
        log_mock = mock.Mock()
        self.archiver.log = mock.Mock()
        self.archiver.log.error = log_mock
        self.assertIsNone(self.archiver.get_datastore(request_id, {}))
        log_mock.assert_called_with('Failed to create datastore: ')

    @mock.patch('lander_archiver.jenkins_utils.get_url_contents')
    def test_get_artifact_list(self, get_url_mock):
        '''Verify an artifact list is returned from jenkins json data.'''
        artifact_list = [
            'http://jenkins:8080/path/to/artifact.1',
            'http://jenkins:8080/path/to/artifact.2']
        jenkins_json = {'artifacts':
                        [{'relativePath': x} for x in artifact_list]}
        get_url_mock.return_value = json.dumps(jenkins_json)
        self.assertEqual(artifact_list,
                         self.archiver.get_artifact_list('url'))

    @mock.patch('lander_archiver.jenkins_utils.get_url_contents')
    def test_get_artifact_list_none(self, get_url_mock):
        '''Verify that a job with no artifacts returns an empty list.'''
        artifact_list = []
        jenkins_json = {'artifacts': artifact_list}
        get_url_mock.return_value = json.dumps(jenkins_json)
        self.assertEqual(artifact_list,
                         self.archiver.get_artifact_list('url'))

    @mock.patch('lander_archiver.jenkins_utils.get_url_contents')
    def test_store_artifact(self, get_url_mock):
        '''Verify that storing an artifact results in a location.'''
        location = 'http://data-store/bucket/file'
        datastore = mock.Mock()
        datastore.put_file = mock.Mock()
        datastore.put_file.return_value = location
        get_url_mock.return_value = 'data'
        self.assertEqual(
            location, self.archiver.store_artifact(datastore, 'name', 'url'))

    @mock.patch('lander_archiver.jenkins_utils.get_url_contents')
    def test_store_artifact_error(self, get_url_mock):
        '''Verify that a datastore error causes a log message.'''
        datastore = mock.Mock()
        datastore.put_file = mock.Mock()
        datastore.put_file.side_effect = data_store.DataStoreException
        log_mock = mock.Mock()
        self.archiver.log = mock.Mock()
        self.archiver.log.error = log_mock
        get_url_mock.return_value = 'data'
        self.assertIsNone(
            self.archiver.store_artifact(datastore, 'name', 'url'))
        log_mock.assert_called_with('Failed to archive artifact [name]: ')

    @mock.patch('lander_archiver.jenkins_utils.get_url_contents')
    def test_archive_console_log(self, get_url_mock):
        '''Verify archiving of the console log.'''
        location = 'http://data-store/bucket/file'

        self.archiver.ticket_api = mock.Mock()

        datastore = mock.Mock()
        datastore.put_file = mock.Mock()
        datastore.put_file.return_value = location
        log_mock = mock.Mock()
        self.archiver.log = mock.Mock()
        self.archiver.log.info = log_mock
        get_url_mock.return_value = 'console data'
        self.archiver.archive_console_log(datastore, 'master',
                                          'http://jenkins:8080/job/master/')
        log_mock.assert_called_with('Console log: {}'.format(location))
        self.assertEqual(1, self.archiver.ticket_api.add_artifact.call_count)

    @mock.patch('urllib2.Request')
    def test_archive_console_log_error(self, request_mock):
        '''Verify a console log load error raises an exception.'''
        request_mock.side_effect = IOError
        location = 'http://data-store/bucket/file'
        datastore = mock.Mock()
        datastore.put_file = mock.Mock()
        datastore.put_file.return_value = location
        log_mock = mock.Mock()
        self.archiver.log = mock.Mock()
        self.archiver.log.info = log_mock
        with self.assertRaises(EnvironmentError):
            self.archiver.archive_console_log(
                datastore, 'master', 'http://jenkins:8080/job/master/')
        self.assertEqual(0, log_mock.call_count)

    @mock.patch('lander_archiver.jenkins_utils.get_url_contents')
    def test_archive_artifacts(self, get_url_mock):
        '''Verify archiving of an artifact.'''
        location = 'http://data-store/bucket/file'
        self.archiver.ticket_api = mock.Mock()
        datastore = mock.Mock()
        datastore.put_file = mock.Mock()
        datastore.put_file.return_value = location
        log_mock = mock.Mock()
        self.archiver.log = mock.Mock()
        self.archiver.log.info = log_mock
        artifact_list = [
            'http://jenkins:8080/path/to/artifact.1']
        jenkins_json = {'artifacts':
                        [{'relativePath': x} for x in artifact_list]}
        get_url_mock.return_value = json.dumps(jenkins_json)
        self.archiver.archive_artifacts(datastore, 'master',
                                        'http://jenkins:8080/job/master/')
        for artifact in artifact_list:
            log_mock.assert_called_with('Artifact [{}]: '
                                        '{}'.format(artifact, location))
        self.archiver.ticket_api.add_artifact.assert_called_once_with(
            'master.http:__jenkins:8080_path_to_artifact.1', location)

    #@mock.patch('lander_archiver.urllib2.Request')
    @mock.patch('urllib2.Request')
    def test_archive_artifacts_error(self, request_mock):
        '''Verify an artifact load error raises an exception.'''
        request_mock.side_effect = IOError
        location = 'http://data-store/bucket/file'
        datastore = mock.Mock()
        datastore.put_file = mock.Mock()
        datastore.put_file.return_value = location
        log_mock = mock.Mock()
        self.archiver.log = mock.Mock()
        self.archiver.log.info = log_mock
        with self.assertRaises(EnvironmentError):
            self.archiver.archive_artifacts(datastore, 'master',
                                            'http://jenkins:8080/job/master/')
        self.assertEqual(0, log_mock.call_count)

    @mock.patch('lander_archiver.jenkins_utils.get_url_contents')
    @mock.patch('lander_archiver.TicketApi')
    def test_call(self, ticket_api, get_url_mock):
        '''Verify __call__ returns 0 when everything works.'''
        config = {'username', 'user',
                  'password', 'pass'}
        location = 'http://jenkins:8080/path/to/artifact'
        artifact_list = [
            'http://jenkins:8080/path/to/artifact.1',
            'http://jenkins:8080/path/to/artifact.2']
        jenkins_json = {'artifacts':
                        [{'relativePath': x} for x in artifact_list]}

        api_inst = mock.Mock()
        ticket_api.return_value = api_inst

        args = mock.Mock()
        args.datastore_config = None
        args.request_id = '1'
        args.job_name = 'lander_master'
        args.source_url = 'http://jenkins:8080/job/lander_master/1/'
        self.archiver.parse_arguments = lambda: args
        self.archiver.get_datastore_config = lambda x: config
        self.archiver.get_datastore = lambda x, y: 'datastore'
        get_url_mock.return_value = json.dumps(jenkins_json)
        self.archiver.store_artifact = mock.Mock(return_value=location)
        log_mock = mock.Mock()
        self.archiver.log = mock.Mock()
        self.archiver.log.info = log_mock
        self.assertEqual(0, self.archiver())
        # 3 log files should have been added
        self.assertEqual(3, api_inst.add_artifact.call_count)

    def test_call_datastore_config_error(self):
        '''Verify __call__ returns -1 on a datastore config loading error.'''
        args = mock.Mock()
        self.archiver.parse_arguments = lambda: args
        self.archiver.get_datastore_config = lambda x: None
        self.assertEqual(-1, self.archiver())

    def test_call_invalid_datastore(self):
        '''Verify __call__ returns -2 on a datastore creation error.'''
        config = {'username', 'user',
                  'password', 'pass'}
        args = mock.Mock()
        args.datastore_config = None
        self.archiver.parse_arguments = lambda: args
        self.archiver.get_datastore_config = lambda x: config
        self.archiver.get_datastore = lambda x, y: None
        self.assertEqual(-2, self.archiver())

    @mock.patch('lander_archiver.TicketApi')
    @mock.patch('urllib2.urlopen')
    def test_call_store_artifact_failure(self, mock_urlopen, ticket_api):
        '''Verify __call__ returns -3 on a store artifact failure.'''
        config = {'username', 'user',
                  'password', 'pass'}

        def side_effect(*args):
            # Don't hang around waiting for the Jenkins server to never reply.
            raise urllib2.URLError('Nope.')
        mock_urlopen.side_effect = side_effect
        api_inst = mock.Mock()
        ticket_api.return_value = api_inst
        args = mock.Mock()
        args.datastore_config = None
        args.request_id = '1'
        args.job_name = 'lander_master'
        args.source_url = 'http://jenkins:8080/job/lander_master/1/'
        self.archiver.parse_arguments = lambda: args
        self.archiver.get_datastore_config = lambda x: config
        datastore = mock.Mock()
        self.archiver.get_datastore = mock.Mock(return_value=datastore)
        log_mock = mock.Mock()
        self.archiver.log = mock.Mock()
        self.archiver.log.error = log_mock
        self.assertEqual(-3, self.archiver())
        self.assertEqual(0, api_inst.add_artifact.call_count)
