# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import os
import StringIO
import sys
import tempfile
import unittest
import urllib2

import mock

sys.path.append(os.path.join(os.path.dirname(__file__), '../../bin'))
import lander_service_wrapper


class TestWrapper(unittest.TestCase):
    '''Sanity tests of the lander_service_wrapper.py script.'''

    def setUp(self):
        super(TestWrapper, self).setUp()
        patch = mock.patch('ci_utils.amqp_utils.get_config')
        patch.start()
        self.addCleanup(patch.stop)

        self.results = tempfile.mkstemp(prefix='lander-test-')[1]
        self.addCleanup(os.unlink, self.results)

        self.params = tempfile.mkstemp(prefix='lander-test-')[1]
        self.addCleanup(os.unlink, self.params)

        if 'BUILD_URL' not in os.environ:
            # required by lander_service_wrapper, but ignore in tests
            os.environ['BUILD_URL'] = 'foo'

        data = {
            'master': {
                'request_id': '1',
                'request_parameters': {
                    'subticket': [{
                        'artifact': [{
                            'type': 'SPU',
                            'reference': 'filename'}]}]},
                'ppa_assigner_url': 'foo',
                'bsb_url': 'bar',
                'imgbuilder_url': 'foo',
                'ts_url': 'bar',
                'tr_url': 'bar',
                'progress_trigger': 'progress-trigger',
                'base_image': {},
                'package_list': {},
            },
            'ppa_assigner': {
                'ppa': 'ppa-name',
                'location': 'ppa-location',
            },
            'bsbuilder': {
                'ppa': 'ppa-name',
            },
            'image_builder': {
                'image_id': 'id',
            }
        }
        with open(self.params, 'w') as f:
            f.write(json.dumps(data))

        self.base_args = [
            '--input-file', self.params, '--output-file', self.results]

    def testBadService(self):
        '''make sure we exit gracefully for a bad service name'''
        args = self.base_args + ['--service', 'bad-name']
        with self.assertRaises(SystemExit):
            lander_service_wrapper._get_parser().parse_args(args)

    @mock.patch('lander_service_wrapper.TicketApi')
    def testGetBinaryPackagesAddedOnly(self, api):
        '''Verify that packages from the current ticket are added.'''
        packages = ['add_package1']
        parameters = {
            'added_binaries': ','.join(packages),
            'removed_binaries': None}
        api.get_binaries = mock.Mock(return_value=[])
        actual = lander_service_wrapper._get_binary_packages(api, parameters)
        self.assertItemsEqual(packages, actual)

    @mock.patch('lander_service_wrapper.TicketApi')
    def testGetBinaryPackagesNone(self, api):
        '''Verify that no packages are added when none are specified.'''
        parameters = {
            'added_binaries': None,
            'removed_binaries': None}
        api.get_binaries = mock.Mock(return_value=[])
        actual = lander_service_wrapper._get_binary_packages(api, parameters)
        self.assertEqual([], actual)

    @mock.patch('lander_service_wrapper.TicketApi')
    def testGetBinaryPackagesRemovedOnly(self, api):
        '''Verify that removing packages that aren't there is ok.'''
        packages = ['remove_package1', 'remove_package2']
        parameters = {
            'added_binaries': None,
            'removed_binaries': ','.join(packages)}
        api.get_binaries = mock.Mock(return_value=[])
        actual = lander_service_wrapper._get_binary_packages(api, parameters)
        self.assertEqual([], actual)

    @mock.patch('lander_service_wrapper.TicketApi')
    def testGetBinaryPackagesGlobalOnly(self, api):
        '''Verify that packages only from the global list are added.'''
        golden = [
            {'name': 'golden_package1'},
            {'name': 'golden_package2'},
        ]
        parameters = {
            'added_binaries': None,
            'removed_binaries': None}
        api.get_binaries = mock.Mock(return_value=golden)
        actual = lander_service_wrapper._get_binary_packages(api, parameters)
        self.assertItemsEqual([x['name'] for x in golden], actual)

    @mock.patch('lander_service_wrapper.TicketApi')
    def testGetBinaryPackagesAddedAndGlobal(self, api):
        '''Verify that packages from the ticket and global list are added.'''
        packages = ['add_package1', 'add_package2']
        golden = [
            {'name': 'golden_package1'},
            {'name': 'golden_package2'},
        ]
        parameters = {
            'added_binaries': ','.join(packages),
            'removed_binaries': None}
        api.get_binaries = mock.Mock(return_value=golden)
        actual = lander_service_wrapper._get_binary_packages(api, parameters)
        expected = [x['name'] for x in golden]
        expected.extend(packages)
        self.assertItemsEqual(expected, actual)

    @mock.patch('lander_service_wrapper.TicketApi')
    def testGetBinaryPackagesAddedRemovedAndGlobal(self, api):
        '''Verify that packages from all three lists are processed.'''
        add_packages = ['add_package1', 'add_package2']
        remove_packages = ['add_package1', 'golden_package2']
        golden = [
            {'name': 'golden_package1'},
            {'name': 'golden_package2'},
        ]
        parameters = {
            'added_binaries': ','.join(add_packages),
            'removed_binaries': ','.join(remove_packages)}
        api.get_binaries = mock.Mock(return_value=golden)
        actual = lander_service_wrapper._get_binary_packages(api, parameters)
        expected = ['golden_package1', 'add_package2']
        self.assertItemsEqual(expected, actual)

    @mock.patch('lander_service_wrapper._post')
    @mock.patch('lander_service_wrapper._get')
    @mock.patch('time.sleep')
    def testPPAAssigner404(self, sleep, get, post):
        '''Ensure we do proper retrying when no ppa's are available'''
        ppa = 'ppa:foo'
        get.return_value = {'name': ppa}
        err = urllib2.HTTPError(
            'url', 404, '404', None, StringIO.StringIO('foo bar'))
        post.side_effect = [err, err, err, 'http://foo/api/v1/ppa/1']

        args = self.base_args + ['--service', 'ppa_assigner']
        args = lander_service_wrapper._get_parser().parse_args(args)
        rc = lander_service_wrapper.main(args)
        self.assertEqual(0, rc)

        with open(self.results) as f:
            data = json.loads(f.read())
            self.assertEqual('PASSED', data['result'])
            self.assertEqual(ppa, data['ppa'])

    @mock.patch('lander_service_wrapper._post')
    def testPPAAssignerUnexpected(self, post):
        '''Ensure unexpected errors still generate a response'''
        err_msg = 'foo bar'
        post.side_effect = RuntimeError(err_msg)

        args = self.base_args + ['--service', 'ppa_assigner']
        args = lander_service_wrapper._get_parser().parse_args(args)
        rc = lander_service_wrapper.main(args)
        self.assertEqual(1, rc)

        with open(self.results) as f:
            data = json.loads(f.read())
            self.assertEqual('FAILED', data['result'])
            self.assertEqual(err_msg, data['error'])

    @mock.patch('lander_service_wrapper._post')
    @mock.patch('lander_service_wrapper._get')
    def testPPAAssignerSucceeds(self, get, post):
        '''Ensure we generate the proper response if things succeed.'''
        ppa = 'ppa:foo'
        post.return_value = 'http://foo/api/v1/ppa/1'
        get.return_value = {'name': ppa}

        args = self.base_args + ['--service', 'ppa_assigner']
        args = lander_service_wrapper._get_parser().parse_args(args)
        rc = lander_service_wrapper.main(args)
        self.assertEqual(0, rc)

        with open(self.results) as f:
            data = json.loads(f.read())
            self.assertEqual('PASSED', data['result'])
            self.assertEqual(ppa, data['ppa'])

    @mock.patch('lander_service_wrapper.TicketApi')
    @mock.patch('lander_service_wrapper._post')
    def testBSBuilderUnexpected(self, post, api):
        '''Ensure unexpected errors still generate a response'''
        err_msg = 'foo bar'
        post.side_effect = RuntimeError(err_msg)

        args = self.base_args + ['--service', 'bsbuilder']
        args = lander_service_wrapper._get_parser().parse_args(args)
        rc = lander_service_wrapper.main(args)
        self.assertEqual(1, rc)
        api().set_pkg_waiting.assert_called_once()
        api().set_pkg_complete.assert_called_once_with(False)

        with open(self.results) as f:
            data = json.loads(f.read())
            self.assertEqual('FAILED', data['result'])
            self.assertEqual(err_msg, data['error'])

    @mock.patch('lander_service_wrapper.TicketApi')
    @mock.patch('lander_service_wrapper._post')
    @mock.patch('ci_utils.amqp_utils.connection')
    def testBSBuilderSucceeds(self, connection, post, api):
        '''Ensure we generate the proper response if things succeed.'''

        # mock the run-forever logic to ensure we return what we expect
        state = {'state': 'COMPLETED', 'result': 'PASSED',
                 'foo': 'bar', 'exit': True}

        def amqp():
            def run_forever(channel, queue, callback):
                m = mock.Mock()
                m.body = json.dumps({'state': 'WAITING'})
                callback(m)
                m.body = json.dumps(state)
                callback(m)
            return run_forever

        args = self.base_args + ['--service', 'bsbuilder']
        args = lander_service_wrapper._get_parser().parse_args(args)

        with mock.patch('ci_utils.amqp_utils._run_forever', new_callable=amqp):
            rc = lander_service_wrapper.main(args)
            self.assertEqual(0, rc)

        api().set_pkg_waiting.assert_called_once()
        api().set_pkg_complete.assert_called_once_with(True)

        with open(self.results) as f:
            data = json.loads(f.read())
            self.assertEqual(state, data)

    @mock.patch('lander_service_wrapper.TicketApi')
    @mock.patch('lander_service_wrapper._post')
    def testImageBuilderUnexpected(self, post, api):
        '''Ensure unexpected errors still generate a response'''
        err_msg = 'foo bar'
        post.side_effect = RuntimeError(err_msg)

        args = self.base_args + ['--service', 'image_builder']
        args = lander_service_wrapper._get_parser().parse_args(args)
        rc = lander_service_wrapper.main(args)
        self.assertEqual(1, rc)
        api().set_image_waiting.assert_called_once()
        api().set_image_complete.assert_called_once_with(False)

        with open(self.results) as f:
            data = json.loads(f.read())
            self.assertEqual('FAILED', data['result'])
            self.assertEqual(err_msg, data['error'])

    @mock.patch('lander_service_wrapper.TicketApi')
    @mock.patch('lander_service_wrapper._post')
    @mock.patch('ci_utils.amqp_utils.connection')
    def testImageBuilderSucceeds(self, connection, post, api):
        '''Ensure we generate the proper response if things succeed.

        While technically "succeeding" its checking for a successful FAILURE
        '''

        # mock the run-forever logic to ensure we return what we expect
        state = {'state': 'COMPLETED', 'result': 'FAILED',
                 'foo': 'bar', 'exit': True}

        def amqp():
            def run_forever(channel, queue, callback):
                m = mock.Mock()
                m.body = json.dumps({'state': 'WAITING'})
                callback(m)
                m.body = json.dumps(state)
                callback(m)
            return run_forever

        args = self.base_args + ['--service', 'image_builder']
        args = lander_service_wrapper._get_parser().parse_args(args)

        with mock.patch('ci_utils.amqp_utils._run_forever', new_callable=amqp):
            rc = lander_service_wrapper.main(args)
            self.assertEqual(1, rc)

        api().set_image_waiting.assert_called_once()
        api().set_image_complete.assert_called_once_with(False)

        with open(self.results) as f:
            data = json.loads(f.read())
            self.assertEqual(state, data)

    @mock.patch('lander_service_wrapper.TicketApi')
    @mock.patch('lander_service_wrapper._post')
    @mock.patch('ci_utils.amqp_utils.connection')
    def testRunnerSucceeds(self, connection, post, api):
        '''Ensure we generate the proper response if things succeed.'''

        # mock the run-forever logic to ensure we return what we expect
        state = {'state': 'COMPLETED', 'result': 'PASSED',
                 'foo': 'bar', 'exit': True}

        def amqp():
            def run_forever(channel, queue, callback):
                m = mock.Mock()
                m.body = json.dumps({'state': 'WAITING'})
                callback(m)
                m.body = json.dumps(state)
                callback(m)
            return run_forever

        args = self.base_args + ['--service', 'test_runner']
        args = lander_service_wrapper._get_parser().parse_args(args)

        with mock.patch('ci_utils.amqp_utils._run_forever', new_callable=amqp):
            rc = lander_service_wrapper.main(args)
            self.assertEqual(0, rc)

        api().set_testing_waiting.assert_called_once()
        api().set_testing_complete.assert_called_once_with(True)

        with open(self.results) as f:
            data = json.loads(f.read())
            self.assertEqual(state, data)

    @mock.patch('lander_service_wrapper.TicketApi')
    @mock.patch('lander_service_wrapper._post')
    def testRunnerUnexpected(self, post, api):
        '''Ensure unexpected errors still generate a response'''
        err_msg = 'foo bar'
        post.side_effect = RuntimeError(err_msg)

        args = self.base_args + ['--service', 'test_runner']
        args = lander_service_wrapper._get_parser().parse_args(args)
        rc = lander_service_wrapper.main(args)
        self.assertEqual(1, rc)
        api().set_testing_waiting.assert_called_once()
        api().set_testing_complete.assert_called_once_with(False)

        with open(self.results) as f:
            data = json.loads(f.read())
            self.assertEqual('FAILED', data['result'])
            self.assertEqual(err_msg, data['error'])

    @mock.patch('lander_service_wrapper._patch')
    def testPPAFreeSucceeds(self, patch):
        '''ensure we can free a PPA'''
        args = self.base_args + ['--service', 'ppa_assigner_release']
        args = lander_service_wrapper._get_parser().parse_args(args)
        rc = lander_service_wrapper.main(args)
        self.assertEqual(0, rc)
        patch.assert_called_once()
