# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import mock
import os
import sys
import unittest

sys.path.append(os.path.join(os.path.dirname(__file__), '../../bin'))
import ticket_api
sys.path.append(os.path.join(os.path.dirname(__file__), '../../ci-utils'))


class TestTicketApi(unittest.TestCase):
    '''Tests for the ticket_api.py module.'''

    ticket_data = {
        'artifact': [{
            'id': 123,
            'name': 'artifact_name',
            'reference': 'artifact_reference',
            'resource_uri': '/api/v1/fullticketartifact/123',
            'type': 'LOGS',
        }],
        'subticket': [{
            'id': 12,
            'artifact': [{
                'id': 321,
                'name': 'subticket_artifact_name',
                'reference': 'subticket_artifact_reference',
                'resource_uri': '/api/v1/fullticketartifact/12',
                'type': 'LOGS'
            }],
        }]
    }

    def setUp(self):
        super(TestTicketApi, self).setUp()
        self.ticket = ticket_api.TicketApi('http://ts-url', '123')

    def test_add_artifact_default(self):
        self.ticket._post = mock.Mock()
        post_url = 'http://ts-url/api/v1/ticketartifact/'
        self.ticket.add_artifact('name', 'url')
        data = {
            'type': 'LOGS',
            'name': 'name',
            'reference': 'url',
            'ticket': '/api/v1/ticket/123/'
        }
        self.ticket._post.assert_called_with(post_url, data)

    def test_add_artifact_subticket(self):
        self.ticket._post = mock.Mock()
        post_url = 'http://ts-url/api/v1/subticketartifact/'
        self.ticket.add_artifact('name', 'url', artifact_type='RESULT',
                                 subticket_id=12)
        data = {
            'type': 'RESULT',
            'name': 'name',
            'reference': 'url',
            'subticket': '/api/v1/subticket/12/'
        }
        self.ticket._post.assert_called_with(post_url, data)

    def test_get_artifact_list_from_ticket(self):
        expected = ['artifact_name']
        self.ticket.get_full_ticket = mock.Mock(return_value=self.ticket_data)
        actual = self.ticket.get_artifact_list()
        self.assertItemsEqual(expected, actual)

    def test_get_artifact_list_from_subticket(self):
        expected = ['subticket_artifact_name']
        self.ticket.get_full_ticket = mock.Mock(return_value=self.ticket_data)
        actual = self.ticket.get_artifact_list(subticket_id=12)
        self.assertItemsEqual(expected, actual)

    def test_get_artifact_list_from_subticket_assert(self):
        self.ticket.get_full_ticket = mock.Mock(return_value=self.ticket_data)
        with self.assertRaises(ValueError):
            self.ticket.get_artifact_list(subticket_id=10)

    def test_process_artifacts(self):
        self.ticket.get_full_ticket = mock.Mock(return_value=self.ticket_data)
        self.ticket.add_artifact = mock.Mock()
        data = {
            'artifacts': [{
                'name': 'new_artifact_name',
                'reference': 'new_artifact_reference',
                'type': 'LOGS'}],
        }
        self.ticket.process_artifacts(data)
        self.ticket.add_artifact.assert_called_with(
            'new_artifact_name',
            'new_artifact_reference',
            'LOGS',
            None)

    def test_process_artifacts_subticket(self):
        self.ticket.get_full_ticket = mock.Mock(return_value=self.ticket_data)
        self.ticket.add_artifact = mock.Mock()
        data = {
            'artifacts': [{
                'name': 'new_artifact_name',
                'reference': 'new_artifact_reference',
                'type': 'LOGS'}],
        }
        self.ticket.process_artifacts(data, subticket_id=12)
        self.ticket.add_artifact.assert_called_with(
            'new_artifact_name',
            'new_artifact_reference',
            'LOGS',
            12)

    def test_process_artifacts_none(self):
        self.ticket.get_full_ticket = mock.Mock(return_value=self.ticket_data)
        self.ticket.add_artifact = mock.Mock()
        data = {}
        self.ticket.process_artifacts(data)
        self.assertEqual(0, self.ticket.add_artifact.called)
