# Ubuntu CI Engine
# Copyright 2013, 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import logging
import time

from optparse import make_option

from django.conf import settings
from django.core.management.base import BaseCommand

from ppa_assigner import launchpad
from ppa_assigner.models import PPA


def _now():
    return datetime.datetime.utcnow()


class Command(BaseCommand):
    help = 'Runs forever to take dirty PPAs and get them ready for re-use.'
    option_list = BaseCommand.option_list + (
        make_option('--interval', dest='interval', default=30, type=int,
                    help='Interval to sleep between checks.'),
        make_option('--exit-on-clean', dest='exit_early', action='store_true',
                    help='''Exit as soon as all PPAs are clean. This
                            is useful for testing.''')
    )

    def handle(self, *args, **options):
        level = logging.INFO
        if int(options['verbosity']) > 1:
            level = logging.DEBUG
        logging.basicConfig(level=level)

        self.assert_oauth()

        interval = options['interval']
        try:
            while True:
                PPA.populate_from_launchpad()
                self.dirty_to_cleaning()
                self.cleaning_to_available()
                if options['exit_early']:
                    self.check_for_exit()
                time.sleep(interval)
        except KeyboardInterrupt:
            print('command cancelled by user')

    def assert_oauth(self):
        required = ['OAUTH_CONSUMER_KEY', 'OAUTH_TOKEN', 'OAUTH_TOKEN_SECRET']
        missing = []
        for x in required:
            if not getattr(settings, x, None):
                missing.append(x)
        if missing:
            print('ERROR: oauth settings are required by this command:')
            print('  ' + '\n  '.join(missing))
            exit(1)

    def check_for_exit(self):
        total = PPA.objects.all().count()
        avail = PPA.objects.filter(state=PPA.AVAILABLE).count()
        if total == avail:
            print('All PPAs are ready for use.')
            exit(0)

    def dirty_to_cleaning(self):
        logging.debug('Check for PPAs that need to be cleaned...')
        qs = PPA.objects.select_for_update().filter(state=PPA.DIRTY)
        if qs.count():
            logging.info('Requesting launchpad clean: %s', qs[0].name)
            try:
                launchpad.request_ppa_clean(qs[0].name)
                qs.filter(pk__in=(qs[0].pk,)).update(
                    state=PPA.CLEANING, timestamp=_now())
            except Exception:
                logging.exception('unable to request clean ppa, will retry')

    def cleaning_to_available(self):
        logging.debug('Check for PPAs that are available...')
        qs = PPA.objects.select_for_update().filter(state=PPA.CLEANING)
        if qs.count():
            if self._lp_is_clean(qs[0]):
                logging.info('ppa is available for use: %s. cleaning took %s',
                             qs[0].name, (_now() - qs[0].timestamp))
                qs.filter(pk__in=(qs[0].pk,)).update(state=PPA.AVAILABLE)

    def _lp_is_clean(self, ppa):
        logging.debug('Checking if ppa is ready: %s', ppa.name)
        try:
            for x in launchpad.get_publishing_history(ppa.name):
                if not x['date_made_pending']:
                    return False
        except Exception:
            logging.exception('error checking if ppa is clean')
            return False
        return True
