# Ubuntu CI Engine
# Copyright 2013, 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import logging
import re
import urllib2

from django.conf import settings
from django.db import models

from ppa_assigner import launchpad


class PPA(models.Model):
    AVAILABLE = 0
    RESERVED = 10
    DIRTY = 20
    CLEANING = 30
    STATES = (
        (AVAILABLE, 'available'),
        (RESERVED, 'reserved'),
        (DIRTY, 'dirty'),
        (CLEANING, 'cleaning'),
    )
    name = models.CharField(max_length=4096, unique=True)

    private = models.BooleanField(default=False)
    state = models.PositiveIntegerField(choices=STATES, default=AVAILABLE)
    ticket_id = models.PositiveIntegerField()
    timestamp = models.DateTimeField(auto_now_add=True)

    def __unicode__(self):
        val = '%s (%s)' % (self.name, self._get_state_str())
        if self.state == PPA.RESERVED:
            val += ' ticket=%s' % self.ticket_id
        return val

    def _get_state_str(self):
        for state, label in PPA.STATES:
            if self.state == state:
                return label
        raise RuntimeError('Unsupported state: %s' % self.state)

    @staticmethod
    def _merge_ppas(lp_ppas):
        # note: LP will throttle calls to list PPAs. So you can occassionally
        # add a PPA, and not see it update for the first few calls to this
        # method, but it will eventually get in sync
        lp_ppas_ = set(lp_ppas.keys())
        model_ppas = set([ppa.name for ppa in PPA.objects.all()])
        add = lp_ppas_ - model_ppas
        remove = model_ppas - lp_ppas_

        logging.info('adding new ppas: %r', add)
        for ppa_name in add:
            lp_ppa = lp_ppas[ppa_name]
            PPA.objects.create(name=ppa_name, ticket_id=0,
                               private=lp_ppa['private'])

        logging.info('removing ppas: %r', remove)
        for ppa_name in remove:
            ppa = PPA.objects.get(name=ppa_name)
            if ppa.state == PPA.RESERVED:
                # log something, but we may as well delete. because if the
                # ppa doesn't exist in LP we are already in a catastrophic
                # situation and shouldn't let any other user re-use this
                logging.error('ppa %s appears to be in use', ppa_name)
            ppa.delete()

    @staticmethod
    def populate_from_launchpad():
        '''Get list of PPAs from launchpad and ensure they exist here.'''
        url = settings.LAUNCHPAD_API_BASE + '/~%s/ppas'
        url = url % settings.LAUNCHPAD_PPA_OWNER
        logging.info('populating ppa model from: %s', url)
        try:
            ppas = {}
            pat = re.compile(settings.PPA_PATTERN)
            for ppa in launchpad.lp_collection(url):
                if pat.match(ppa['name']):
                    ppa_name = 'ppa:{}/{}'.format(settings.LAUNCHPAD_PPA_OWNER,
                                                  ppa['name'])
                    ppas[ppa_name] = {
                        'private': ppa['private'],
                    }
            PPA._merge_ppas(ppas)
        except urllib2.URLError:
            logging.exception('Could not populate PPAs from launchpad')
            raise

    @staticmethod
    def reserve(ticket_id):
        # "select_for_update" provides proper db-level locking so we
        # don't accidentally have 2 concurrent requests find and reserve the
        # same PPA
        qs = PPA.objects.select_for_update().filter(
            state=PPA.AVAILABLE, private=settings.PRIVATE_PPAS_ONLY)
        if qs.count() == 0:
            raise PPA.DoesNotExist('No free PPAs exist in pool')
        pk = qs[0].pk
        now = datetime.datetime.utcnow()
        qs.filter(pk__in=(pk,)).update(
            state=PPA.RESERVED, ticket_id=ticket_id, timestamp=now)
        return PPA.objects.get(id=pk)
