# adapted from: http://code.launchpad.net/ppa-dev-tools

import logging
import subprocess

from ci_utils.launchpad import lp_login


def get_package_data(ppa, series, package):
    source = ppa.getPublishedSources(exact_match=True,
                                     source_name=package,
                                     distro_series=series)[0]
    publishing_id = str(source).rpartition('/')[-1]
    source_data = dict()
    source_data['package'] = package
    source_data['binaries'] = source.getPublishedBinaries()
    source_data['version'] = source.source_package_version
    source_data['records'] = ppa.getBuildRecords(source_name=package)
    source_data['build_status'] = ppa.getBuildSummariesForSourceIds(
        source_ids=publishing_id)[publishing_id]
    return source_data


def is_built(source_data):
    try:
        status = source_data['build_status']['status']
        if status not in 'FULLYBUILT':
            logging.info(
                'SKIP %s has status %s', source_data['package'], status)
            return False
        else:
            return True
    except KeyError:
        logging.info('SKIP %s not present in PPA', source_data['package'])
        return False


def records_built(records, package, version):
    arches = set()
    for record in records:
        title = record.title.split()
        # title[0]:arch
        # title[3] package name
        # title[4] version
        if package == title[3] and version == title[4]:
            if record.buildstate != 'Successfully built':
                raise EnvironmentError('%s - %s has not built yet' %
                                       (package, version))
            arches.add(title[0])
    if not arches:
        raise EnvironmentError('%s - %s seems to be a binary copy' %
                               (package, version))
    return arches


def records_published(binaries):
    built_arches = set()
    for build in binaries:
        if build.status == 'Pending':
            raise EnvironmentError('%s has not publised yet' %
                                   (build.display_name))
        if not build.architecture_specific:
            return None
        built_arches.add(build.display_name.split()[4])
    return built_arches


def _get_team_ppa(lp, ppaurl):
    # convert ppa:team/name into a tuple of (team, name)
    team, ppa = ppaurl.split(':')[1].split('/')
    team = lp.people[team]
    ppa = team.getPPAByName(name=ppa)
    return team, ppa


def _should_update(ppa, package, series, source_version):
    try:
        target = ppa.getPublishedSources(
            exact_match=True, source_name=package, distro_series=series)[0]
        if target.date_made_pending:
            return True
        tver = target.source_package_version
        rc = subprocess.call(
            ['dpkg', '--compare-versions', source_version, 'gt', tver])
        return rc == 0
    except:
        return True
    return False


def copy(source_ppa, target_ppa, packages, series):
    launchpad = lp_login()

    ppa_team, ppa = _get_team_ppa(launchpad, source_ppa)
    target_ppa_team, target_ppa = _get_team_ppa(launchpad, target_ppa)

    ubuntu = launchpad.distributions['ubuntu']
    series = ubuntu.getSeries(name_or_version=series)

    for package in packages:
        source_data = get_package_data(ppa, series, package)
        if not is_built(source_data):
            continue
        try:
            arches = records_built(source_data['records'], package,
                                   source_data['version'])
            # Arches built None means binary independent
            arches_built = records_published(source_data['binaries'])
        except EnvironmentError as e:
            logging.info('Skipping %s: %s', package, e.message)
            continue

        if arches_built is not None and arches != arches_built:
            logging.info(
                '%s is not fully published in the source PPA', package)
            continue

        logging.info('Copying %s - %s for arches %s',
                     package, source_data['version'], arches_built)
        target_ppa.syncSource(from_archive=ppa,
                              include_binaries=True,
                              source_name=package,
                              to_pocket='Release',
                              version=source_data['version'],
                              to_series=series.name)
