# Ubuntu CI Engine
# Copyright 2013, 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import json
import threading
import time
import urllib2

import mock

from django.conf import settings
from django.db import connection, transaction
from django.db.utils import ConnectionHandler, DEFAULT_DB_ALIAS
from django.test import TestCase, TransactionTestCase, skipUnlessDBFeature
from django.utils import timezone

from ci_utils.tastypie.test import TastypieTestCase

import ppa_assigner
from ppa_assigner.models import PPA
from ppa_assigner.management.commands.clean_ppas import Command
from ucitests import styles


class TestModel(TestCase):

    def setUp(self):
        super(TestModel, self).setUp()
        settings.PRIVATE_PPAS_ONLY = False
        for n in range(5):
            name = 'ci_pool-%s' % str(n).zfill(3)
            PPA.objects.create(name=name, ticket_id=0)

    def testReserveTimestamp(self):
        ticket = 1234
        p = PPA.reserve(ticket)
        delta = timezone.now() - p.timestamp
        # the lock should have happened in less than a second
        self.assertEqual(PPA.RESERVED, p.state)
        self.assertEqual(ticket, p.ticket_id)
        self.assertGreater(1, delta.seconds)

    def testReserveNoFree(self):
        PPA.objects.update(state=PPA.RESERVED)
        with self.assertRaises(PPA.DoesNotExist):
            PPA.reserve(1234)

    def testReservePrivateOnly(self):
        PPA.objects.create(name='ci_pool-998', ticket_id=0)
        settings.PRIVATE_PPAS_ONLY = True
        with self.assertRaises(PPA.DoesNotExist):
            PPA.reserve(1234)
        PPA.objects.create(name='ci_pool-999', ticket_id=0, private=True)
        p = PPA.reserve(1234)
        self.assertEquals(p.ticket_id, 1234)
        self.assertEquals(p.private, True)


class TestConcurrency(TransactionTestCase):
    # inspired by, but heavily modified:
    # https://github.com/django/django/blob/master/tests/select_for_update/

    def setUp(self):
        super(TestConcurrency, self).setUp()
        settings.PRIVATE_PPAS_ONLY = False
        transaction.enter_transaction_management()
        for n in range(5):
            name = 'ci_pool-%s' % str(n).zfill(3)
            PPA.objects.create(name=name, ticket_id=0)
        transaction.commit()
        self.addCleanup(transaction.abort)

    @contextlib.contextmanager
    def blocking_transaction(self):
        # We need another database connection to test that one connection
        # issuing a SELECT ... FOR UPDATE will block.
        con = ConnectionHandler(settings.DATABASES)[DEFAULT_DB_ALIAS]
        con.enter_transaction_management()
        cursor = con.cursor()
        sql = 'SELECT * FROM %(db_table)s %(for_update)s;' % {
            'db_table': PPA._meta.db_table,
            'for_update': con.ops.for_update_sql(),
        }
        cursor.execute(sql, ())
        cursor.fetchone()
        try:
            yield
        finally:
            con.rollback()
            con.close()

    def _reserve(self, params):
        params['signal'].set()
        start = time.time()
        PPA.reserve(1234)
        params['time'] = time.time() - start
        # connections are per-thread, so close this
        connection.close()

    @skipUnlessDBFeature('has_select_for_update')
    def testReserveConcurrency(self):
        params = {'signal': threading.Event()}
        t = threading.Thread(target=self._reserve, args=(params,))
        locktime = 2
        with self.blocking_transaction():
            t.start()
            # wait for thread to let us know its running
            params['signal'].wait()
            time.sleep(locktime)
        t.join()
        # it won't be exact but it should be pretty close
        self.assertLess(abs(locktime - params['time']), 0.05)


class TestApi(TastypieTestCase):

    def setUp(self):
        super(TestApi, self).setUp('/api/v1')
        settings.PRIVATE_PPAS_ONLY = False
        for n in range(5):
            name = 'ci_pool-%s' % str(n).zfill(3)
            PPA.objects.create(name=name, ticket_id=0)

    def testGetAll(self):
        obj = self.get('ppa/')
        self.assertEqual(PPA.objects.all().count(), len(obj['objects']))

    def testJSONP(self):
        resp = self.client.get('/api/v1/ppa/?format=jsonp')
        content = resp.content.strip()
        self.assertTrue(content.startswith('callback({'))
        self.assertTrue(content.endswith(')'))

    def testGetQuery(self):
        p = PPA.reserve(1234)
        obj = self.get('ppa/', {'state': PPA.AVAILABLE})
        count = PPA.objects.filter(state=PPA.AVAILABLE).count()
        self.assertEqual(count, len(obj['objects']))

        obj = self.getResource('ppa/', {'state': PPA.RESERVED})
        self.assertEqual(p.name, obj['name'])
        self.assertEqual(p.ticket_id, obj['ticket_id'])

    def testReserve(self):
        params = {'ticket_id': 1234}
        obj = self.createResource('ppa/', params)
        self.assertEqual(obj['ticket_id'], params['ticket_id'])

    def testReservePrivateOnly(self):
        PPA.objects.create(name='ci-pool_996', ticket_id=0)
        params = {'ticket_id': 1234}
        # There are only public PPAs available.
        settings.PRIVATE_PPAS_ONLY = True
        with self.assertRaisesRegexp(AssertionError, '404 != 201'):
            self.post('ppa/', params)

        # Creating one private PPA to be reserved.
        PPA.objects.create(name='ci-pool_997', ticket_id=0, private=True)
        obj = self.createResource('ppa/', params)
        self.assertEqual(obj['ticket_id'], params['ticket_id'])

    def testReserveNoFree(self):
        PPA.objects.update(state=PPA.DIRTY)
        params = {'ticket_id': 1234}
        # the post method should raise an AssertionError because it
        # expects a 201 response and gets a 404(not found)
        with self.assertRaisesRegexp(AssertionError, '404 != 201'):
            self.post('ppa/', params)

    def testFreeNumeric(self):
        '''check we can free a ppa with the numeric value in the db'''
        p = PPA.reserve(1234)
        self.patch('ppa/%d/' % p.id, {'state': PPA.DIRTY})
        p = PPA.objects.get(pk=p.id)
        self.assertEqual(p.state, PPA.DIRTY)

    def testFreeString(self):
        '''check we can free a ppa with a string value'''
        p = PPA.reserve(1234)
        self.patch('ppa/%d/' % p.id, {'state': 'dirty'})
        p = PPA.objects.get(pk=p.id)
        self.assertEqual(p.state, PPA.DIRTY)

    @mock.patch('urllib2.urlopen')
    def testPopulateFail(self, urlopen):
        '''ensure a lp error is handled correctly.'''
        def side_effect(arg):
            raise urllib2.URLError('foo')
        urlopen.side_effect = side_effect
        resp = self.client.patch(
            self._resource('ppa/'),
            data={'populate': True},
            authentication=self.auth)
        self.assertEqual(500, resp.status_code)
        self.assertEqual('<urlopen error foo>', resp.content)

    @mock.patch('urllib2.urlopen')
    def testPopulateMerge(self, urlopen):
        ignored = [
            {'name': 'foo', 'private': False},
            {'name': 'ci-pool', 'private': False},
        ]
        expected = [
            {'name': 'ci-pool-001', 'private': False},
            {'name': 'ci-pool-002', 'private': True},
            {'name': 'ci-pool-003', 'private': False},
        ]
        data = {
            'total_size': 2,
            'entries': ignored + expected
        }

        resp = mock.Mock()
        resp.read.return_value = json.dumps(data)
        urlopen.return_value = resp
        ppa_owner = settings.LAUNCHPAD_PPA_OWNER
        self.patch('ppa/', {'populate': True})
        expected = dict(('ppa:{}/{}'.format(ppa_owner, x['name']),
                         x['private']) for x in expected)
        found = dict((x.name, x.private) for x in PPA.objects.all())
        self.assertEqual(len(found), len(expected))
        self.assertListEqual(sorted(expected.keys()), sorted(found.keys()))
        for ppa_name, ppa_privacy in found.iteritems():
            self.assertEqual(expected[ppa_name], ppa_privacy)
        private_ppa = PPA.objects.filter(private=True)
        self.assertEqual(private_ppa.count(), 1)
        self.assertEqual(PPA.objects.filter(private=False).count(), 2)
        self.assertEqual(PPA.objects.all().count(), 3)

    def testCopyBadParams(self):
        resource = self._resource('ppa_copy/')
        resp = self.client.post(
            resource, data={}, authentication=self.auth)
        self.assertHttpBadRequest(resp)

    @mock.patch('ppa_assigner.ppa_sync.copy')
    def testCopy(self, copy):
        resource = self._resource('ppa_copy/')
        data = {
            'source': 'ppa:t1/ppa',
            'target': 'ppa:t2/ppa',
            'series': 'trusty',
            'packages': ['autopilot'],
        }
        resp = self.client.post(
            resource, data=data, authentication=self.auth)
        self.assertHttpAccepted(resp)
        copy.assert_called_once()

    def testStatus(self):
        status = self.get('status/')
        total_ppas = PPA.objects.filter(private=settings.PRIVATE_PPAS_ONLY)
        total = total_ppas.count()
        avail = total_ppas.filter(state=PPA.AVAILABLE).count()
        self.assertEqual('private ppas only', status[1]['label'])
        self.assertEqual(False, status[1]['value'])
        self.assertEqual('total ppas', status[2]['label'])
        self.assertEqual(total, status[2]['value'])
        self.assertEqual('available ppas', status[3]['label'])
        self.assertEqual(avail, status[3]['value'])


class TestCleaner(TestCase):

    def setUp(self):
        super(TestCleaner, self).setUp()
        settings.PRIVATE_PPAS_ONLY = False
        for n in range(5):
            name = 'ci_pool-%s' % str(n).zfill(3)
            PPA.objects.create(name=name, ticket_id=0)

    @mock.patch('ppa_assigner.launchpad._lp_post')
    @mock.patch('ppa_assigner.launchpad.get_publishing_history')
    def testDirtyToClean(self, get_publishing_history, lp_post):
        p = PPA.objects.get(name='ci_pool-002')
        p.state = PPA.DIRTY
        p.save()

        # verify we make one _lp_post call
        get_publishing_history.return_value = [
            {'self_link': 'foo', 'date_made_pending': None},
            {'self_link': 'foo', 'date_made_pending': '123'},
        ]
        resp_200 = mock.Mock()
        resp_200.getcode.return_value = 200
        lp_post.return_value = resp_200
        Command().dirty_to_cleaning()
        p = PPA.objects.get(name='ci_pool-002')
        self.assertEqual(PPA.CLEANING, p.state)
        lp_post.assert_called_once()

    @mock.patch('ppa_assigner.launchpad.request_ppa_clean')
    def testDirtyToCleanFailure(self, request_clean):
        '''make sure state only changes if the call passes.'''
        p = PPA.objects.get(name='ci_pool-002')
        p.state = PPA.DIRTY
        p.save()
        request_clean.side_effect = RuntimeError('foo bar mocked error')

        Command().dirty_to_cleaning()
        p = PPA.objects.get(name='ci_pool-002')
        self.assertEqual(PPA.DIRTY, p.state)

    @mock.patch('ppa_assigner.launchpad.get_publishing_history')
    def testCleaning(self, get_publishing_history):
        p = PPA.objects.get(name='ci_pool-002')
        p.state = PPA.CLEANING
        p.save()

        # first pass we still have an item, so the state shouldn't change
        get_publishing_history.return_value = [
            {'date_made_pending': None},
            {'date_made_pending': '123'},
        ]
        Command().cleaning_to_available()
        p = PPA.objects.get(name='ci_pool-002')
        self.assertEqual(PPA.CLEANING, p.state)

        get_publishing_history.return_value = [
            {'date_made_pending': '123'},
            {'date_made_pending': '123'},
        ]
        Command().cleaning_to_available()
        p = PPA.objects.get(name='ci_pool-002')
        self.assertEqual(PPA.AVAILABLE, p.state)

    @mock.patch('ppa_assigner.launchpad.get_publishing_history')
    def testCleaningFailure(self, get_publishing_history):
        '''make sure it handles an exception from launchpad gracefully.'''
        p = PPA.objects.get(name='ci_pool-002')
        p.state = PPA.CLEANING
        p.save()
        get_publishing_history.side_effect = RuntimeError('foo bar mocked')

        Command().cleaning_to_available()
        p = PPA.objects.get(name='ci_pool-002')
        self.assertEqual(PPA.CLEANING, p.state)


class TestPep8(styles.TestPep8):

    packages = [ppa_assigner]
    exclude = ['migrations/']


class TestPyflakes(styles.TestPyflakes):

    packages = [ppa_assigner]
    excludes = ['ppa_assigner/migrations/0001_initial.py']
