#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ci_utils import (
    amqp_utils,
    amqp_worker,
)
import tstrun
from tstrun import testbed


class TestRunnerWorker(amqp_worker.AMQPWorker):

    def __init__(self):
        super(TestRunnerWorker, self).__init__('image_test')
        self.data_store = None

    def save_subunit(self, log, package, stream, results):
        # make this exception-safe since we can already report pass/fail
        # with or without this file
        try:
            log.info('Saving subunit results for {}'.format(package))
            name = 'test-runner.{}-subunit-stream'.format(package)
            url = self.data_store.put_file(name, stream, 'text/plain')
            results.setdefault('artifacts', []).append({
                'name': name,
                'reference': url,
                'type': 'RESULTS',
            })
        except:
            log.exception(
                'Unable to upload subunit result for {}'.format(package))

    def save_testbed_console(self, log, test_bed, results):
        # make this exception-safe since we can already report pass/fail
        # with or without this file
        try:
            log.info('Saving testbed console')
            name = 'test-runner.testbed-cloud-init.log'
            console = test_bed.get_cloud_init_console()
            url = self.data_store.put_file(name, console, 'text/plain')
            results.setdefault('artifacts', []).append({
                'name': name,
                'reference': url,
                'type': 'LOGS',
            })
        except:
            log.exception('unable to upload testbed console')

    def handle_request(self, log, params):
        ticket_id = params['ticket_id']
        progress_queue = params['progress_trigger']
        image_id = params['image_id']
        package_list = params['package_list']

        results = {}

        def status_cb(msg):
            log.info(msg)
            amqp_utils.progress_update(progress_queue, {'message': msg})

        self.data_store = self._create_data_store(ticket_id)
        try:
            status_cb('Setting up the testbed for ticket {}'.format(ticket_id))
            flavors = tstrun.get_auth_config()['tr_flavors']
            test_bed = testbed.TestBed('testbed-{}'.format(progress_queue),
                                       flavors, image_id, status_cb)
        except:
            log.exception(
                'The testbed creation for {} failed'.format(ticket_id))
            return amqp_utils.progress_failed, results

        try:
            test_bed.setup()
        except:
            log.exception(
                'The testbed setup for {} failed'.format(ticket_id))
            if test_bed.instance is not None:
                self.save_testbed_console(log, test_bed, results)
                test_bed.teardown()
            return amqp_utils.progress_failed, results

        status_cb('The test bed is ready')
        # The tests will succeed unless they fail ;)
        notify = amqp_utils.progress_completed
        try:
            for package in package_list:
                if params.get('cancelled'):
                    log.error('The request  for {} has been cancelled,'
                              ' exiting'.format(ticket_id))
                    return amqp_utils.progress_failed, results
                status_cb('Testing {}'.format(package))
                # uci-vms shell adt-run ...  --- adt-virt-null
                return_code, subunit_stream = test_bed.run_test(package)
                # 0 is success, 8 is skipped and considered a success
                if return_code not in (0, 8):
                    # At least one test failed
                    notify = amqp_utils.progress_failed
                self.save_subunit(log, package, subunit_stream, results)
            status_cb('Test completed for ticket {}'.format(ticket_id))
            return notify, results
        except:
            log.exception(
                'Exception while handling ticket {}'.format(ticket_id))
            raise
        finally:
            test_bed.teardown()


if __name__ == '__main__':
    TestRunnerWorker().main(tstrun.internal_rabbit_queue)
