#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import requests
import subprocess
import time
import yaml

from novaclient import exceptions
from novaclient.v1_1 import client

import tstrun


logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


class NovaClientException(Exception):
    pass


class CloudInitUserData(object):

    """Maps configuration data into cloud-init user-data.

    This is a container for the data that will ultimately be encoded into a
    cloud-config-archive user-data file.
    """

    def __init__(self):
        super(CloudInitUserData, self).__init__()
        # The objects we will populate before creating a yaml encoding as a
        # cloud-config-archive file
        conf = {}
        # Common pre-requisites:
        conf['final_message'] = 'testbed setup completed.'
        conf['apt_update'] = True
        # Default value but we may want to parametrize in the future
        conf['password'] = 'ubuntu'
        # We don't want to be queried to change the password
        conf['chpasswd'] = dict(expire=False)
        # The pre-requisites packages to be able to use adt-run (it's probably
        # a bug in autopkgtest that we need to specify pbuilder here).
        conf['packages'] = ['autopkgtest', 'pbuilder',
                            'python-subunit', 'python-testtools',
                            # We ned rng-tools to generate enough entropy so
                            # gpg can generate its keys.
                            'rng-tools',
                            ]
        self.cloud_config = conf
        self.ubuntu_hook = None

    def _hook_content(self, host_path, user, hook_path, mode='0700'):
        with open(host_path) as f:
            fcontent = f.read()
        hook_content = '''#!/bin/sh
cat >{__guest_path} <<'EOSETUPVMUNIQUECONTENTDONTBREAKFORFUN'
{__fcontent}
EOSETUPVMUNIQUECONTENTDONTBREAKFORFUN
chown {__user}:{__user} {__guest_path}
chmod {__mode} {__guest_path}
'''
        return hook_content.format(__user=user, __fcontent=fcontent,
                                   __mode=mode,
                                   __guest_path=hook_path)

    def set_ubuntu_hook(self):
        """Install run_test.py in ~ubuntu in the testbed."""
        # FIXME: Needs a test ensuring we execute as ubuntu -- vila 2013-03-07
        host_path = os.path.join(os.path.dirname(__file__), '..',
                                 'run_test.py')
        # FIXME: Investigate why this happens before /home/ubuntu exists in the
        # testbed
        hook_path = '/tmp/run_test.py'
        content = self._hook_content(host_path, 'ubuntu', hook_path,
                                     mode='0755')
        if content is not None:
            self.ubuntu_hook = content

    def append_cmd(self, cmd):
        cmds = self.cloud_config.get('runcmd', [])
        cmds.append(cmd)
        self.cloud_config['runcmd'] = cmds

    def add_boot_hook(self, parts, hook):
        if hook is not None:
            parts.append({'content': '#cloud-boothook\n' + hook})

    def dump(self):
        parts = [{'content': '#cloud-config\n'
                  + yaml.safe_dump(self.cloud_config)}]
        self.add_boot_hook(parts, self.ubuntu_hook)
        # Wrap the lot into a cloud config archive
        return '#cloud-config-archive\n' + yaml.safe_dump(parts)


class TestBed(object):

    def __init__(self, name, flavors, image, status_cb, auth_conf=None):
        if auth_conf is None:
            auth_conf = tstrun.get_auth_config()
        self.auth_conf = auth_conf
        self.name = name
        self.flavors = flavors
        self.image = image
        self.status_cb = status_cb
        self.instance = None
        args = [auth_conf.get('auth_user'),
                auth_conf.get('auth_password'),
                auth_conf.get('auth_tenant_name'),
                auth_conf.get('auth_url')]
        kwargs = {'region_name': auth_conf.get('auth_region'),
                  'service_type': 'compute'}
        nova = client.Client(*args, **kwargs)
        self.nova = nova
        self.nova_ssh_key_name = 'test-runner-ssh-key'
        self.test_bed_key_path = None
        self.test_bed_ssh_key_content = None

    def ensure_ssh_key_is_available(self):
        # From the test runner, we need an ssh key that can be used to connect
        # to the test bed. For testing purposes, we rely on self.auth_conf to
        # provide this key.
        if not os.path.exists(self.test_bed_key_path):
            # First time the test runner instance needs to create a test bed
            # instance, we need to create the ssh key pair.
            subprocess.call(
                ['ssh-keygen', '-t', 'rsa',
                 '-f', self.test_bed_key_path, '-N', ''])

        # MISSINGTEST: the key (or the .pub) doesn't exist (related to errors
        # occuring in the ssh-keygen call above but can happen for other
        # reasons) -- vila 2014-01-30
        with open(self.test_bed_key_path + '.pub') as f:
            self.test_bed_ssh_key_content = f.read()

        # Register the key in nova
        if not self.nova.keypairs.findall(name=self.nova_ssh_key_name):
            self.nova.keypairs.create(
                name=self.nova_ssh_key_name,
                public_key=self.test_bed_ssh_key_content)

    def create_user_data(self):
        self.test_bed_key_path = self.auth_conf.get('auth_testbed_sshkey',
                                                    '~/.ssh/id_rsa')
        self.test_bed_key_path = os.path.expanduser(self.test_bed_key_path)
        self.ensure_ssh_key_is_available()
        ud = CloudInitUserData()
        # MISSINGTEST: Since we tell nova to use that key (see setup()), we may
        # not need to use cloud-init to ensure it's authorized on the test bed
        # anymore. -- vila 2014-01-30
        ud.cloud_config['ssh_authorized_keys'] = [
            self.test_bed_ssh_key_content]
        ud.set_ubuntu_hook()
        self.cloud_init_completion_message = ud.cloud_config['final_message']
        self.user_data = ud.dump()

    def find_flavor(self):
        for flavor in self.flavors:
            try:
                return self.nova.flavors.find(name=flavor)
            except exceptions.NotFound:
                pass
        # MISSINGTEST: some cloud doesn't provide one of our expected flavors
        # -- vila 2014-01-06
        raise NovaClientException(
            'None of {} can be found'.format(self.flavors))

    def setup(self):
        flavor = self.find_flavor()
        image = self.nova.images.find(name=self.image)
        self.create_user_data()
        self.instance = self.nova.servers.create(
            name=self.name, flavor=flavor, image=image,
            userdata=self.user_data, keypair=self.nova_ssh_key_name)
        # FIXME: This times out for hpcloud (when used for testing from dev
        # host) because it requires a floating ip for the instance see
        # http://pad.lv/1288622 -- vila 2014-03-06
        self.wait_for_ip()
        self.wait_for_cloud_init()

    def wait_for_ip(self):
        status = self.instance.status
        # FIXME: get_ip_timeout should be a config option -- vila 2014-01-30
        get_ip_timeout = 300  # in seconds so 5 minutes
        timeout_limit = time.time() + get_ip_timeout
        while time.time() < timeout_limit and status != 'ACTIVE':
            # Always query nova to get updated data about the instance
            # MISSINGTEST: What if the instance disappear ? -- vila 2014-01-30
            try:
                server = self.nova.servers.get(self.instance.id)
                status = server.status
            except requests.ConnectionError:
                # The reported status is the one known before the last attempt.
                log.warn('Received connection error for {},'
                         ' retrying (status was: {})'.format(self.instance.id,
                                                             status))
                time.sleep(5)
                continue
            networks = server.networks.values()
            if networks:
                # The network name is arbitrary, can vary for different clouds
                # but there should be only one IP address defined so we just
                # get the first one and avoid the need for a config option for
                # the network name.

                # FIXME: ev encountered a case where the instance is in an
                # error state and has no networks leading to a KeyError
                # exception. It should be handled in its own branch based on
                # 'status == "ERROR"'. -- vila 2014-02-18
                self.ip = networks[0][0]
                log.info('Got IP {} for {}'.format(self.ip, self.instance.id))
                # FIXME: Right place to report how long it took to spin up the
                # instance as far as nova is concerned. -- vila 2014-01-30
                return
            else:
                log.info(
                    'IP not yet available for {}'.format(self.instance.id))
            time.sleep(5)
        # MISSINGTEST: What if the instance doesn't come up after max_tries ?
        if status == 'BUILD':
            raise NovaClientException('Instance never came up')
        else:
            raise NovaClientException('Instance never provided an IP')

    def get_cloud_init_console(self, length=None):
        return self.instance.get_console_output(length)

    def wait_for_cloud_init(self):
        # FIXME: cloud_init_timeout should be a config option (related to
        # get_ip_timeout and probably the two can be merged) -- vila 2014-01-30
        cloud_init_timeout = 300  # in seconds so 5 minutes
        timeout_limit = time.time() + cloud_init_timeout
        while time.time() < timeout_limit:
            # A relatively cheap way to catch cloud-init completion is to watch
            # the console for the specific message we specified in user-data).
            try:
                console = self.get_cloud_init_console(10)
            except requests.ConnectionError:
                # The reported status is the one known before the last attempt.
                log.warn('Received connection error for {},'
                         ' retrying)'.format(self.instance.id))
                time.sleep(5)
                continue
            if self.cloud_init_completion_message in console:
                # We're good to go
                log.info(
                    'cloud-init completed for {}'.format(self.instance.id))
                # FIXME: Right place to report how long it
                # took for cloud-init to finish. -- vila 2014-01-30
                return
            time.sleep(5)
        raise NovaClientException('Instance never completed cloud-init')

    def teardown(self):
        self.nova.servers.delete(self.instance.id)
        # FIXME: Now we can remove the testbed key from known_hosts (see
        # ssh()). -- vila 2014-01-30

    def ssh(self, command, *args, **kwargs):
        out = kwargs.pop('out', None)
        err = kwargs.pop('err', None)
        user = 'ubuntu'
        host = self.ip
        cmd = ['ssh',
               # FIXME: It would be better to ssh-keygen -f
               # "~/.ssh/known_hosts" -R self.ip once we're done with the test
               # bed to avoid polluting ssh commands stdout, but that will do
               # for now (that's what juju is doing after all ;)
               # -- vila 2014-01-29
               '-oStrictHostKeyChecking=no',
               '-i', self.test_bed_key_path,
               '{}@{}'.format(user, host)]
        if command is not None:
            cmd += [command]
        if args:
            cmd += args
        proc = subprocess.Popen(cmd, stdout=out, stderr=err)
        out, err = proc.communicate()
        return proc, out, err

    def run_test(self, package):
        proc, _, _ = self.ssh('/tmp/run_test.py', package)
        # FIXME: Does that look like a pipe ? -- vila 2014-02-06
        _, subunit_stream, _ = self.ssh('cat', 'result.subunit',
                                        out=subprocess.PIPE)
        return proc.returncode, subunit_stream


# The following are helpers for local manual tests, they should disappear once
# we have proper integration tests.

# Those helpers rely on a unit_config file at the root of the branch (/!\ NEVER
# to be put under version control) containing the nova credentials (in effect,
# doing manually what juju-deployer does on the juju units). An additional
# 'auth_testbed_sshkey' can be defined there if you don't use ~/.ssh/id_rsa.

def test_print_ip():
    conf = tstrun.get_auth_config()
    # The following images are "known" to work when they are added while
    # testing manually. In general, they need to be updated because they
    # disappear to be replaced by newer ones. They also vary depending on
    # whether canonistack or hpcloud is used.
    test_images = dict(
        cs_precise=('ubuntu-released/ubuntu-precise-12.04'
                    '-amd64-server-20131205-disk1.img'),
        cs_saucy=('ubuntu-released/ubuntu-saucy-13.10'
                  '-amd64-server-20140212-disk1.img'),
        hp_precise=('Ubuntu Server 12.04.4 LTS (amd64 20140227)'
                    ' - CI Engineering'),
        hp_saucy=('Ubuntu Server 13.10 (amd64 20131030) - Partner Image'),
    )
    test_bed = TestBed('localtest',
                       conf['tr_flavors'],
                       test_images['cs_saucy'],
                       lambda x: None)
    test_bed.setup()
    print 'cloud-init user-data: {}'.format(test_bed.user_data)
    print 'IP: {}'.format(test_bed.ip)
    return test_bed


if __name__ == '__main__':
    # the worker might not have installed this module, so determine the path
    # and add it, so we can always safely import stuff
    import sys
    sys.path.insert(0, os.path.join(
        os.path.dirname(__file__), '../../ci-utils'))
    from ci_utils import dump_stack
    dump_stack.install_stack_dump_signal()
    test_bed = test_print_ip()
    print test_bed.run_test('libpng')  # Known to run no tests
    print test_bed.run_test('juju-core')
    test_bed.teardown()
