#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2013, 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import urllib2
import unittest
import sys
import yaml

sys.path.append(os.path.join(
    os.path.dirname(__file__), '..', '..', 'ci-utils'))

from ci_utils import data_store


class TestDataStore(unittest.TestCase):

    def setUp(self):
        super(TestDataStore, self).setUp()
        default_config = os.path.join(os.environ["HOME"], '.ubuntu-ci')
        if not os.path.exists(default_config):
            self.skipTest('{} is not available'.format(default_config))
        with open(default_config) as fp:
            self.auth_config = yaml.safe_load(fp)

        self.ds = data_store.DataStore("test", auth_config=self.auth_config)
        self.assert_(self.ds, "Failed to create DataStore")
        self.addCleanup(self.ds.delete, recursive=True)

        self.ds.clear()
        self.filename = 'test_file'
        self.contents = 'some data'

    def test_list_files_empty(self):
        """ Test listing files on an empty container. """
        files = self.ds.list_files()
        self.assertEqual([], files)

    def test_add_file(self):
        """ Test adding a file. """

        self.ds.put_file(self.filename, self.contents)
        self.addCleanup(self.ds.delete_file, self.filename)
        files = self.ds.list_files()
        self.assertEqual([self.filename], files)

    def test_add_file_contents(self):
        """ Test adding a file has correct contents. """

        self.ds.put_file(self.filename, self.contents)
        self.addCleanup(self.ds.delete_file, self.filename)
        contents = self.ds.get_file(self.filename)

        self.assertEqual(contents, self.contents)

    def test_public_container(self):
        """ Test accesssing a public container. """

        self.ds.put_file(self.filename, self.contents)
        self.addCleanup(self.ds.delete_file, self.filename)
        self.ds.change_visibility(public=True)

        response = urllib2.urlopen(self.ds.file_path(self.filename))
        self.assertEqual(response.code, 200)

    def test_delete_file(self):
        """ Test delete file removes the file. """

        self.ds.put_file(self.filename, self.contents)
        files = self.ds.list_files()
        self.assertEqual([self.filename], files)

        self.ds.delete_file(self.filename)
        files = self.ds.list_files()
        self.assertEqual([], files)

    def test_auth_missing(self):
        """ Test that missing auth details is handled correctly. """

        with self.assertRaises(data_store.DataStoreException):
            data_store.DataStore("test", auth_config={})

    def test_put_file_link(self):
        """ Test put file returns a valid link. """

        link = self.ds.put_file(self.filename, self.contents)
        self.addCleanup(self.ds.delete_file, self.filename)
        self.ds.change_visibility(public=True)
        resp = urllib2.urlopen(link)
        self.assertEqual(200, resp.getcode())

    def test_missing_region(self):
        """ Test that not passing a region fails. """

        del self.auth_config['auth_region']
        with self.assertRaises(data_store.DataStoreException):
            data_store.DataStore("test2", auth_config=self.auth_config)

    def test_bad_region(self):
        """ Test that passing an invalid region fails. """

        self.auth_config['auth_region'] = 'bad_region'
        with self.assertRaises(data_store.DataStoreException):
            data_store.DataStore("test2", auth_config=self.auth_config)


if __name__ == "__main__":
    unittest.main()
