# Ubuntu Continuous Integration Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import base64
import os
import shutil
import subprocess
import unittest

import yaml

import amulet

defdir = os.path.join(os.path.dirname(__file__), '../juju-deployer')
DEPLOYER_DIR = os.environ.get('JUJU_DEPLOYER_DIR', defdir)


class AmuletDeployment(object):
    '''Fixture for a deployed set of juju services from a deployer config'''

    def __init__(self, deployer_cfg, juju_env=None, sentries=False):
        self.deployment = amulet.Deployment(juju_env, sentries=sentries)
        with open(deployer_cfg) as f:
            jd_script = yaml.safe_load(f.read())
        self.fixup_includes(os.path.dirname(deployer_cfg), jd_script)
        # Use juju_env given or defaulted above rather than taking the name
        # from the deployer config and assuming an env of that name exists
        script = {self.deployment.juju_env: jd_script.values()[0]}
        self.deployment.load(script)

    def setUp(self, timeout):
        self.deployment.setup(timeout)

    def fixup_includes(self, path, jd_script):
        # amulet doesn't deal with "include-base64" options, we have to
        # patch the config ourselves
        for deployment in jd_script:
            for service in jd_script[deployment]['services'].values():
                for key, val in service.get('options', {}).iteritems():
                    include_token = 'include-base64://'
                    if type(val) is str and val.startswith(include_token):
                        val = val.replace(include_token, '')
                        p = os.path.join(path, val)
                        with open(p) as f:
                            val = base64.b64encode(f.read())
                        service['options'][key] = val


class DeployerTest(unittest.TestCase):
    '''Base class for building juju deployer based tests.'''

    deployer_cfg = None
    timeout = 600

    def setUp(self):
        super(DeployerTest, self).setUp()
        self.deployer = AmuletDeployment(self.deployer_cfg)
        self.addCleanup(self._clean_amulet)
        if self.timeout:
            self.deployer.setUp(self.timeout)
        self.status = amulet.waiter.status(self.deployer.deployment.juju_env)

    def _clean_amulet(self):
        if self.deployer.deployment.deployer_dir:
            shutil.rmtree(self.deployer.deployment.deployer_dir)

    def get_ip_and_port(self, service, unit=0):
        units = self.status['services'][service]['units']
        ip = units['%s/%d' % (service, unit)]['public-address']
        port, _ = units['%s/%d' % (service, unit)]['open-ports'][0].split('/')
        return ip, int(port)

    def assert_job_running(self, upstart_job, service, unit=0):
        '''ensure the given upstart job is running on a unit'''
        unit = '{}/{}'.format(service, unit)
        args = [
            'juju', 'ssh', unit,
            '-o', 'StrictHostKeyChecking=no',
            '-o', 'UserKnownHostsFile=/dev/null',
            '/sbin/status', upstart_job
        ]
        output = subprocess.check_output(args)
        self.assertIn('start/running', output)
