#!/usr/bin/env python

import json
import os
import unittest

import httplib2

from tests.deployers import DEPLOYER_DIR, DeployerTest


class PPAAssignerTest(DeployerTest):
    """Integration tests for ppa-assigner service run on a juju deployment"""

    deployer_cfg = os.path.join(DEPLOYER_DIR, 'ppa-assigner.yaml')

    def get_server_status_and_content(self, url):
        final_url = url.format(*self.get_ip_and_port('ppa-django'))

        # Issue a simple command against the ppa assigner API
        client = httplib2.Http()
        resp, content = client.request(final_url, 'GET')
        jcontent = json.loads(content)
        return resp, jcontent

    def reserve_ppa(self, ticket_id):
        url = 'http://{}:{}/api/v1/ppa/'
        final_url = url.format(*self.get_ip_and_port('ppa-django'))

        client = httplib2.Http()
        resp, content = client.request(
            uri=final_url,
            method='POST',
            headers={'Content-Type': 'application/json'},
            body=json.dumps({"ticket_id": ticket_id}),
        )
        return resp, content

    def release_ppa(self, url):
        client = httplib2.Http()
        resp, content = client.request(
            uri=url,
            method='PATCH',
            headers={'Content-Type': 'application/json'},
            body=json.dumps({"state": "dirty"}),
        )
        return resp, content

    def reset_ppa(self, url):
        client = httplib2.Http()
        resp, content = client.request(
            uri=url,
            method='PATCH',
            headers={'Content-Type': 'application/json'},
            body=json.dumps({"state": "available", "ticket_id": 0}),
        )
        return resp, content

    def get_total_free_ppa(self):
        url = 'http://{}:{}/api/v1/ppa'
        resp, content = self.get_server_status_and_content(url)
        tot_free = 0
        for i in range(0, len(content["objects"])):
            if content["objects"][i]["state"] == 0:
                tot_free += 1
        return tot_free

    def test_reserve_more_than_in_the_pool(self):
        tot_free_ppa = self.get_total_free_ppa()
        ppa_url = []
        for ticket_id in range(1, tot_free_ppa+1):
            resp, content = self.reserve_ppa(ticket_id)
            self.assertEqual('201', resp['status'])
            ppa_url.append(resp['location'])

        #  Attmpt to assign another ppa
        resp, content = self.reserve_ppa(tot_free_ppa+1)
        self.assertEqual('404', resp['status'])
        self.assertIn('No free PPAs exist in pool', content)

        for ticket_id in range(1, tot_free_ppa+1):
            resp, content = self.release_ppa(ppa_url[ticket_id-1])
            self.assertEqual('202', resp['status'])  # ACCEPTED

            # reset the ppa to put it back to the pool
            resp, content = self.reset_ppa(ppa_url[ticket_id-1])
            self.assertEqual('202', resp['status'])  # ACCEPTED

    def test_reserve_and_release(self):
        resp, content = self.reserve_ppa(101)
        self.assertEqual('201', resp['status'])  # CREATED
        ppa_url = resp['location']

        resp, content = self.release_ppa(ppa_url)
        self.assertEqual('202', resp['status'])

        # put the ppa back to the available pool
        resp, content = self.reset_ppa(ppa_url)
        self.assertEqual('202', resp['status'])

    def test_ppa_api(self):
        url = 'http://{}:{}/api/v1/ppa'
        resp, content = self.get_server_status_and_content(url)

        self.assertEqual('200', resp['status'])
        self.assertNotEqual([], content['objects'])

    def test_ppa_name(self):
        url = 'http://{}:{}/api/v1/ppa'
        resp, content = self.get_server_status_and_content(url)
        expected = 'ppa:{}/ci-pool-'.format(
            os.environ['CI_LAUNCHPAD_PPA_OWNER'])
        self.assertIn(expected, content['objects'][0]['name'])

    def test_next_page_ppa(self):
        url = 'http://{}:{}/api/v1/ppa/?format=json&limit=1'
        resp, content = self.get_server_status_and_content(url)

        # if there's more than one PPA
        if content['meta']['next']:
            self.assertEqual('/api/v1/ppa/?offset=1&limit=1&format=json',
                             content['meta']['next'])
        self.assertEqual('/api/v1/ppa/1/',
                         content['objects'][0]['resource_uri'])

    def test_previous_page_ppa(self):
        url = 'http://{}:{}/api/v1/ppa/?format=json&limit=1&offset=1'
        resp, content = self.get_server_status_and_content(url)

        # if there's more than one PPA
        if content['meta']['next']:
            self.assertEqual('/api/v1/ppa/?format=json&limit=1&offset=0',
                             content['meta']['previous'])
            self.assertEqual('/api/v1/ppa/2/',
                             content['objects'][0]['resource_uri'])

if __name__ == "__main__":
    unittest.main()
