#!/usr/bin/python

import atexit
import logging
import os
import shutil
import signal
import subprocess
import sys
import tempfile
import time
import yaml

import swiftclient

from novaclient.v1_1 import client
from novaclient.exceptions import NotFound

SSH_OPTS = ['-o', 'UserKnownHostsFile=/dev/null',
            '-o', 'StrictHostKeyChecking=no']
SUDOERS_FILE = '/etc/sudoers.d/sshuttle'
SUDOERS = 'ALL = (root) NOPASSWD: /usr/bin/python /usr/lib/sshuttle/main.py *'


logging.basicConfig(level=logging.INFO, format='%(msg)s')
log = logging.getLogger('integration-test-runner')


HERE = os.path.abspath(os.path.dirname(__file__))


# deploy.py probably isn't in our pythonpath
sys.path.append(os.path.join(os.path.dirname(__file__), '../juju-deployer'))
import deploy


def _get_status():
    """Returns a dictionary of the output of `juju status`."""

    r = subprocess.check_output(['juju', 'status'], stderr=subprocess.STDOUT)
    return yaml.safe_load(r)


def _destroy_machine(machine):
    """Terminate a nova instance used by juju. This does not destroy the juju
    bootstrap node's cache of the local service, so it is important to call
    _destroy_service as well."""

    args = ['juju', 'destroy-machine', machine, '--force']
    try:
        subprocess.check_output(args)
    except subprocess.CalledProcessError as e:
        if ('machine %s does not exist' % machine) in e.output:
            return
        else:
            raise


def _destroy_service(service):
    """Destroy a deployed service."""

    subprocess.check_output(['juju', 'destroy-service', service])


def _has_only_bootstrap_node(status):
    return status.get('machines').keys() == ['0']


def _wait_for_death():
    """Wait for the environment to return to just the bootstrap node."""

    tries = 0
    while tries < 60:
        status = _get_status()
        services = status.get('services')
        if not services and _has_only_bootstrap_node(status):
            return

        time.sleep(3)
        tries += 1
    else:
        raise Exception('Timed out waiting for death.')


def assert_no_deployed_services():
    """Assert that there are no deployed services or nodes beyond the bootstrap
    node."""

    # First ensure that anything dying has made the transition to the other
    # side.
    _wait_for_death()

    status = _get_status()

    assert status.get('services') == {}, status.get('services')
    # We should still have a bootstrap node ('0').
    assert status.get('machines').keys() == ['0'], status.get('machines')


def destroy_services():
    """This is a quick way of destroying the services without having to
    re-bootstrap for the next test. We would just use juju-deployer -T here,
    but https://bugs.launchpad.net/juju-deployer/+bug/1269783 prevents that for
    now."""

    log.info("Destroying services.")

    status = _get_status()
    machines = status.get('machines', {}).keys()
    for machine in machines:
        if machine != '0':
            _destroy_machine(machine)

    status = _get_status()
    services = status.get('services', {}).keys()
    for service in services:
        _destroy_service(service)


def _find_test_files(test_dir, tests):
    for test in tests:
        actual_test = os.path.join(test_dir, test, 'test.py')
        if not os.path.exists(actual_test):
            msg = "Skipping %s as %s does not exist." % (test, actual_test)
            log.error(msg)
        else:
            yield actual_test


def _find_dirs_in_dir(test_dir):
    """Return an iterator of only the directories in test_dir."""

    for x in os.listdir(test_dir):
        if os.path.isdir(os.path.join(test_dir, x)):
            yield x


def get_tests(test_dir):
    """Returns a list of full relative paths to test files."""

    tests = os.environ.get('TESTS', '')
    if tests:
        tests = tests.split(' ')
    else:
        tests = _find_dirs_in_dir(test_dir)
    return list(_find_test_files(test_dir, tests))


def find_failed_units():
    """Return an iterator of all units that are in an error state."""

    status = _get_status()
    for service in status['services'].itervalues():
        for unit, data in service.get('units', {}).iteritems():
            if data.get('agent-state') == 'error':
                yield unit


def dump_log(unit):
    """Print the contents of the juju for the service to stdout. If the tests
    fail, there may be clues in this output."""

    logfile = '/var/log/juju/unit-%s.log' % unit.replace('/', '-')
    cmd = ['juju', 'ssh', unit, 'cat', logfile]
    log.error('Errors found in %s (%s):' % (unit, logfile))
    subprocess.check_call(cmd)


def install_charmworld(where=None):
    """Recent versions of amulet need charmworldlib. Install it."""
    if where is None:
        where = os.path.join(HERE, '..', 'branches')
    dest = os.path.abspath(os.path.join(where, 'charmworldlib'))
    deploy.get_or_pull_branch('lp:charmworldlib', dest)
    deploy.prepend_path('PYTHONPATH', dest)
    return dest


def install_amulet(where=None):
    if where is None:
        where = os.path.join(HERE, '..', 'branches')
    dest = os.path.abspath(os.path.join(where, 'amulet'))
    deploy.get_or_pull_branch('lp:amulet', dest)
    deploy.prepend_path('PYTHONPATH', dest)
    return dest


def remove_branch(branch_dir):
    """Clean up the bzr branch if we didn't already do so as a result of
    bzr failing."""

    if os.path.exists(branch_dir):
        shutil.rmtree(branch_dir)


def needs_bootstrap():
    """Check to see if there's a Swift bucket for juju. If not, we need to
    boostrap. Check to see if the IP of the bootstrap node is owned by a
    running Nova instance. If not, we need to bootstrap."""

    log.info("Checking to see if a bootstrap is needed.")

    def inconsistent_state(msg):
        log.warning(msg)
        _delete_control_bucket()

    try:
        instance = get_bootstrap_nova_instance()
    except swiftclient.exceptions.ClientException:
        instance = None
    if instance is None:
        # We don't have an instance. We need to bootstrap.
        inconsistent_state('No bootstrap instance in the swift bucket')
        return True

    try:
        get_bootstrap_ip(instance, 1)
    except NotFound:
        # Juju claims we have an instance that does not exist. We're in an
        # inconsistent state. Delete the bucket and bootstrap.
        inconsistent_state('No IP for the bootstrap instance')
        return True

    return False


def _get_pid_of_sshuttle(ip_addr):
    try:
        cmd = ['pgrep', '-f', 'sshuttle.*%s' % ip_addr]
        pid = subprocess.check_output(cmd)
        # pgrep returns a list of pids matching the search argument
        pid.split('\n')[0]
        return int(pid)
    except subprocess.CalledProcessError:
        return -1


def needs_tunnel():
    """Check to see if juju's Swift bucket exists, and the IP it refers to is
    pointed at by a sshuttle instance."""

    log.info("Checking to see if a tunnel is needed.")

    instance = get_bootstrap_nova_instance()
    ip_addr = get_bootstrap_ip(instance, 1)
    sshuttle_pid = _get_pid_of_sshuttle(ip_addr)
    return sshuttle_pid < 0


def _connect_to_swift():
    opts = {'tenant_name': os.environ['OS_TENANT_NAME'],
            'region_name': os.environ['OS_REGION_NAME'],
            }
    conn = swiftclient.client.Connection(
        os.environ['OS_AUTH_URL'],
        os.environ['OS_USERNAME'],
        os.environ['OS_PASSWORD'],
        os_options=opts,
        auth_version='2.0')
    return conn


def _get_from_swift(bucket, path):
    """Download the file path from swift at bucket."""

    conn = _connect_to_swift()
    # get_object return a tuple of (metadata, data)
    return conn.get_object(bucket, path)[1]


def _get_control_bucket():
    """Get the name of the bucket in Swift that juju uses for state
    information."""

    default_home = os.path.join(os.path.expanduser('~'), '.juju')
    juju_home = os.environ.get('JUJU_HOME') or default_home

    with open(os.path.join(juju_home, 'environments.yaml')) as fp:
        envs = yaml.safe_load(fp.read())

    juju_env = os.environ.get('JUJU_ENV')
    if juju_env is None:
        juju_env = envs.get('default')
    if not juju_env:
        log.error('No juju environment specified.')
        sys.exit(1)

    return envs['environments'][juju_env]['control-bucket']


def _delete_control_bucket():
    """Remove the juju control bucket and all of its files."""

    bucket = _get_control_bucket()
    conn = _connect_to_swift()
    try:
        files = [x['name'] for x in conn.get_container(bucket)[1]]
        for f in files:
            conn.delete_object(bucket, f)
        conn.delete_container(bucket)
    except swiftclient.exceptions.ClientException:
        # Assume the bucket is not present
        pass


def get_bootstrap_nova_instance():
    """Get the Nova instance identifier for the juju bootstrap node.

    Raises swiftclient.exceptions.ClientException if the instance cannot be
    found."""

    bucket = _get_control_bucket()

    state = yaml.safe_load(_get_from_swift(bucket, 'provider-state'))
    if not state:
        # Cause unknown (yet) but this has been seen in real life. I can't
        # reproduce anymore but I suspect 'juju bootstrap' failed mid-way so
        # returning None and letting the caller delete the bucket is the Right
        # Thing to do -- vila 2014-02-07
        log.warning("Empty 'provider-state' in the juju bucket")
        return None
    return state['state-instances'][0]


def get_bootstrap_ip(instance, max_tries=120):
    """Wait for the juju bootstrap node to boot and obtain an IP address so we
    can tunnel through it.

    Raises NotFound if the instance has not been spawned."""

    log.info("Getting the bootstrap IP address.")

    args = [os.environ['OS_USERNAME'], os.environ['OS_PASSWORD'],
            os.environ['OS_TENANT_NAME'], os.environ['OS_AUTH_URL']]
    kwargs = {'region_name': os.environ['OS_REGION_NAME'],
              'service_type': 'compute'}
    nt = client.Client(*args, **kwargs)

    net = ''
    tries = 0
    while tries < max_tries:
        # Raises NotFound if the instance does not exist.
        server = nt.servers.get(instance)
        net = server.networks.get('canonistack')
        if net:
            net = net[0]
            # Don't sleep anymore. We have our IP; break out of the loop.
            break

        tries += 1
        if tries != max_tries:
            time.sleep(3)
    else:
        raise Exception('Timed out waiting for bootstrap IP.')

    return net


def destroy_tunnel():
    if os.path.exists('sshuttle.pid'):
        with open('sshuttle.pid') as fp:
            pid = int(fp.read())
            os.kill(pid, signal.SIGTERM)


def tunnel(ip):
    """Tunnel to the bootstrap node using sshuttle."""

    log.info("Tunneling to the bootstrap node.")

    host = 'ubuntu@%s' % ip
    mask = '10.55.0.0/16'
    cmd = ['sshuttle', '-D', '-r', host, mask]
    opt = ['-e', 'ssh %s' % ' '.join(SSH_OPTS)]
    subprocess.check_call(cmd + opt)
    atexit.register(destroy_tunnel)


def wait_for_node(ip):
    """Wait for the bootstrap node to become available on ssh. We need to wait
    for this state before attempting to set up a sshuttle."""

    log.info("Waiting for the bootstrap node to respond to ssh.")

    cmd = ['ssh'] + SSH_OPTS + ['ubuntu@%s' % ip, 'exit']
    tries = 0
    while tries < 120:
        try:
            subprocess.check_call(cmd)
            return
        except subprocess.CalledProcessError:
            tries += 1
            time.sleep(5)

    raise Exception('Timed out waiting for bootstrap node.')


def bootstrap():
    """Bootstrap juju. Specify needed as bootstrap to just bootstrap, otherwise
    tunnel to bootstrap and tunnel."""

    args = ['juju', 'bootstrap', '--constraints=mem=1024M']
    try:
        subprocess.check_output(args)
    except subprocess.CalledProcessError as e:
        # FIXME: if check_output raise an exception, we don't get a proper
        # output to check for the text below -- vila 2014-02-07
        if 'environment is already bootstrapped' not in e.output:
            raise


def wait_then_tunnel():
    instance = get_bootstrap_nova_instance()
    ip = get_bootstrap_ip(instance)
    wait_for_node(ip)
    tunnel(ip)


def check_sudoers(sudoers_path=SUDOERS_FILE):
    """Check to see if /etc/sudoers.d/sshuttle exists and is of the right
    form."""

    if not os.path.exists(sudoers_path):
        return False

    try:
        with open(sudoers_path) as fp:
            contents = fp.read()
            if not contents.strip('\n').endswith(SUDOERS):
                return False
    except IOError as e:
        if e.errno == 13:
            msg = "Unable to read %s. Assuming it's correct." % sudoers_path
            log.error(msg)
            return True
        else:
            raise

    return True


def check_nova_env():
    """Ensure we have the right keys set in the environment for talking to
    Canonistack."""

    if os.environ.get('OS_USERNAME'):
        return True
    else:
        return False


def check_juju_version():
    """Return True if the correct version of juju is installed, False
    otherwise."""

    # Juju 1.16.5 introduced the --force flag to terminate-machine.
    version = subprocess.check_output(['juju', '--version']).strip()
    rc = subprocess.call(
        ['dpkg', '--compare-versions', version, 'ge', '1.16.5'])
    return rc == 0


def _prepare_deployer():
    deployer_dir = os.path.join(os.path.dirname(__file__), '../juju-deployer')
    temp = tempfile.mkdtemp()
    atexit.register(lambda: shutil.rmtree(temp))
    working_dir = os.path.join(temp, 'deployer')
    shutil.copytree(deployer_dir, working_dir)
    deploy.upload_payload_to_swift()
    deploy.populate_templates(deployer_dir, working_dir)
    deploy.build_charms()
    os.environ['JUJU_DEPLOYER_DIR'] = working_dir


def main():
    test_dir = os.path.dirname(sys.argv[0])
    deploy.check_environment()
    tests = get_tests(test_dir)
    if not tests:
        sys.exit(0)

    if not check_juju_version():
        msg = 'Please install a more recent version of juju (1.16.5 or later).'
        log.error(msg)
        sys.exit(1)

    if not check_nova_env():
        msg = 'Please source a novarc file before running this harness.'
        log.error(msg)
        sys.exit(1)

    if not check_sudoers():
        msg = 'Please run:\necho "$USER %s" > %s' % (SUDOERS, SUDOERS_FILE)
        log.error(msg)
        sys.exit(1)

    _prepare_deployer()

    install_amulet()
    install_charmworld()
    deploy.install_deployer()
    deploy.install_jujuclient()

    # Now, insert the 'tests' dir in front of PYTHONPATH so we get our tests
    # without colliding with other 'tests' directories defined by other
    # packages (including our deps).
    deploy.prepend_path('PYTHONPATH', os.path.dirname(test_dir))

    if needs_bootstrap():
        bootstrap()
        wait_then_tunnel()
    elif needs_tunnel():
        wait_then_tunnel()

    destroy_services()
    assert_no_deployed_services()
    atexit.register(destroy_services)

    returncode = 0
    for test in tests:
        log.error('========================================')
        log.error('Testing %s' % test)
        log.error('========================================')
        try:
            subprocess.check_call(['python', test])
        except subprocess.CalledProcessError as e:
            returncode = e.returncode

            # Let's see if there are some failed instances that we can gather
            # logs from.
            units = find_failed_units()
            for unit in units:
                dump_log(unit)

        destroy_services()
        assert_no_deployed_services()
    return returncode


if __name__ == '__main__':
    sys.exit(main())
