#!/usr/bin/env python

import unittest
import mock
import run
import yaml
import tempfile
import shutil
import os
import logging
import stat
import swiftclient
from novaclient.exceptions import NotFound


from ucitests import fixtures


logging.disable(logging.CRITICAL)


class TestRunTests(unittest.TestCase):
    def _destroy_services(self, fake_status, destroy_machine, destroy_service):
        fake_status = yaml.safe_dump(fake_status)

        with mock.patch('subprocess.check_output', return_value=fake_status):
            run.destroy_services()

    @mock.patch('run._destroy_service')
    @mock.patch('run._destroy_machine')
    def test_destroy_services(self, destroy_machine, destroy_service):
        fake_status = {'services': {'imagebuild-restish': {}},
                       'machines': {'2': {}, '0': {}}}
        self._destroy_services(fake_status, destroy_machine, destroy_service)

        destroy_machine.assert_called_with('2')
        destroy_service.assert_called_with('imagebuild-restish')

    @mock.patch('run._destroy_service')
    @mock.patch('run._destroy_machine')
    def test_destroy_no_services(self, destroy_machine, destroy_service):
        fake_status = {'services': {}, 'machines': {}}
        self._destroy_services(fake_status, destroy_machine, destroy_service)

    def test_get_tests_no_tests(self):
        temp_dir = tempfile.mkdtemp()
        self.addCleanup(lambda: shutil.rmtree(temp_dir))
        self.assertEqual(run.get_tests(temp_dir), [])

    def test_get_tests(self):
        temp_dir = tempfile.mkdtemp()
        self.addCleanup(lambda: shutil.rmtree(temp_dir))
        os.mkdir(os.path.join(temp_dir, 'foo'))
        os.mkdir(os.path.join(temp_dir, 'bar'))
        expect = [os.path.join(temp_dir, x, 'test.py') for x in ('foo', 'bar')]

        for x in expect:
            with open(x, 'w') as fp:
                fp.write('#!/usr/bin/python\nprint "hello"\n')
        with open(os.path.join(temp_dir, 'README'), 'w') as fp:
            fp.write('Empty.')
        os.mkdir(os.path.join(temp_dir, 'baz'))
        with open(os.path.join(temp_dir, 'baz', 'README'), 'w') as fp:
            fp.write('Empty.')

        self.assertEqual(sorted(run.get_tests(temp_dir)), sorted(expect))

    def test_install_amulet(self):
        fixtures.set_uniq_cwd(self)
        with mock.patch.dict('os.environ', {'PYTHONPATH': ''}):
            amulet_dir = run.install_amulet('.')
            self.addCleanup(shutil.rmtree, amulet_dir)
            self.assertTrue(os.path.exists(os.path.join(amulet_dir, '.bzr')))
            self.assertEqual(amulet_dir, os.environ['PYTHONPATH'])

    @mock.patch('run._get_control_bucket',
                return_value='juju-d911b43e427c204cfcfd8c7276c4b2aa')
    def test_needs_bootstrap(self, mock_get_control_bucket):
        provider_state = '{ state-instances: ["abcdefg"] }'
        with mock.patch('run._get_from_swift', return_value=provider_state):
            # Bootstrap needed - the bucket exists but the instance does not.
            e = NotFound('Four oh four')
            with mock.patch('run.get_bootstrap_ip', side_effect=e):
                with mock.patch('run._delete_control_bucket') as dcb:
                    self.assertTrue(run.needs_bootstrap())
                    dcb.assert_called_with()

            # Bootstrap not needed.
            ip_addr = '127.0.0.1'
            with mock.patch('run.get_bootstrap_ip', return_value=ip_addr):
                self.assertFalse(run.needs_bootstrap())

        # Bootstrap needed - the bucket does not exist.
        e = swiftclient.exceptions.ClientException('not found')
        with mock.patch('run._get_from_swift', side_effect=e):
            with mock.patch('run._delete_control_bucket') as dcb:
                self.assertTrue(run.needs_bootstrap())
                dcb.assert_called_with()

    def test_check_sudoers(self):
        with tempfile.NamedTemporaryFile() as fp:
            # Empty.
            self.assertFalse(run.check_sudoers(fp.name))

            # With correct contents.
            fp.write('ev %s' % run.SUDOERS)
            fp.flush()
            self.assertTrue(run.check_sudoers(fp.name))

            os.fchmod(fp.fileno(), stat.S_IWUSR)
            self.assertTrue(run.check_sudoers(fp.name))

    def test_check_juju_version(self):
        with mock.patch('subprocess.check_output', return_value='1.16.4'):
            self.assertFalse(run.check_juju_version())
        with mock.patch('subprocess.check_output', return_value='1.16.5'):
            self.assertTrue(run.check_juju_version())

    def test_find_failed_units(self):
        # No services.
        fake_status = {'services': {}, 'machines': {}}
        fake_status = yaml.safe_dump(fake_status)
        with mock.patch('subprocess.check_output', return_value=fake_status):
            self.assertEqual([x for x in run.find_failed_units()], [])

        # One service in an error state.
        fake_status = {
            'services': {'ci-juju-gui':
                         {'units': {
                          'ci-juju-gui/0': {'agent-state': 'error'},
                          'ci-juju-gui/1': {'agent-state': 'started'},
                          }}
                         },
            'machines': {}}
        fake_status = yaml.safe_dump(fake_status)
        with mock.patch('subprocess.check_output', return_value=fake_status):
            result = [x for x in run.find_failed_units()]
        self.assertEqual(result, ['ci-juju-gui/0'])

if __name__ == '__main__':
    unittest.main()
