#!/usr/bin/env python

import json
import os
import unittest

import httplib2

from tests.deployers import DEPLOYER_DIR, DeployerTest


class TicketSystemTest(DeployerTest):
    """Integration tests for ticket-system service run on a juju deployment"""

    deployer_cfg = os.path.join(DEPLOYER_DIR, 'ticket-system.yaml')

    def get_server_status_and_content(self, url):
        final_url = url.format(*self.get_ip_and_port('ts-django'))

        # Issue a simple command against the ticket API
        client = httplib2.Http()
        resp, content = client.request(final_url, 'GET')
        jcontent = json.loads(content)
        return resp, jcontent

    def create_ticket(self, ticket_content):
        url = 'http://{}:{}/api/v1/ticket/'
        final_url = url.format(*self.get_ip_and_port('ts-django'))

        client = httplib2.Http()
        resp, content = client.request(
            uri=final_url,
            method='POST',
            headers={'Content-Type': 'application/json'},
            body=json.dumps(ticket_content),
        )
        return resp, content

    def test_ticket_sys_api(self):
        url = 'http://{}:{}/api/v1/ticket'
        resp, content = self.get_server_status_and_content(url)

        self.assertEqual('200', resp['status'])

    def test_create_ticket(self):
        ticket_content = {"owner": "create@example.com",
                          "title": "Created ticket",
                          "description": "This ticket is for creation test",
                          "bug_id": "111111",
                          "added_binaries": "binary-1,binary-2",
                          "removed_binaries": "binary-3,binary-4"}
        resp, content = self.create_ticket(ticket_content)
        self.assertEqual('201', resp['status'])

    def test_prohibit_delete(self):
        ticket_content = {'owner': "dont_delete@example.com",
                          "title": "Delete test ticket",
                          "description": "Attempt to delete ticket details",
                          "bug_id": "22222",
                          "added_binaries": "binary-4-delete-1",
                          "removed_binaries": "binary-4-delete-2"}
        resp, content = self.create_ticket(ticket_content)
        self.assertEqual('201', resp['status'])
        ticket_loc = resp['location']

        client = httplib2.Http()
        resp, content = client.request(
            uri=ticket_loc,
            method='DELETE',
        )
        self.assertEqual('405', resp['status'])

        resp, content = self. get_server_status_and_content(ticket_loc)
        self.assertEqual('200', resp['status'])

    def test_prohibit_change_ticket_details(self):
        ticket_content = {'owner': "before_change@example.com",
                          "title": "Change test ticket",
                          "description": "Attempt to change ticket details",
                          "bug_id": "33333",
                          "added_binaries": "binary-4-change-1",
                          "removed_binaries": "binary-4-change-2"}
        resp, content = self.create_ticket(ticket_content)
        self.assertEqual('201', resp['status'])
        ticket_loc = resp['location']

        client = httplib2.Http()
        resp, content = client.request(
            uri=ticket_loc,
            method='PATCH',
            headers={'Content-Type': 'application/json'},
            body=json.dumps({'owner': "after_change@example.com"}),
        )
        self.assertEqual('405', resp['status'])

        resp, content = self. get_server_status_and_content(ticket_loc)
        self.assertEqual('200', resp['status'])
        self.assertEqual('before_change@example.com', content['owner'])

if __name__ == "__main__":
    unittest.main()
