# Ubuntu Continuous Integration Engine
# Copyright 2013, 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django.test import TestCase
from ci_utils.tastypie.test import TastypieTestCase
from people.models import Person


def load_tests(loader, std_tests, pattern):
    from ticket_system import settings
    if 'people' in settings.LOCAL_APPS:
        return std_tests
    else:
        # if people is not in LOCAL_APPS the tests cannot pass so get rid of
        # them by returning an empty test suite.
        return loader.suiteClass()


def create_person(name="Johnny Test",
                  email="johnny.test@example.com", is_team=False):
    person = Person()
    person.name = name
    person.email = email
    person.is_team = is_team
    person.save()
    return person


class PersonModelTest(TestCase):

    def test_creating_a_person(self):
        person = create_person()
        people_in_database = Person.objects.all()
        self.assertEquals(len(people_in_database), 1)
        only_person_in_database = people_in_database[0]
        self.assertEquals(only_person_in_database, person)

        self.assertEquals(only_person_in_database.name, "Johnny Test")
        self.assertEquals(only_person_in_database.email,
                          "johnny.test@example.com")


class PersonResourceTest(TastypieTestCase):

    def setUp(self):
        super(PersonResourceTest, self).setUp('/api/v1')
        create_person()
        self.person_1 = Person.objects.get(name='Johnny Test')
        self.detail_url = 'person/{0}/'.format(self.person_1.pk)
        self.post_person_data = {
            'name': 'Judy Test',
            'email': 'judy.test@example.com',
        }
        self.post_team_data = {
            'name': 'CI Test Team',
            'email': 'ci.team@example.com',
            'is_team': 'True',
        }

    def test_get_person_list_json(self):
        obj = self.getResource('person/')
        self.assertEqual(obj['objects'][0], {
            u'email': u'johnny.test@example.com',
            u'id': self.person_1.pk,
            u'is_team': False,
            u'name': u'Johnny Test',
            u'resource_uri': u'/api/v1/person/{0}/'.format(self.person_1.pk)
        })

    def test_get_person_detail_json(self):
        obj = self.getResource(self.detail_url)

        self.assertKeys(obj,
                        ['email', 'id', 'is_team', 'name', 'resource_uri'])
        self.assertEqual(obj['name'], 'Johnny Test')

    def test_post_person_list(self):
        # Check how many are there first.
        self.assertEqual(Person.objects.count(), 1)
        self.post(resource='person/', params=self.post_person_data)
        # Verify a new one has been added.
        self.assertEqual(Person.objects.count(), 2)

    def test_post_team_list(self):
        # Check how many are there first.
        self.assertEqual(Person.objects.filter(is_team=True).count(), 0)
        self.post(resource='person/', params=self.post_team_data)
        # Verify a new one has been added.
        self.assertEqual(Person.objects.filter(is_team=True).count(), 1)

    def test_patch_detail(self):
        # Grab the current data & modify it slightly.
        original_data = self.getResource(self.detail_url)
        new_data = original_data.copy()
        new_data['email'] = 'jtest@example.com'

        self.assertEqual(Person.objects.count(), 1)
        self.patch(resource=self.detail_url, params=new_data)
        # Make sure the count hasn't changed & we did an update.
        self.assertEqual(Person.objects.count(), 1)
        # Check for updated data.
        self.assertEqual(Person.objects.get(pk=self.person_1.pk).name,
                         'Johnny Test')
        self.assertEqual(Person.objects.get(pk=self.person_1.pk).email,
                         'jtest@example.com')
        self.assertEqual(Person.objects.get(pk=self.person_1.pk).is_team,
                         False)

    def test_delete_person_not_allowed(self):
        resp = self.delete(resource=self.detail_url)
        self.assertHttpMethodNotAllowed(resp)
