# Ubuntu Continuous Integration Engine
# Copyright 2013, 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django.db import IntegrityError
from django.test import TestCase
from model_mommy import mommy
from model_mommy.recipe import Recipe, seq
from ci_utils.tastypie.test import TastypieTestCase
import project
from project.models import SourcePackage, BinaryPackage
from ucitests import styles


sourcepackage_recipe = Recipe(SourcePackage, name=seq('foobar'))


class SourcePackageModelTest(TestCase):

    def setUp(self):
        self.sourcepackage = sourcepackage_recipe.make()

    def test_creating_sourcepackage(self):
        package_in_database = SourcePackage.objects.all()
        self.assertEquals(len(package_in_database), 1)
        only_package_in_database = package_in_database[0]
        self.assertEquals(only_package_in_database, self.sourcepackage)

        self.assertEquals(only_package_in_database.name,
                          self.sourcepackage.name)


class BinaryPackageModelTest(TestCase):

    def setUp(self):
        self.sourcepackage = sourcepackage_recipe.make()
        self.binarypackage = mommy.make('BinaryPackage')

    def test_creating_binarypackage(self):
        binarypackage_in_database = BinaryPackage.objects.all()
        self.assertEquals(len(binarypackage_in_database), 1)
        only_binarypackage_in_database = binarypackage_in_database[0]
        self.assertEquals(only_binarypackage_in_database, self.binarypackage)

        self.assertEquals(only_binarypackage_in_database.name,
                          self.binarypackage.name)


class SourcePackageResourceTest(TastypieTestCase):

    def setUp(self):
        super(SourcePackageResourceTest, self).setUp('/api/v1')
        self.sourcepackage = sourcepackage_recipe.make()
        self.detail_url = 'sourcepackage/{0}/'.format(self.sourcepackage.pk)
        self.post_sourcepackage_data = {
            'name': 'ubuntu-ci',
        }

    def test_get_sourcepackage_list_json(self):
        obj = self.getResource('sourcepackage/')
        self.assertEqual(obj['objects'][0], {
            u'id': self.sourcepackage.pk,
            u'name': unicode(self.sourcepackage.name),
            u'resource_uri': u'/api/v1/sourcepackage/{0}/'.format(
                self.sourcepackage.pk)
        })

    def test_get_sourcepackage_detail_json(self):
        obj = self.getResource(self.detail_url)
        # We use ``assertKeys`` here to just verify the keys, not all the data.
        self.assertKeys(obj, ['id', 'name', 'resource_uri'])
        self.assertEqual(obj['name'], self.sourcepackage.name)

    def test_post_sourcepackage(self):
        # Check how many are there first.
        self.assertEqual(SourcePackage.objects.count(), 1)
        self.post(resource='sourcepackage/',
                  params=self.post_sourcepackage_data)
        # Verify a new one has been added.
        self.assertEqual(SourcePackage.objects.count(), 2)

    def test_post_existing_sourcepackage(self):
        self.assertEqual(SourcePackage.objects.count(), 1)
        self.post(resource='sourcepackage/',
                  params=self.post_sourcepackage_data)
        # Verify a new one has been added.
        self.assertEqual(SourcePackage.objects.count(), 2)
        self.assertRaises(IntegrityError, self.post, resource='sourcepackage/',
                          params=self.post_sourcepackage_data)

    def test_patch_detail(self):
        # Grab the current data & modify it slightly.
        original_data = self.getResource(self.detail_url)
        new_data = original_data.copy()
        new_data['name'] = 'ci-ubuntu'

        self.assertEqual(SourcePackage.objects.count(), 1)
        self.patch(resource=self.detail_url, params=new_data)
        # Make sure the count hasn't changed & we did an update.
        self.assertEqual(SourcePackage.objects.count(), 1)
        # Check for updated data.
        self.assertEqual(SourcePackage.objects.get(
                         pk=self.sourcepackage.pk).name, new_data['name'])

    def test_delete_sourcepackage_not_allowed(self):
        resp = self.delete(resource=self.detail_url)
        self.assertHttpMethodNotAllowed(resp)


class BinaryPackageResourceTest(TastypieTestCase):

    def setUp(self):
        super(BinaryPackageResourceTest, self).setUp('/api/v1')
        self.sourcepackage = sourcepackage_recipe.make()
        self.binarypackage = mommy.make('BinaryPackage')
        self.detail_url = 'binarypackage/{0}/'.format(self.binarypackage.pk)
        self.post_binarypackage_data = {'name': 'ubuntu-ci'}

    def test_get_binarypackage_list_json(self):
        obj = self.getResource('binarypackage/')
        self.assertEqual(obj['objects'][0], {
            u'id': self.binarypackage.pk,
            u'name': unicode(self.binarypackage.name),
            u'resource_uri': u'/api/v1/binarypackage/{0}/'.format(
                self.binarypackage.pk)
        })

    def test_get_binarypackage_detail_json(self):
        obj = self.getResource(self.detail_url)

        # We use ``assertKeys`` here to just verify the keys, not all the data.
        self.assertKeys(obj, ['id', 'name', 'resource_uri'])
        self.assertEqual(obj['name'], self.binarypackage.name)

    def test_post_binarypackage(self):
        # Check how many are there first.
        self.assertEqual(BinaryPackage.objects.count(), 1)
        self.post(resource='binarypackage/',
                  params=self.post_binarypackage_data)
        # Verify a new one has been added.
        self.assertEqual(BinaryPackage.objects.count(), 2)

    def test_patch_detail(self):
        # Grab the current data & modify it slightly.
        original_data = self.getResource(self.detail_url)
        new_data = original_data.copy()
        new_data['name'] = 'ci-ubuntu'

        self.assertEqual(BinaryPackage.objects.count(), 1)
        self.patch(resource=self.detail_url, params=new_data)
        # Make sure the count hasn't changed & we did an update.
        self.assertEqual(BinaryPackage.objects.count(), 1)
        # Check for updated data.
        self.assertEqual(BinaryPackage.objects.get(
                         pk=self.binarypackage.pk).name, new_data['name'])

    def test_delete_binarypackage_not_allowed(self):
        resp = self.delete(resource=self.detail_url)
        self.assertHttpMethodNotAllowed(resp)


class TestPep8(styles.TestPep8):

    packages = [project]
    exclude = ['migrations/']


class TestPyflakes(styles.TestPyflakes):

    packages = [project]
    # pyflakes dislikes 'from module import *'
    excludes = ['project/tests/__init__.py',
                # Ignore generated migrations files
                'project/migrations/0001_initial.py']
