# Ubuntu Continuous Integration Engine
# Copyright 2013, 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django.db import models, IntegrityError
from django.conf import settings
from project.models import SourcePackage

from ci_utils.ticket_states import (
    SubTicketWorkflowStep,
    SubTicketWorkflowStepStatus,
    TicketWorkflowStep,
    TicketWorkflowStepStatus,
)


def _choices(enum):
    return [
        (int(item.value), item.title) for item in enum.items]


def get_enum_title(value, enumcls):
    return enumcls.items.mapping[value].title


class Ticket(models.Model):
    class Meta:
        db_table = 'ticket'

    SERIES_CHOICES = (
        ('PRECISE', 'Precise'),
        ('QUANTAL', 'Quantal'),
        ('SAUCY', 'Saucy'),
        ('TRUSTY', 'Trusty'),
    )
    owner = models.EmailField(max_length=254)
    title = models.CharField(max_length=4096)
    description = models.TextField()
    bug_id = models.IntegerField(null=True, blank=True)
    current_workflow_step = models.IntegerField(
        choices=_choices(TicketWorkflowStep), max_length=4096,
        default=000)
    status = models.IntegerField(
        choices=_choices(TicketWorkflowStepStatus), max_length=4096,
        default=000)
    base_image = models.CharField(max_length=4096,
                                  default=settings.BASE_IMAGE_DEFAULT)
    series = models.CharField(max_length=4096, choices=SERIES_CHOICES,
                              default=settings.SERIES_DEFAULT)
    master_ppa = models.CharField(max_length=4096,
                                  default=settings.MASTER_PPA_DEFAULT)
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)
    # Comma-separated list of binary names to install or not install when
    # building the image for ticket testing.
    added_binaries = models.CharField(max_length=4096, null=True,
                                      blank=True)
    removed_binaries = models.CharField(max_length=4096, null=True,
                                        blank=True)

    def __unicode__(self):
        return self.title

    def save(self, *args, **kwargs):
        """Overridden to validate added/removed package list formatting."""

        def is_valid_package_list(package_list_str):
            if package_list_str is None:
                return True
            if " " in package_list_str:
                raise IntegrityError("Added and removed packages have to be "
                                     "comma-separated lists with no spaces.")
            return len(package_list_str.split(",")) > 0

        if (not is_valid_package_list(self.added_binaries) or
                not is_valid_package_list(self.removed_binaries)):
            raise IntegrityError("Added and removed packages have to be "
                                 "comma-separated lists with no spaces.")
        super(Ticket, self).save(*args, **kwargs)


class SourcePackageUpload(models.Model):
    class Meta:
        db_table = 'sourcepackageupload'

    sourcepackage = models.ForeignKey(SourcePackage)
    version = models.CharField(max_length=4096)

    def __unicode__(self):
        return "{} {}".format(self.sourcepackage, self.version)


class SubTicket(models.Model):
    class Meta:
        db_table = 'subticket'

    current_workflow_step = models.IntegerField(
        choices=_choices(SubTicketWorkflowStep), max_length=4096,
        default=000)
    status = models.IntegerField(
        choices=_choices(SubTicketWorkflowStepStatus), max_length=4096,
        default=000)
    ticket = models.ForeignKey(Ticket)
    assignee = models.EmailField(max_length=254)
    source_package_upload = models.ForeignKey(SourcePackageUpload)

    def __unicode__(self):
        return "{} {}".format(self.ticket, self.source_package_upload)


class ArtifactBase(models.Model):
    class Meta:
        abstract = True

    # 'reference' provided by the artifact manager
    reference = models.CharField(max_length=4096)
    name = models.CharField(max_length=4096)


class SubTicketArtifact(ArtifactBase):
    SUBTICKET_ARTIFACT_TYPE = (
        ('RESULTS', 'Test Results'),
        ('SPU', 'Source Package Upload'),
        ('LOGS', 'Logs'),
    )

    class Meta:
        db_table = 'subticket_artifact'

    type = models.CharField(choices=SUBTICKET_ARTIFACT_TYPE, max_length=4096)
    subticket = models.ForeignKey(SubTicket, null=True)

    def __unicode__(self):
        return "{} {}".format(self.name, self.type)


class TicketArtifact(ArtifactBase):
    TICKET_ARTIFACT_TYPE = (
        ('RESULTS', 'Test Results'),
        ('LOGS', 'Logs'),
        ('IMAGE', 'Image'),
    )

    class Meta:
        db_table = 'ticket_artifact'

    type = models.CharField(choices=TICKET_ARTIFACT_TYPE, max_length=4096)
    ticket = models.ForeignKey(Ticket, null=True)

    def __unicode__(self):
        return "{} {}".format(self.name, self.type)
