# Ubuntu Continuous Integration Engine
# Copyright 2013, 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django.db import IntegrityError
from django.test import TestCase
from project.models import SourcePackage
from ticket.models import (TicketArtifact, SubTicketArtifact, SubTicket,
                           SourcePackageUpload, SubTicketWorkflowStep,
                           SubTicketWorkflowStepStatus, Ticket,
                           TicketWorkflowStep, TicketWorkflowStepStatus)


def create_sourcepackage(name='my-package'):
    sourcepackage = SourcePackage()
    sourcepackage.name = name
    sourcepackage.save()
    return sourcepackage


def create_sourcepackageupload(sourcepackage='my-package', version='1.1'):
    spu = SourcePackageUpload()
    spu.sourcepackage = sourcepackage
    spu.version = version
    spu.save()
    return spu


def create_ticket(title="This is my ticket",
                  description="this is my ticket description",
                  bug_id='12345', owner='test@example.com',
                  current_workflow_step=TicketWorkflowStep.PKG_BUILDING,
                  status=TicketWorkflowStepStatus.PKG_BUILDING_INPROGRESS,
                  base_image='17', added_binaries="foo,bar", series='saucy',
                  removed_binaries="baz", master_ppa='ci-engine/ppa'):
    ticket = Ticket()
    ticket.owner = owner
    ticket.title = title
    ticket.description = description
    ticket.bug_id = bug_id
    ticket.current_workflow_step = current_workflow_step
    ticket.status = status
    ticket.base_image = base_image
    ticket.added_binaries = added_binaries
    ticket.removed_binaries = removed_binaries
    ticket.series = series
    ticket.master_ppa = master_ppa
    ticket.save()
    return ticket


def create_subticket(current_workflow_step=SubTicketWorkflowStep.PKG_BUILDING,
                     ticket=None, status=(
                         SubTicketWorkflowStepStatus.PKG_BUILDING_INPROGRESS),
                     assignee='test@example.com', source_package_upload=None):
    subticket = SubTicket()
    subticket.current_workflow_step = current_workflow_step
    subticket.status = status
    subticket.ticket = ticket
    subticket.assignee = assignee
    subticket.source_package_upload = source_package_upload
    subticket.save()
    return subticket


def create_ticketartifact(name="test.log", ticket=None, reference='123abc'):
    artifact = TicketArtifact()
    artifact.name = name
    artifact.ticket = ticket
    artifact.reference = reference
    artifact.save()
    return artifact


def create_subticketartifact(name="test.log", subticket=None,
                             reference='123abc'):
    artifact = SubTicketArtifact()
    artifact.name = name
    artifact.subticket = subticket
    artifact.reference = reference
    artifact.save()
    return artifact


class TicketModelsTestCase(TestCase):
    def setUp(self):
        self.sourcepackage = create_sourcepackage()
        self.spu = create_sourcepackageupload(
            sourcepackage=self.sourcepackage)
        self.ticket = create_ticket(owner='test@example.com')
        self.subticket = create_subticket(ticket=self.ticket,
                                          assignee='test@example.com',
                                          source_package_upload=self.spu)

    def test_creating_an_artifact_subticket(self):
        self.artifact_1 = create_subticketartifact(subticket=self.subticket)
        artifact_in_database = SubTicketArtifact.objects.exclude(
            subticket=None)
        self.assertEquals(len(artifact_in_database), 1)
        only_artifact_in_database = artifact_in_database[0]
        self.assertEquals(only_artifact_in_database, self.artifact_1)

        self.assertEquals(only_artifact_in_database.name, "test.log")

    def test_creating_an_artifact_ticket(self):
        self.artifact_2 = create_ticketartifact(ticket=self.ticket)
        artifact_in_database = TicketArtifact.objects.exclude(ticket=None)
        self.assertEquals(len(artifact_in_database), 1)
        only_artifact_in_database = artifact_in_database[0]
        self.assertEquals(only_artifact_in_database, self.artifact_2)

        self.assertEquals(only_artifact_in_database.name, "test.log")

    def test_creating_a_ticket(self):
        ticket_in_database = Ticket.objects.all()
        self.assertEquals(len(ticket_in_database), 1)
        only_ticket_in_database = ticket_in_database[0]
        self.assertEquals(only_ticket_in_database, self.ticket)

        self.assertEquals(only_ticket_in_database.title, "This is my ticket")

        self.assertEquals(only_ticket_in_database.added_binaries, "foo,bar")
        self.assertEquals(only_ticket_in_database.removed_binaries, "baz")

    def test_creating_a_subticket(self):
        subticket_in_database = SubTicket.objects.all()
        self.assertEquals(len(subticket_in_database), 1)
        only_subticket_in_database = subticket_in_database[0]
        self.assertEquals(only_subticket_in_database, self.subticket)

        self.assertEquals(only_subticket_in_database.assignee,
                          "test@example.com")

    def test_creating_a_sourcepackageupload(self):
        spu_in_database = SourcePackageUpload.objects.all()
        self.assertEquals(len(spu_in_database), 1)
        only_spu_in_database = spu_in_database[0]
        self.assertEquals(only_spu_in_database, self.spu)

        self.assertEquals(only_spu_in_database.sourcepackage.name,
                          "my-package")

    def test_added_binary_list_invalid(self):
        added_binaries = "foo bar"
        self.assertRaises(IntegrityError, create_ticket,
                          added_binaries=added_binaries)
