# Ubuntu Continuous Integration Engine
# Copyright 2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from model_mommy import mommy
from model_mommy.recipe import Recipe, seq
from ci_utils.tastypie.test import TastypieTestCase
from project.models import SourcePackage
from ticket.models import (TicketArtifact, SourcePackageUpload, SubTicket,
                           SubTicketWorkflowStep, SubTicketWorkflowStepStatus,
                           Ticket, TicketWorkflowStep, SubTicketArtifact,
                           TicketWorkflowStepStatus)
from project.models import BinaryPackage


def create_ticket(added=None, removed=None):
    return mommy.make('Ticket', removed_binaries=removed, added_binaries=added)


sourcepackage_recipe = Recipe(SourcePackage, name=seq('foobar'))


class APICreateTicketResourceTest(TastypieTestCase):

    def setUp(self):
        super(APICreateTicketResourceTest, self).setUp('/api/v1')
        self.ticket = create_ticket()
        self.resource = 'ticket/'
        self.detail_url = self.resource + '{0}/'.format(self.ticket.pk)
        self.post_ticket_data = {
            'owner': 'owner@example.com',
            'title': 'My first ticket!',
            'description': 'This if my first ticket. See what it can do',
            'bug_id': '12345',
            'added_binaries': 'binary-1,binary-2',
            'removed_binaries': 'binary-3,binary-4',
        }

    def test_post_ticket(self):
        # Check how many are there first.
        self.assertEqual(Ticket.objects.count(), 1)
        self.post(resource=self.resource, params=self.post_ticket_data)
        # Verify a new one has been added.
        self.assertEqual(Ticket.objects.count(), 2)

    def test_patch_detail(self):
        # Grab the current data & modify it slightly.
        original_data = self.getResource(self.detail_url)
        new_data = original_data.copy()
        new_data['owner'] = 'ci@example.com'

        self.assertEqual(Ticket.objects.count(), 1)
        resp = self.client.patch('/api/v1/' + self.detail_url, data=new_data)
        self.assertHttpMethodNotAllowed(resp)
        # Make sure the count hasn't changed & we did an update.
        self.assertEqual(Ticket.objects.count(), 1)
        # Check for updated data.
        self.assertEqual(Ticket.objects.get(pk=self.ticket.pk).owner,
                         self.ticket.owner)

    def test_delete_ticket_not_allowed(self):
        resp = self.delete(resource=self.detail_url)
        self.assertHttpMethodNotAllowed(resp)


class APICreateSPUResourceTest(TastypieTestCase):

    def setUp(self):
        super(APICreateSPUResourceTest, self).setUp('/api/v1')
        self.sourcepackage = sourcepackage_recipe.make()
        self.spu = mommy.make('SourcePackageUpload',
                              sourcepackage=self.sourcepackage)
        self.resource = 'spu/'
        self.detail_url = self.resource + '{0}/'.format(self.spu.pk)
        self.sourcepackage_uri = '/api/v1/sourcepackage/{0}/'.format(
            self.sourcepackage.pk)
        self.post_spu_data = {
            'sourcepackage': self.sourcepackage_uri,
            'version': '1.0',
        }

    def test_post_spu(self):
        # Check how many are there first.
        self.assertEqual(SourcePackageUpload.objects.count(), 1)
        self.post(resource=self.resource, params=self.post_spu_data)
        # Verify a new one has been added.
        self.assertEqual(SourcePackageUpload.objects.count(), 2)

    def test_patch_detail(self):
        # Grab the current data & modify it slightly.
        original_data = self.getResource(self.detail_url)
        new_data = original_data.copy()
        new_data['version'] = '2.0'

        self.assertEqual(SourcePackageUpload.objects.count(), 1)
        resp = self.client.patch('/api/v1/' + self.detail_url, data=new_data)
        self.assertHttpMethodNotAllowed(resp)
        # Make sure the count hasn't changed & we did an update.
        self.assertEqual(SourcePackageUpload.objects.count(), 1)
        # Check for updated data.
        self.assertEqual(SourcePackageUpload.objects.get(
                         pk=self.spu.pk).version, self.spu.version)

    def test_delete_spu_not_allowed(self):
        resp = self.delete(resource=self.detail_url)
        self.assertHttpMethodNotAllowed(resp)


class APICreateSubTicketResourceTest(TastypieTestCase):

    def setUp(self):
        super(APICreateSubTicketResourceTest, self).setUp('/api/v1')
        self.ticket = mommy.make('Ticket')
        self.sourcepackage = sourcepackage_recipe.make()
        self.spu = mommy.make('SourcePackageUpload',
                              sourcepackage=self.sourcepackage)
        self.subticket = mommy.make('SubTicket', ticket=self.ticket,
                                    source_package_upload=self.spu)
        self.resource = 'subticket/'
        self.detail_url = self.resource + '{0}/'.format(self.subticket.pk)
        self.ticket_uri = '/api/v1/ticket/{0}/'.format(self.ticket.pk)
        self.spu_uri = '/api/v1/spu/{0}/'.format(self.spu.pk)
        self.post_subticket_data = {
            'source_package_upload': self.spu_uri,
            'ticket': self.ticket_uri,
            'assignee': 'test@example.com',
        }

    def test_post_subticket_subticket(self):
        # Check how many are there first.
        self.assertEqual(SubTicket.objects.count(), 1)
        self.post(resource=self.resource, params=self.post_subticket_data)
        # Verify a new one has been added.
        self.assertEqual(SubTicket.objects.count(), 2)

    def test_patch_detail(self):
        # Grab the current data & modify it slightly.
        original_data = self.getResource(self.detail_url)
        new_data = original_data.copy()
        new_data['assignee'] = 'me@example.com'

        self.assertEqual(SubTicket.objects.count(), 1)
        resp = self.client.patch('/api/v1/' + self.detail_url, data=new_data)
        self.assertHttpMethodNotAllowed(resp)
        # Make sure the count hasn't changed & we did an update.
        self.assertEqual(SubTicket.objects.count(), 1)
        # Check for updated data.
        self.assertEqual(SubTicket.objects.get(pk=self.subticket.pk).assignee,
                         self.subticket.assignee)

    def test_delete_subticket_not_allowed(self):
        resp = self.delete(resource=self.detail_url)
        self.assertHttpMethodNotAllowed(resp)


class APICreateTicketArtifactResourceTest(TastypieTestCase):

    def setUp(self):
        super(APICreateTicketArtifactResourceTest, self).setUp('/api/v1')
        self.ticket = mommy.make('Ticket')
        self.artifact = mommy.make('TicketArtifact',
                                   ticket=self.ticket)
        self.resource = 'ticketartifact/'
        self.detail_url = self.resource + '{0}/'.format(self.artifact.pk)
        self.ticket_uri = '/api/v1/ticket/{0}/'.format(self.ticket.pk)
        self.post_artifact_data_ticket = {
            'type': 'LOGS',
            'name': 'my_artifact',
            'ticket': self.ticket_uri,
            'reference': 'http://www.example.com/m1CTLg5FHY',
        }

    def test_post_artifact_ticket(self):
        """
        Test creating an artifact for a ticket via the API
        """
        # Check how many are there first.
        self.assertEqual(TicketArtifact.objects.count(), 1)
        self.post(resource=self.resource,
                  params=self.post_artifact_data_ticket)
        # Verify a new one has been added.
        self.assertEqual(TicketArtifact.objects.count(), 2)

    def test_patch_detail(self):
        # Grab the current data & modify it slightly.
        original_data = self.getResource(self.detail_url)
        new_data = original_data.copy()
        new_data['name'] = 'my_upload'

        self.assertEqual(TicketArtifact.objects.count(), 1)
        resp = self.client.patch('/api/v1/' + self.detail_url, data=new_data)
        self.assertHttpMethodNotAllowed(resp)
        # Make sure the count hasn't changed & we did an update.
        self.assertEqual(TicketArtifact.objects.count(), 1)
        # Check for updated data.
        self.assertEqual(TicketArtifact.objects.get(
                         pk=self.artifact.pk).name, self.artifact.name)

    def test_delete_artifact_not_allowed(self):
        resp = self.delete(resource=self.detail_url)
        self.assertHttpMethodNotAllowed(resp)


class APICreateSubTicketArtifactResourceTest(TastypieTestCase):

    def setUp(self):
        super(APICreateSubTicketArtifactResourceTest, self).setUp('/api/v1')
        self.sourcepackage = sourcepackage_recipe.make()
        self.spu = mommy.make('SourcePackageUpload',
                              sourcepackage=self.sourcepackage)
        self.ticket = mommy.make('Ticket')
        self.subticket = mommy.make('SubTicket', ticket=self.ticket,
                                    source_package_upload=self.spu)
        self.artifact = mommy.make('SubTicketArtifact',
                                   subticket=self.subticket)
        self.resource = 'subticketartifact/'
        self.detail_url = self.resource + '{0}/'.format(self.artifact.pk)
        self.subticket_uri = '/api/v1/subticket/{0}/'.format(self.subticket.pk)
        self.post_artifact_data_subticket = {
            'type': 'SPU',
            'name': 'my_artifact',
            'subticket': self.subticket_uri,
            'reference': 'http://www.example.com/m1CTLg5FHY',
        }

    def test_post_artifact_subticket(self):
        """
        Test creating an artifact for a subticket via the API
        """
        # Check how many are there first.
        self.assertEqual(SubTicketArtifact.objects.count(), 1)
        self.post(resource=self.resource,
                  params=self.post_artifact_data_subticket)
        # Verify a new one has been added.
        self.assertEqual(SubTicketArtifact.objects.count(), 2)

    def test_patch_detail(self):
        # Grab the current data & modify it slightly.
        original_data = self.getResource(self.detail_url)
        new_data = original_data.copy()
        new_data['name'] = 'my_upload'

        self.assertEqual(SubTicketArtifact.objects.count(), 1)
        resp = self.client.patch('/api/v1/' + self.detail_url, data=new_data)
        self.assertHttpMethodNotAllowed(resp)
        # Make sure the count hasn't changed & we did an update.
        self.assertEqual(SubTicketArtifact.objects.count(), 1)
        # Check for updated data.
        self.assertEqual(SubTicketArtifact.objects.get(
                         pk=self.artifact.pk).name, self.artifact.name)

    def test_delete_artifact_not_allowed(self):
        resp = self.delete(resource=self.detail_url)
        self.assertHttpMethodNotAllowed(resp)


class APIUpdateTicketStatuses(TastypieTestCase):

    def setUp(self):
        super(APIUpdateTicketStatuses, self).setUp('/api/v1')
        mommy.make('BinaryPackage', name='binary3')
        mommy.make('BinaryPackage', name='binary4')
        mommy.make('BinaryPackage', name='binary5')
        self.ticket = create_ticket(added='binary1,binary2,binary6',
                                    removed='binary3,binary4')
        self.ticket_detail_url = 'updateticketstatus/{0}/'.format(
            self.ticket.pk)

    def test_patch_ticket_status(self):
        new_data = {
            'current_workflow_step': str(
                TicketWorkflowStep.IMAGE_BUILDING.value),
            'status': str(
                TicketWorkflowStepStatus.IMAGE_BUILDING_COMPLETED.value),
        }
        self.assertEqual(Ticket.objects.count(), 1)
        self.assertEqual(BinaryPackage.objects.count(), 3)
        binaries = BinaryPackage.objects.all()
        self.assertQuerysetEqual(binaries,
                                 ["<BinaryPackage: binary3>",
                                  "<BinaryPackage: binary4>",
                                  "<BinaryPackage: binary5>"], ordered=False)
        self.patch('/api/v1/' + self.ticket_detail_url, new_data)
        # Make sure the count hasn't changed & we did an update.
        self.assertEqual(Ticket.objects.count(), 1)
        # Check for updated data.
        self.assertEqual(BinaryPackage.objects.count(), 3)
        binaries = BinaryPackage.objects.all()
        self.assertQuerysetEqual(binaries,
                                 ["<BinaryPackage: binary3>",
                                  "<BinaryPackage: binary4>",
                                  "<BinaryPackage: binary5>"], ordered=False)
        self.assertEqual(Ticket.objects.get(
            pk=self.ticket.pk).current_workflow_step,
            int(TicketWorkflowStep.IMAGE_BUILDING))
        self.assertEqual(
            Ticket.objects.get(pk=self.ticket.pk).status,
            int(TicketWorkflowStepStatus.IMAGE_BUILDING_COMPLETED))

    def test_update_ticket_status_complete(self):
        new_data = {
            "current_workflow_step": TicketWorkflowStep.COMPLETED.value,
            "status": TicketWorkflowStepStatus.COMPLETED.value,
        }
        self.assertEqual(Ticket.objects.count(), 1)
        self.assertEqual(BinaryPackage.objects.count(), 3)
        binaries = BinaryPackage.objects.all()
        self.assertQuerysetEqual(binaries,
                                 ["<BinaryPackage: binary3>",
                                  "<BinaryPackage: binary4>",
                                  "<BinaryPackage: binary5>"], ordered=False)
        self.assertEqual(Ticket.objects.get(
            pk=self.ticket.pk).current_workflow_step,
            self.ticket.current_workflow_step)
        self.assertEqual(Ticket.objects.get(
            pk=self.ticket.pk).status,
            self.ticket.status)
        self.patch('/api/v1/' + self.ticket_detail_url, new_data)
        self.assertEqual(Ticket.objects.count(), 1)
        self.assertEqual(BinaryPackage.objects.count(), 4)
        binaries = BinaryPackage.objects.all()
        self.assertQuerysetEqual(binaries,
                                 ["<BinaryPackage: binary1>",
                                  "<BinaryPackage: binary2>",
                                  "<BinaryPackage: binary5>",
                                  "<BinaryPackage: binary6>"], ordered=False)
        self.assertEqual(Ticket.objects.get(
            pk=self.ticket.pk).current_workflow_step,
            int(TicketWorkflowStep.COMPLETED))
        self.assertEqual(Ticket.objects.get(
            pk=self.ticket.pk).status,
            int(TicketWorkflowStepStatus.COMPLETED))

    def test_update_ticket_status_failed(self):
        '''Ensure we fail a ticket properly

        Let's not assume users of this API will update the ticket-status to
        failed when a workflow step fails. Ensure the API corrects this for
        the user.
        '''
        new_data = {
            "current_workflow_step": str(TicketWorkflowStep.COMPLETED.value),
            "status": TicketWorkflowStepStatus.PKG_BUILDING_FAILED.value,
        }
        self.assertEqual(Ticket.objects.count(), 1)
        self.assertEqual(Ticket.objects.get(
            pk=self.ticket.pk).current_workflow_step,
            self.ticket.current_workflow_step)
        self.assertEqual(Ticket.objects.get(
            pk=self.ticket.pk).status,
            self.ticket.status)
        self.patch('/api/v1/' + self.ticket_detail_url, new_data)
        self.assertEqual(Ticket.objects.count(), 1)
        self.assertEqual(Ticket.objects.get(
            pk=self.ticket.pk).current_workflow_step,
            int(TicketWorkflowStep.FAILED))
        self.assertEqual(Ticket.objects.get(
            pk=self.ticket.pk).status,
            int(TicketWorkflowStepStatus.PKG_BUILDING_FAILED))

    def test_try_adding_binary_already_in_list(self):
        """
        Test what happens when a ticket attempts to add a binary to the master
        list that is already in the master list
        """
        new_data = {
            "current_workflow_step": TicketWorkflowStep.COMPLETED.value,
            "status": TicketWorkflowStepStatus.COMPLETED.value,
        }
        self.ticket_2 = create_ticket(added='binary1,binary2,binary3')
        self.assertEqual(Ticket.objects.count(), 2)
        self.assertEqual(BinaryPackage.objects.count(), 3)
        binaries = BinaryPackage.objects.all()
        self.assertQuerysetEqual(binaries,
                                 ["<BinaryPackage: binary3>",
                                  "<BinaryPackage: binary4>",
                                  "<BinaryPackage: binary5>"], ordered=False)
        self.client.patch('/api/v1/updateticketstatus/{0}/'.format(
                          self.ticket_2.pk), data=new_data)
        self.assertEqual(BinaryPackage.objects.count(), 5)
        binaries = BinaryPackage.objects.all()
        self.assertQuerysetEqual(binaries,
                                 ["<BinaryPackage: binary1>",
                                  "<BinaryPackage: binary2>",
                                  "<BinaryPackage: binary3>",
                                  "<BinaryPackage: binary4>",
                                  "<BinaryPackage: binary5>"], ordered=False)

    def test_try_removing_binary_not_in_list(self):
        """
        Test what happens when a ticket attempts to remove a binary from the
        master list when it isn't in the master list
        """
        new_data = {
            "current_workflow_step": TicketWorkflowStep.COMPLETED.value,
            "status": TicketWorkflowStepStatus.COMPLETED.value,
        }
        self.ticket_2 = create_ticket(removed='binary2,binary3')
        self.assertEqual(Ticket.objects.count(), 2)
        self.assertEqual(BinaryPackage.objects.count(), 3)
        binaries = BinaryPackage.objects.all()
        self.assertQuerysetEqual(binaries,
                                 ["<BinaryPackage: binary3>",
                                  "<BinaryPackage: binary4>",
                                  "<BinaryPackage: binary5>"], ordered=False)
        self.client.patch('/api/v1/updateticketstatus/{0}/'.format(
                          self.ticket_2.pk), data=new_data)
        binaries = BinaryPackage.objects.all()
        self.assertEqual(BinaryPackage.objects.count(), 2)
        self.assertQuerysetEqual(binaries,
                                 ["<BinaryPackage: binary4>",
                                  "<BinaryPackage: binary5>"], ordered=False)

    def test_update_ticket_status_complete_404(self):
        """
        Test trying to complete the status of a ticket that doesn't exist
        """
        new_data = {
            "current_workflow_step": str(TicketWorkflowStep.COMPLETED.value),
            "status": str(TicketWorkflowStepStatus.COMPLETED.value),
        }
        self.assertEqual(Ticket.objects.count(), 1)
        self.assertEqual(BinaryPackage.objects.count(), 3)
        binaries = BinaryPackage.objects.all()
        self.assertQuerysetEqual(binaries,
                                 ["<BinaryPackage: binary3>",
                                  "<BinaryPackage: binary4>",
                                  "<BinaryPackage: binary5>"], ordered=False)
        self.assertEqual(Ticket.objects.get(
            pk=self.ticket.pk).current_workflow_step,
            self.ticket.current_workflow_step)
        self.assertEqual(Ticket.objects.get(
            pk=self.ticket.pk).status,
            self.ticket.status)
        resp = self.client.patch('updateticketstatus/17394/', data=new_data)
        self.assertEqual(Ticket.objects.count(), 1)
        self.assertEqual(BinaryPackage.objects.count(), 3)
        binaries = BinaryPackage.objects.all()
        self.assertQuerysetEqual(binaries,
                                 ["<BinaryPackage: binary3>",
                                  "<BinaryPackage: binary4>",
                                  "<BinaryPackage: binary5>"], ordered=False)
        self.assertEqual(resp.status_code, 404)

    def test_delete_ticket_not_allowed(self):
        resp = self.delete(resource=self.ticket_detail_url)
        self.assertHttpMethodNotAllowed(resp)


class APIUpdateSubTicketStatuses(TastypieTestCase):

    def setUp(self):
        super(APIUpdateSubTicketStatuses, self).setUp('/api/v1')
        self.sourcepackage = sourcepackage_recipe.make()
        self.spu = mommy.make('SourcePackageUpload',
                              sourcepackage=self.sourcepackage)
        self.subticket = mommy.make('SubTicket',
                                    source_package_upload=self.spu)
        self.subticket_detail_url = (
            'updatesubticketstatus/{0}/'.format(self.subticket.pk))

    def test_patch_subticket_status(self):
        new_data = {
            'current_workflow_step': int(SubTicketWorkflowStep.PKG_BUILDING),
            'status': int(SubTicketWorkflowStepStatus.PKG_BUILDING_WAITING),
        }

        self.assertEqual(SubTicket.objects.count(), 1)
        self.patch('/api/v1/' + self.subticket_detail_url,
                   new_data)
        # Make sure the count hasn't changed & we did an update.
        self.assertEqual(SubTicket.objects.count(), 1)
        # Check for updated data.
        self.assertEqual(SubTicket.objects.get(
                         pk=self.subticket.pk).current_workflow_step,
                         int(SubTicketWorkflowStep.PKG_BUILDING))
        self.assertEqual(SubTicket.objects.get(pk=self.subticket.pk).status,
                         int(SubTicketWorkflowStepStatus.PKG_BUILDING_WAITING))

    def test_delete_subticket_not_allowed(self):
        resp = self.delete(resource=self.subticket_detail_url)
        self.assertHttpMethodNotAllowed(resp)
